/* abstract-document.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * The base class for text document
 */
public abstract class Valide.AbstractTextDocument : Gtk.VBox, iDocument
{
  /**
   * The text view
   */
  public SplitSourceView split_view;

  private bool _save;
  private uint64 mtime;
  private Searching searching;
  private DocumentState _state;
  private GotoDialog goto_dialog = null;

  /**
   * The cursor_moved signal will be emitted when the user moved the caret
   */
  public signal void cursor_moved (int row, int col);

  /**
   * The function create an appropriete source view for display a content of a
   * file
   */
  protected abstract SplitSourceView create_source_view ();

  /**
   * The SourceBuffer
   */
  public SourceBuffer buffer
  {
    get
    {
      return this.text_view.buffer;
    }
  }

  /**
   * The current focus text view
   */
  public SourceView text_view
  {
    get
    {
      return this.split_view.active_view;
    }
  }

  /**
   * First text view of SplitView (on the top)
   */
  public SourceView view1
  {
    get
    {
      return this.split_view.view1;
    }
  }

  /**
   * Second text view of SplitView (on the bottom)
   */
  public SourceView view2
  {
    get
    {
      return this.split_view.view2;
    }
  }

  /**
   * @see Valide.iDocument.is_save
   */
  public bool is_save
  {
    get
    {
      return this._save;
    }
    set
    {
      this._save = value;
      if (value)
      {
        this.saved ();
      }
      this.title_change ();
    }
  }

  /**
   * @see Vaide.iDocument.state
   */
  public new DocumentState state
  {
    get
    {
      return this._state;
    }
    set
    {
      this._state = value;
      this.title_change ();
      switch (this.state)
      {
        case DocumentState.NORMAL:
    	    this.editable = true;
        break;
        case DocumentState.EXTERNALLY_MODIFIED:
        case DocumentState.BAD_ENCODING:
    	    this.editable = false;
        break;
      }
      this.state_change (this.state);
    }
  }

  /**
   * The document is editable?
   */
  public bool editable
  {
    get
    {
      return this.text_view.get_editable ();
    }
    set
    {
      this.view1.set_editable (value);
      this.view2.set_editable (value);
    }
  }

  /**
   * The current selected text
   */
  public string selected_text
  {
    owned get
    {
      string text = "";
      Gtk.TextIter start;
      Gtk.TextIter end;

      if (this.buffer.get_selection_bounds (out start, out end))
      {
        text = this.buffer.get_text (start, end, false);
      }
      return text;
    }
  }

  /*
   * Note: When the path is modified, the signal title_change has emitted and
   *       the old path is store in the object data.
   */
  /**
   * @see Valide.iDocument.path
   */
  public new string path
  {
    owned get
    {
      return this.buffer.path;
    }

    set
    {
      string old_path;

      old_path = this.buffer.path;
      this.set_data ("old_path", old_path);
      this.buffer.path = value;
      this.title_change ();
      this.set_data ("old_path", null);
    }

    default = _("New file");
  }

  /**
   * @see Vaide.iDocument.filename
   */
  public string filename
  {
    owned get
    {
      return Path.get_basename (this.path);
    }
  }

  /**
   * @see Valide.iDocument.title
   */
  public string title
  {
    owned get
    {
      string text = "";

      if (this.buffer.get_modified ())
      {
        text = "* ";
      }
      text += this.filename;
      return text;
    }
  }

  private void check_externally_modified ()
  {
    if (this.path != _("New file") && this.state == DocumentState.NORMAL)
    {
      if (this.mtime < Utils.get_mtime (this.path))
      {
        DocumentMessage msg;

        this.state = DocumentState.EXTERNALLY_MODIFIED;
        msg = new DocumentMessage (this);
        this.pack_start (msg, false, true, 0);
        this.reorder_child (msg, 0);
        msg.show_all ();
      }
    }
  }

  private void scroll_to_iter (Gtk.TextIter iter)
  {
    Gtk.TextMark mark;

    mark = this.buffer.create_mark ("cursor", iter, true);
    this.text_view.scroll_to_mark (mark, 0.4, false, 0, 0);
    this.buffer.delete_mark (mark);
  }

  private void cursor_move ()
  {
	  int row;
	  int col;
    Gtk.TextIter iter;
	  Gtk.TextIter start;
	  uint tab_size;

    buffer.get_iter_at_mark (out iter, this.buffer.get_insert ());

    row = iter.get_line ();

    start = iter;
    start.set_line_offset (0);
    col = 0;

    tab_size = this.text_view.get_tab_width ();

    while (!start.equal (iter))
    {
      /* FIXME: Are we Unicode compliant here? */
      if (start.get_char () == '\t')
      {
        col += (int)(tab_size - (col  % tab_size));
      }
      else
      {
        col++;
      }

      start.forward_char ();
    }

    this.cursor_moved (row + 1, col + 1);
  }

  /**
   * Return the entire file content in UTF-8 encoding
   *
   * @return The content of the document
   */
  protected virtual string get_contents () throws Error
  {
    return this.buffer.get_contents ();
  }

  /**
   * Write the entire document content in file
   */
  protected virtual void set_contents (string? new_path = null) throws Error
  {
    this.buffer.set_contents (new_path);
  }

  construct
  {
    this.is_save = true;
    this.split_view = this.create_source_view ();
    this.path = _("New file");
    this.pack_start (this.split_view, true, true, 0);
    this.buffer.changed.connect (() => {
      this.cursor_move ();
    });
    this.buffer.modified_changed.connect (() => {
      this.is_save = !this.buffer.get_modified ();
      this.title_change ();
    });
    this.buffer.mark_set.connect (() => {
      this.cursor_move ();
    });

    this.searching = new Searching (this);
    this.pack_start (this.searching, false, true, 0);
    this.searching.hide ();

    this.cursor_move ();

    this.state = DocumentState.NORMAL;
    this.text_view.focus_in_event.connect (() => {
      this.check_externally_modified ();
      this.cursor_move ();
      return false;
    });
  }

  /**
   * @see Gtk.Widget.grab_focus
   */
  public new void grab_focus ()
  {
    this.text_view.grab_focus ();
  }

  /**
   * @see Gtk.Widget.show
   */
  public new void show ()
  {
    this.show_all ();
    this.searching.hide ();
  }

  /**
   * @see Gtk.Widget.show_all
   */
  public override void show_all ()
  {
    base.show_all ();
    this.searching.hide ();
  }

  /**
   * Refresh the content of the widget with the hard disk file
   */
  public void reload ()
  {
    string contents = "";

    try
    {
      contents = this.get_contents ();
    }
    catch (Error e)
    {
      DocumentMessage msg;

      this.state = DocumentState.BAD_ENCODING;
      msg = new DocumentMessage (this);
      this.pack_start (msg, false, true, 0);
      this.reorder_child (msg, 0);
      msg.show_all ();
    }
    finally
    {
      Gtk.TextIter start;
      Gtk.TextIter end;

      this.buffer.get_start_iter (out start);
      this.buffer.get_end_iter (out end);
      this.buffer.begin_not_undoable_action ();
      this.buffer.delete (start, end);
      this.buffer.insert (start, contents, -1);
	    this.buffer.set_modified (false);
      this.buffer.end_not_undoable_action ();
      this.is_save = true;
      try
      {
        this.mtime = Utils.get_mtime (this.path);
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }
  }

  /**
   * @see Valide.iDocument.open
   */
  public void open (string path) throws DocumentError
  {
    if (FileUtils.test (path, FileTest.EXISTS))
    {
      this.path = path;
      this.reload ();
    }
    else
    {
      throw new DocumentError.BAD_URI (_("cannot open file '%s': No such file or directory"), path);
    }
  }

  /**
   * @see Valide.iDocument.save
   */
  public void save (string? current_folder = null)
  {
    if (!this.is_save)
    {
      Project project;
      int response = Gtk.ResponseType.ACCEPT;

      if (this.path == _("New file"))
      {
        Gtk.FileChooserDialog dialog;

        dialog = new Gtk.FileChooserDialog (_("Save file"), null,
                                            Gtk.FileChooserAction.SAVE,
                                            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                            Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
                                            null);

        if (current_folder != null)
        {
          dialog.set_current_folder (current_folder);
        }
        response = dialog.run ();
        if (response == Gtk.ResponseType.ACCEPT)
        {
          this.path = dialog.get_filename ();
        }
        dialog.destroy ();
      }
      if (response == Gtk.ResponseType.ACCEPT)
      {
        try
        {
          this.set_contents ();
	        this.buffer.set_modified (false);
          this.mtime = Utils.get_mtime (this.path);
          this.is_save = true;
        }
        catch (Error e)
        {
          warning (e.message);
        }
      }
    }
  }

  /**
   * @see Valide.iDocument.save_as
   */
  public void save_as (string? current_folder = null)
  {
    Project project;
    string path = null;
    int response = Gtk.ResponseType.ACCEPT;

    Gtk.FileChooserDialog dialog;

    dialog = new Gtk.FileChooserDialog (_("Save as file"), null,
                                        Gtk.FileChooserAction.SAVE,
                                        Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT,
                                        null);

    if (current_folder != null)
    {
      dialog.set_current_folder (current_folder);
    }
    response = dialog.run ();
    if (response == Gtk.ResponseType.ACCEPT)
    {
      path = dialog.get_filename ();
    }
    dialog.destroy ();

    if (response == Gtk.ResponseType.ACCEPT)
    {
      try
      {
        this.set_contents (path);
        this.path = path;
        this.buffer.set_modified (false);
        this.mtime = Utils.get_mtime (this.path);
        this.is_save = true;
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
  }

  /**
   * @see Valide.iDocument.rename
   */
  public void rename ()
  {
    Gtk.Entry entry;
    Gtk.Dialog dialog;
    Gtk.Window parent;

    dialog = new Gtk.Dialog ();
    dialog.set_title (_("Rename"));
    parent = this.get_ancestor (typeof (Gtk.Window)) as Gtk.Window;
    dialog.set_transient_for (parent);
    dialog.set_modal (true);
    dialog.add_button (Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY);
    dialog.add_button (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL);

    entry = new Gtk.Entry ();
    entry.set_data ("dialog", dialog);
    entry.activate.connect ((s) => {
      Gtk.Dialog dlg;

      dlg = s.get_data ("dialog") as Gtk.Dialog;
      dlg.response (Gtk.ResponseType.APPLY);
    });
    entry.set_text (this.path);
    dialog.vbox.pack_start (entry, true, true, 10);

    dialog.show_all ();
    if (dialog.run () == Gtk.ResponseType.APPLY)
    {
      string new_path;

      new_path = entry.get_text ();
      FileUtils.rename (this.path, new_path);
      this.path = new_path;
    }
    dialog.destroy ();
  }

  /**
   * @see Valide.iDocument.close
   */
  public bool close ()
  {
    bool is_close = true;

    if (!this.is_save)
    {
      Gtk.Window parent;
      Gtk.Dialog dialog;
      Gtk.Label label;
      int response;

      dialog = new Gtk.Dialog ();
      dialog.set_title (_("Are you sure?"));
      dialog.add_button (Gtk.STOCK_YES, Gtk.ResponseType.YES);
      dialog.add_button (Gtk.STOCK_NO, Gtk.ResponseType.NO);
      dialog.add_button (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL);
      dialog.set_modal (true);
      parent = this.get_ancestor (typeof (Gtk.Window)) as Gtk.Window;
      dialog.set_transient_for (parent);

      label = new Gtk.Label (_("The document has unsaved changes. Save changes before closing?"));
      dialog.vbox.pack_start (label, true, false, 10);
      dialog.show_all ();
      response = dialog.run ();
      dialog.destroy ();
      switch (response)
      {
        case Gtk.ResponseType.YES:
          this.save ();
          is_close = this.is_save;
        break;
        case Gtk.ResponseType.NO:
        break;
        case Gtk.ResponseType.CANCEL:
          is_close =  false;
        break;
      }
    }
    if (is_close)
    {
      this.closed ();
    }
    return is_close;
  }

  /**
   * @see Valide.iDocument.redo
   */
  public void redo ()
  {
    this.buffer.redo ();
  }

  /**
   * @see Valide.iDocument.search
   */
  public void search ()
  {
    this.searching.find ();
  }

  /**
   * @see Valide.iDocument.find_next
   */
  public void find_next ()
  {
    this.searching.find_next ();
  }

  /**
   * @see Valide.iDocument.find_prev
   */
  public void find_prev ()
  {
    this.searching.find_prev ();
  }

  /**
   * @see Valide.iDocument.replace
   */
  public void replace ()
  {
    this.searching.replace ();
  }

  /**
   * @see Valide.iDocument.goto_line
   */
  public bool goto_line (int line = -1)
  {
    bool ret = false;
	  uint line_count;
	  Gtk.TextIter iter;

	  if (line < 0)
	  {
	    if (this.goto_dialog == null)
	    {
        Gtk.Window toplevel;

        this.goto_dialog = new GotoDialog (this);
        toplevel = this.get_ancestor (typeof (Gtk.Window)) as Gtk.Window;
        if (toplevel != null)
        {
          Gtk.WindowGroup group;

          group = toplevel.group;
          if (group != null)
          {
            group.add_window (this.goto_dialog);
          }
        }
      }
	    this.goto_dialog.show ();
	  }
    else
	  {
    	line_count = this.buffer.get_line_count ();

    	if (line <= line_count)
	    {
	      this.grab_focus ();
	    	this.buffer.get_iter_at_line (out iter, line);
      	this.buffer.place_cursor (iter);
        this.scroll_to_iter (iter);
	    	ret = true;
	    }
    }
    return ret;
  }

  /**
   * @see Valide.iDocument.get_icon
   */
  public virtual Gdk.Pixbuf get_icon ()
  {
    Gdk.Pixbuf pixbuf;

    switch (this.state)
    {
      case DocumentState.NORMAL:
        try
        {
          pixbuf = Utils.get_pixbuf_for_file (this.path, Gtk.IconSize.MENU);
        }
        catch (Error e)
        {
          pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_FILE, Gtk.IconSize.MENU);
        }
      break;
      case DocumentState.EXTERNALLY_MODIFIED:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_DIALOG_WARNING, Gtk.IconSize.MENU);
      break;
      case DocumentState.BAD_ENCODING:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_DIALOG_ERROR, Gtk.IconSize.MENU);
      break;
      default:
        pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_DIALOG_QUESTION, Gtk.IconSize.MENU);
      break;
    }
    return pixbuf;
  }

  /**
   * @see Valide.iDocument.get_tooltips
   */
  public virtual string get_tooltips ()
  {
    string tooltips = null;

    if (this.path != _("New file"))
    {
      string content_type;
      string mime_type;
      string content_description;

      content_type = Utils.get_mime_type (this.path);
      content_description = g_content_type_get_description (content_type);
      if (content_description != null)
      {
        mime_type = "%s (%s)".printf (content_description, content_type);
      }
      else
      {
        mime_type = content_type;
      }

      tooltips = "<b>%s</b> %s\n\n<b>%s</b> %s\n<b>%s</b> %s".printf (_("Name:"), this.path,
			                                                                _("MIME Type:"), mime_type,
			                                                                _("Encoding:"), this.buffer.encoding);
    }
    return tooltips;
  }

  /**
   * @see Valide.iDocument.setup_ui
   */
  public void setup_ui (UIManager ui_manager)
  {
    ui_manager.get_action ("edit-undo").sensitive = this.buffer.can_undo;
    ui_manager.get_action ("edit-redo").sensitive = this.buffer.can_redo;
  }

  /**
   * Select a range
   *
   * @param pos The position of the beginning and end of the selection
   */
  public void select_range (Position pos)
  {
    Gtk.TextIter start;
    Gtk.TextIter end;

    if (pos.start_line >= 0)
    {
      this.buffer.get_start_iter (out start);
      start.set_line (pos.start_line);
      if (pos.start_row >= 0 && pos.start_row < start.get_chars_in_line ())
      {
        start.set_line_offset (pos.start_row);
      }
      end = start;
      if (pos.end_line >= 0 && pos.end_line < this.buffer.get_line_count ())
      {
        end.set_line (pos.end_line);
        if (pos.end_row >= 0 && pos.end_row < end.get_chars_in_line ())
        {
          end.set_line_offset (pos.end_row);
        }
      }

      this.buffer.select_range (start, end);
      this.scroll_to_iter (start);
    }
    else
    {
      debug (_("Invalid range"));
    }
  }
}

