/* completion.vala
 *
 * Copyright (C) 2008-2009 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * The completion manager
 */
public class Valide.Completion.Manager : Object
{
  private List<Provider> providers;

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    construct;
  }

  /**
   * The completion popup
   */
  public Popup popup
  {
    get;
    construct;
  }

  /**
   * The completion calltips
   */
  public Calltips calltips
  {
    get;
    construct;
  }

  private void get_window_position_in_cursor (Gtk.Window window,
                                              ref int x, ref int y)
  {
    int w, h, xtext, ytext;
    int sw, sh;

    sw = Gdk.Screen.width ();
    sh = Gdk.Screen.height ();

    window.get_size (out w, out h);
    if (x + w > sw) x = sw - w - 4;
    // If we cannot show it down, we show it up.
    if (y + h > sh)
    {
      Pango.Layout layout;

      layout = window.create_pango_layout (null);
      layout.get_pixel_size (out xtext, out ytext);
      y = y - ytext - h;
    }
  }

  private void get_cursor_pos (Gtk.TextView text_view, 
                               out int x, out int y)
  {
    Gdk.Window win;
    Gtk.TextMark insert_mark;
    Gtk.TextBuffer text_buffer;
    Gtk.TextIter start;
    Gdk.Rectangle location;
    int win_x, win_y;
    int xx, yy;

    text_buffer = text_view.get_buffer ();
    insert_mark = text_buffer.get_insert ();
    text_buffer.get_iter_at_mark (out start,insert_mark);
    text_view.get_iter_location (start, out location);
    text_view.buffer_to_window_coords (Gtk.TextWindowType.WIDGET, location.x,
                                       location.y, out win_x, out win_y);

    win = text_view.get_window (Gtk.TextWindowType.WIDGET);
    win.get_origin (out xx, out yy);

    x = win_x + xx;
    y = win_y + yy + location.height;
  }

  private bool key_press_event_cb (Gtk.Widget sender, Gdk.EventKey event)
  {
    int x, y;
    bool catched = false;
    SourceView source_view;

    source_view = sender as SourceView;
    /* Trigger custom keys */
    if (event.keyval == '(' && !this.calltips.visible)
    {
      this.get_cursor_pos (source_view, out x, out y);
      this.calltips.show (source_view.buffer);
      this.calltips.move (x, y);
    }
    else if (event.keyval == ')' && this.calltips.visible)
    {
      this.calltips.hide ();
    }
    else if (!this.popup.visible && event.keyval == Gdk.KeySyms.Return
             && (event.state & Gdk.ModifierType.CONTROL_MASK) == Gdk.ModifierType.CONTROL_MASK)
    {
      string filter;
      string current_word;

      this.get_cursor_pos (source_view, out x, out y);
      this.trigger_event (source_view, x, y);
      current_word = source_view.buffer.get_last_word ();
      this.popup.filter (Utils.get_last_element (current_word, '.'));
      catched = true;
    }
    else if (this.popup.visible)
    {
      switch (event.keyval)
      {
        case Gdk.KeySyms.Left:
        case Gdk.KeySyms.Right:
          catched = true;
        break;
        case Gdk.KeySyms.Down:
          this.popup.select_next (1);
          catched = true;
        break;
        case Gdk.KeySyms.Page_Down:
          this.popup.select_next (5);
          catched = true;
        break;
        case Gdk.KeySyms.Up:
          this.popup.select_previous (1);
          catched = true;
        break;
        case Gdk.KeySyms.Page_Up:
          this.popup.select_previous (5);
          catched = true;
        break;
        case Gdk.KeySyms.Home:
          this.popup.select_first ();
          catched = true;
        break;
        case Gdk.KeySyms.End:
          this.popup.select_last ();
          catched = true;
        break;
        case Gdk.KeySyms.Return:
        case Gdk.KeySyms.Tab:
          Proposal prop = null;

          if (this.popup.get_selected_proposal (out prop))
          {
            this.popup.proposal_selected (prop);
            catched = true;
          }
        break;
        case Gdk.KeySyms.BackSpace:
        case Gdk.KeySyms.Escape:
          this.popup.hide ();
        break;
      }
      if (!catched)
      {
        uint32 c;
        string word;

        word = source_view.buffer.get_last_word ();
        c = Gdk.keyval_to_unicode (event.keyval);
        if (c > 0)
        {
          word = "%s%c".printf (word, (int)c);
        }
        else if (event.keyval == Gdk.KeySyms.BackSpace)
        {
          if (word != "")
          {
            word = word.ndup (word.len () - 1);
          }
        }
        this.popup.filter (word);
        if (!this.popup.has_proposals ())
        {
          this.popup.hide ();
        }
      }
    }
    else if (this.calltips.visible)
    {
      switch (event.keyval)
      {
        case Gdk.KeySyms.BackSpace:
        case Gdk.KeySyms.Escape:
          this.calltips.hide ();
        break;
      }
    }
    return catched;
  }

  private bool focus_out_event_cb (Gtk.Widget sender, Gdk.EventFocus event)
  {
    if (this.popup.visible)
    {
      this.popup.hide ();
    }
    if (this.calltips.visible)
    {
      this.calltips.hide ();
    }
    return false;
  }

  private bool button_press_event_cb (Gtk.Widget sender, Gdk.EventButton event)
  {
    if (this.popup.visible)
    {
      this.popup.hide ();
    }
    if (this.calltips.visible)
    {
      this.calltips.hide ();
    }
    return false;
  }

  private void proposal_selected_cb (Popup sender, Proposal proposal)
  {
    sender.source_buffer.replace_actual_word (proposal.label);
    this.popup.hide ();
  }

  /**
   * Create a new Valaide.Completion.Manager
   */
  public Manager (DocumentManager documents)
  {
    this.popup = new Popup ();
    this.calltips = new Calltips ();
    this.documents = documents;
  }

  construct
  {
    this.providers = new List<Provider> ();

    this.popup.proposal_selected.connect (this.proposal_selected_cb);

    this.documents.tab_added.connect ((s, d) => {
      if (d is AbstractTextDocument)
      {
        AbstractTextDocument atd;

        atd = d as AbstractTextDocument;
        atd.view1.focus_out_event.connect (this.focus_out_event_cb);
        atd.view1.key_press_event.connect (this.key_press_event_cb);
        atd.view1.button_press_event.connect (this.button_press_event_cb);

        atd.view2.focus_out_event.connect (this.focus_out_event_cb);
        atd.view2.key_press_event.connect (this.key_press_event_cb);
        atd.view2.button_press_event.connect (this.button_press_event_cb);
      }
    });

    this.documents.tab_removed.connect ((s, d) => {
      if (d is AbstractTextDocument)
      {
        AbstractTextDocument atd;

        atd = d as AbstractTextDocument;
        atd.view1.focus_out_event.disconnect (this.focus_out_event_cb);
        atd.view1.key_press_event.disconnect (this.key_press_event_cb);
        atd.view1.button_press_event.disconnect (this.button_press_event_cb);

        atd.view2.focus_out_event.disconnect (this.focus_out_event_cb);
        atd.view2.key_press_event.disconnect (this.key_press_event_cb);
        atd.view2.button_press_event.disconnect (this.button_press_event_cb);
      }
    });
  }

  /**
   * Register a new provider
   *
   * @param provider The provider
   *
   * @return true if the provider is registered
   */
  public bool register_provider (Provider provider)
  {
    bool register = false;

    this.providers.append (provider);
    register = true;
    return register;
  }

  /**
   * Unregister a provider
   *
   * @param provider The provider
   *
   * @return true if the provider is unregistered
   */
  public bool unregister_provider (Provider provider)
  {
    bool unregister = false;

    this.providers.remove (provider);
    unregister = true;
    return unregister;
  }

  /**
   * Trigger the completion event
   *
   * @param source_view The trigger emiter
   * @param x Event x position
   * @param y Event y position
   */
  public void trigger_event (SourceView source_view, int x, int y)
  {
    List<Proposal> proposals;

    proposals = new List <Proposal> ();

    foreach (Provider provider in this.providers)
    {
      foreach (Proposal proposal in provider.get_proposals (source_view))
      {
        proposals.append (proposal);
      }
    }
    this.popup.proposals = proposals;

    this.get_window_position_in_cursor (this.popup, ref x, ref y);
    this.popup.move (x, y);
    this.popup.show (source_view.buffer);
  }
}

