/* popup.vala
 *
 * Copyright (C) 2008-2009 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * The completion popup
 */
public class Valide.Completion.Popup : Gtk.Window
{
  private const int WIDTH = 350;
  private const int HEIGHT = 200;
  private Tree completion_tree;

  /**
   * The associed source buffer
   */
  public SourceBuffer source_buffer
  {
    get;
    private set;
  }

  /**
   * @see Valide.Completion.Tree.proposal_selected
   */
  public signal void proposal_selected (Proposal proposal);

  /**
   * The list of proposals
   */
  public unowned List<Proposal> proposals
  {
    set
    {
      this.completion_tree.clear ();
      foreach (Proposal proposal in value)
      {
        this.completion_tree.add_data (proposal);
      }
    }
  }

  private bool delete_event_cb (Gtk.Widget widget, Gdk.Event  event)
  {
    // Prevent the alt+F4 keys
    return true;
  }

  private void proposal_selected_cb (Tree sender, Proposal proposal)
  {
    this.proposal_selected (proposal);
  }

  /**
   * Create a new completion popup
   */
  public Popup ()
  {
    this.type = Gtk.WindowType.POPUP;

    this.completion_tree = new Tree ();
    this.completion_tree.can_focus = false;
    this.completion_tree.proposal_selected.connect (this.proposal_selected_cb);
    this.completion_tree.show ();
    this.add (this.completion_tree);

    this.delete_event.connect (this.delete_event_cb);
  }

  /**
   * @see Gtk.Widget.show
   */
  public new void show (SourceBuffer source_buffer)
  {
    if (!this.visible && this.has_proposals ())
    {
      this.source_buffer = source_buffer;
      base.show ();
    }
  }

  /**
   * @]see Gtk.Widget.hide
   */
  public new void hide ()
  {
    this.source_buffer = null;
    base.hide ();
  }

  /**
   * @see Gtk.Widget.realize
   */
  public override void realize ()
  {
    this.set_border_width (1);
    this.set_size_request (this.WIDTH, this.HEIGHT);
    this.set_resizable (true);
    base.realize ();
  }

  /**
   * @see Valide.Completion.Tree.select_first
   */
  public void select_first ()
  {
    this.completion_tree.select_first ();
  }

  /**
   * @see Valide.Completion.Tree.select_last
   */
  public void select_last ()
  {
    this.completion_tree.select_last ();
  }

  /**
   * @see Valide.Completion.Tree.select_previous
   */
  public void select_previous (int rows)
  {
    this.completion_tree.select_previous (rows);
  }

  /**
   * @see Valide.Completion.Tree.select_next
   */
  public void select_next (int rows)
  {
    this.completion_tree.select_next (rows);
  }

  /**
   * @see Valide.Completion.Tree.get_selected_proposal
   */
  public bool get_selected_proposal (out Proposal proposal)
  {
    return this.completion_tree.get_selected_proposal (out proposal);
  }

  /**
   * @see Valide.Completion.Tree.get_num_proposals
   */
  public int get_num_proposals ()
  {
    return this.completion_tree.get_num_proposals ();
  }

  /**
   * Check if popup has proposal
   *
   * @return true if proposals are available
   */
  public bool has_proposals ()
  {
    return (this.get_num_proposals () > 0);
  }

  /**
   * @see Valide.Completion.Tree.filter
   */
  public void filter (string filter)
  {
    this.completion_tree.filter (filter);
  }
}

