/* document-manager.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public errordomain Valide.DocumentError
{
  UNKNOW_DOCUMENT,
  BAD_URI,
  NO_DOCUMENT,
  MIME
}

/**
 * The manager for opened documents
 */
public class Valide.DocumentManager : Gtk.Notebook
{
  /**
   * The UI manager
   */
  public UIManager ui_manager
  {
    get;
    construct;
  }

  /**
   * New document is adding
   */
  public signal void tab_added (iDocument document);

  /**
   * Current document is saving
   */
  public signal void tab_saved (iDocument document);

  /**
   * Current document is closing
   */
  public signal void tab_removed (iDocument document);

  /**
   * The current document changed
   */
  public signal void tab_changed (iDocument document);

  /**
   * The state of the current document changed
   */
  public signal void tab_state_changed (iDocument document);

  /**
   * The tab_cursor_moved signal will be emitted when the user moved the caret
   */
  public signal void tab_cursor_moved (iDocument document, int row, int col);

  /**
   * The current opened document.
   */
  public iDocument current
  {
    get
    {
      return this.get_nth_page (this.page) as iDocument;
    }
    set
    {
      this.set_current_page (this.page_num (value));
    }
  }

  /**
   * All opened documents.
   */
  public weak List<iDocument> documents
  {
    get;
    private set;
  }

  private void title_change (iDocument sender)
  {
    Gtk.HBox title;
    Gtk.Image image;
    Gtk.Label label;
    Gtk.Button button;

    title = new Gtk.HBox (false, 4);
    title.set_tooltip_markup (sender.get_tooltips ());

    image = new Gtk.Image.from_pixbuf (sender.get_icon ());
    title.pack_start (image, false, true, 0);

    label = new Gtk.Label (sender.title);
    title.pack_start (label, false, false, 0);

    /* Close button */
  	image = new Gtk.Image.from_stock (Gtk.STOCK_CLOSE, Gtk.IconSize.MENU);
	  button = Utils.create_small_button (image);
    button.set_data ("document", sender);
    button.clicked.connect ((s) => {
      iDocument document;

      document = s.get_data ("document") as iDocument;
      document.close ();
    });
    button.set_tooltip_text (_("Close document"));
    title.pack_start (button, false, false, 0);

    title.show_all ();
    this.set_tab_label (sender, title);
    this.tab_state_changed (sender);
  }

  private static void switch_page_cb (Gtk.Notebook sender, void *page,
                                      int page_num, DocumentManager self)
  {
    self.current.setup_ui (self.ui_manager);
    self.tab_changed (self.current);
  }

  private void launch_external (string? filename) throws Error
  {
    if (filename != null)
    {
      string msg;
      Gtk.Window toplevel;
      Gtk.MessageDialog dialog;

      msg = _("%s doesn't support this type of document, would you like open this with the default program?").printf (Config.APPNAME);

      toplevel = this.get_toplevel () as Gtk.Window;
      dialog = new Gtk.MessageDialog (toplevel, Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.QUESTION,
                                      Gtk.ButtonsType.YES_NO, msg);

      if (dialog.run () == Gtk.ResponseType.YES)
      {
        AppInfo.launch_default_for_uri ("file://" + filename, null);
      }
      dialog.destroy ();
    }
  }

  /**
   * Update the user interface
   */
  protected virtual void setup_ui ()
  {
    bool active = false;

    if (this.current != null)
    {
      active = true;
    }
    this.ui_manager.get_action ("file-save").sensitive = active;
    this.ui_manager.get_action ("file-save-as").sensitive = active;
    this.ui_manager.get_action ("file-rename").sensitive = active;
    this.ui_manager.get_action ("file-close").sensitive = active;
    this.ui_manager.get_action ("file-close-all").sensitive = active;
    this.ui_manager.get_action ("edit-copy").sensitive = active;
    this.ui_manager.get_action ("edit-cut").sensitive = active;
    this.ui_manager.get_action ("edit-paste").sensitive = active;
    this.ui_manager.get_action ("edit-search").sensitive = active;
    this.ui_manager.get_action ("edit-find-next").sensitive = active;
    this.ui_manager.get_action ("edit-find-prev").sensitive = active;
    this.ui_manager.get_action ("edit-replace").sensitive = active;
    this.ui_manager.get_action ("edit-goto-line").sensitive = active;
    this.ui_manager.get_action ("edit-goto-definition").sensitive = active;
    this.ui_manager.get_action ("edit-undo").sensitive = active;
    this.ui_manager.get_action ("edit-redo").sensitive = active;
    if (this.current != null)
    {
      this.current.setup_ui (this.ui_manager);
    }
  }

  /**
   * Create a new Valide.AbstractTextDocument for filename
   *
   * @param filename A name of file
   *
   * @throws DocumentError
   */
  protected virtual AbstractTextDocument create_document (string? filename) throws DocumentError
  {
    /* @todo Use plugins */
    Type type = 0;
    string ext = "Unknow";

    if (filename != null)
    {
      string tmp;

      tmp = Utils.get_extension (filename);
      if (tmp != null)
      {
        ext = tmp.up (1) + tmp.offset (1);
      }
    }
    else
    {
      ext = "Vala";
    }

    /* Search standard behavior */
    type = Type.from_name ("Valide" + ext + "Document");
    
    /* Or, open as text */
    if (type == 0)
    {
      string? mime;
      
      try
      {
        mime = Utils.get_mime_type (filename);
      }
      catch (Error e)
      {
        throw new DocumentError.MIME (_("Unable to retrieve mime type"));
      }

      if(mime == null)
      {
        throw new DocumentError.MIME (_("Unable to retrieve mime type"));
      }
      
      if (mime.has_prefix ("text/"))
      {
        type = typeof (TxtDocument);
      }

      /* Default system behavior */
      if (type == 0)
      {
        throw new DocumentError.UNKNOW_DOCUMENT (_("Couldn't open this document"));
      }
    }
    return Object.new (type) as AbstractTextDocument;
  }

  private void drag_data_received_cb (Gtk.Widget sender,
                                      Gdk.DragContext drag_context,
                                      int x, int y,
                                      Gtk.SelectionData data,
                                      uint info, uint time)
  {
    SList<string> files;

    files = new SList<string> ();
    foreach (string uri in ((string)data.data).split ("\r\n"))
    {
      try
      {
        if (uri != "")
        {
          File file;
          string filename;

          file = File.new_for_uri (uri);
          filename = file.get_path ();
          this.create (filename);
        }
      }
      catch (Error e)
      {
        debug (e.message);
      }
    }
  }

  private int find_tab_num_at_pos (int abs_x, int abs_y)
  {
    Gtk.Widget page;
    int page_num = 0;
    Gtk.PositionType tab_pos;

    tab_pos = this.get_tab_pos ();

    if (this.first_tab == null)
    {
      return -1;
    }

    while ((page = this.get_nth_page (page_num)) != null)
    {
      Gtk.Widget tab;
      int max_x, max_y;
      int x_root, y_root;

      tab = this.get_tab_label (page);
      return_val_if_fail (tab != null, -1);

      if (!tab.is_mapped ())
      {
        page_num++;
        continue;
      }

      tab.window.get_origin (out x_root, out y_root);

      max_x = x_root + tab.allocation.x + tab.allocation.width;
      max_y = y_root + tab.allocation.y + tab.allocation.height;

      if ((tab_pos == Gtk.PositionType.TOP || tab_pos == Gtk.PositionType.BOTTOM)
          && abs_x <= max_x)
      {
        return page_num;
      }
      else if ((tab_pos == Gtk.PositionType.LEFT || tab_pos == Gtk.PositionType.RIGHT)
               &&  abs_y <= max_y)
      {
        return page_num;
      }

      page_num++;
    }
    return -1;
  }

  private bool on_button_press (Gtk.Widget sender, Gdk.EventButton event)
  {
    bool catched = false;

    if (event.button == 3)
    {
      int tab_clicked;

      tab_clicked = this.find_tab_num_at_pos ((int)event.x_root, (int)event.y_root);
      if (tab_clicked == -1)
      {
        // CHECK: do we really need it?

        /* consume event, so that we don't pop up the context menu when
         * the mouse if not over a tab label
         */
        catched = true;
      }
      else
      {
        Gtk.Menu menu;

        /* Switch to the page the mouse is over, but don't consume the event */
        this.set_current_page (tab_clicked);

        menu = this.ui_manager.get_widget ("/document-popup") as Gtk.Menu;
        menu.popup (null, null, null, event.button, event.time);
      }
    }
    return catched;
  }

  /**
   * Create a new Valide.DocumentManager
   */
  public DocumentManager (UIManager ui_manager)
  {
    this.ui_manager = ui_manager;
    /* Force type register */
    typeof (TxtDocument);
    typeof (GsDocument);
    typeof (ValaDocument);
    typeof (VapiDocument);
    typeof (XmlDocument);
    typeof (GladeDocument);
    typeof (UiDocument);
  }

  construct
  {
    Gtk.TargetEntry[] targets;

    this.setup_ui ();
    Signal.connect_after (this, "switch-page", (Callback)switch_page_cb, this);
    this.button_press_event.connect (this.on_button_press);

    /* Seting up the drag & drop */
    targets = new Gtk.TargetEntry[1];
    targets[0].target = "text/uri-list";
    targets[0].flags = 0;
    targets[0].info = 0;

    Gtk.drag_dest_set (this, Gtk.DestDefaults.ALL, targets,
                       Gdk.DragAction.COPY);
    this.drag_data_received.connect (this.drag_data_received_cb);
  }

  /**
   * Create a new document.
   *
   * @param filename a name of a file or null for create a new document.
   *
   * @return The new document.
   */
  public iDocument create (string? filename = null) throws DocumentError
  {
    int pos;

    if (!this.is_open (filename, out pos))
    {
      AbstractTextDocument document;

      try
      {
        document = this.create_document (filename);

        if (filename != null)
        {
          try
          {
            document.open (filename);
            Symbols.Cache.get_instance ().add_file (filename);
          }
          catch (Error e)
          {
            warning (e.message);
          }
        }
        document.title_change.connect (this.title_change);
        this.append_page (document);
        document.show ();
        this.current = document;
        this.current.grab_focus ();
        if (this.current is AbstractTextDocument)
        {
          AbstractTextDocument atd;

          atd = this.current as AbstractTextDocument;
          atd.buffer.notify["can-undo"].connect (() => {
            this.setup_ui ();
          });
          atd.buffer.notify["can-redo"].connect (() => {
            this.setup_ui ();
          });
          atd.buffer.changed.connect (() => {
            this.setup_ui ();
          });
          atd.cursor_moved.connect ((s, r, c) => {
            this.tab_cursor_moved (s, r, c);
          });
        }
        this.current.closed.connect ((s) => {
          this.remove_page (s);
        });
        this.current.saved.connect ((s) => {
          Symbols.Cache.get_instance ().add_file (s.path);
          this.tab_saved (s);
        });
        this.current.state_change.connect ((sender, state) => {
          this.tab_state_changed (sender);
        });
      }
      catch (DocumentError e)
      {
        if (e is DocumentError.UNKNOW_DOCUMENT)
        {
          this.launch_external (filename);
        }
        else
        {
          warning (e.message);
        }
      }
    }
    else
    {
      this.set_current_page (pos);
    }

    Utils.process_gtk_events ();
    return this.current;
  }

  /**
   * Open document with #Gtk.FileChosserDialog.
   */
  public void open () throws DocumentError, ProjectError
  {
    Gtk.FileChooserDialog dialog;
    SList<string> files;

    files = new SList<string> ();
    dialog = new Gtk.FileChooserDialog (_("Open file"), null,
                                        Gtk.FileChooserAction.OPEN,
                                        Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                        Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
                                        null);

    dialog.set_select_multiple (true);
    if (dialog.run () == Gtk.ResponseType.ACCEPT)
    {
      files = dialog.get_filenames ().copy ();
    }
    dialog.destroy ();

    foreach (string file in files)
    {
      this.create (file);
      this.current.is_save = true;
    }
  }

  /**
   * Save all opening documents.
   *
   * @return true if all documents are really saved.
   */
  public bool save_all ()
  {
    bool saved = true;

    foreach (iDocument document in documents)
    {
      document.save ();
      if (!document.is_save)
      {
        saved = false;
        break;
      }
    }
    return saved;
  }

  /**
   * Close all opening documents.
   *
   * @return true if all documents are really closed.
   */
  public bool close_all ()
  {
    int unsave = 0;
    bool close = true;

    foreach (iDocument document in documents)
    {
      if (!document.is_save)
      {
        unsave++;
      }
    }

    if (unsave > 0)
    {
      DocumentCloseDialog dialog;
      Gtk.Window parent;
      int response;

      dialog = new DocumentCloseDialog (this.documents);
      parent = this.get_ancestor (typeof (Gtk.Window)) as Gtk.Window;
      dialog.set_transient_for (parent);
      dialog.set_modal (true);
      response = dialog.run ();
      dialog.destroy ();

      switch (response)
      {
        case Gtk.ResponseType.OK:
          foreach (iDocument document in dialog.selected)
          {
            document.save ();
          }
          close = true;
        break;
        case Gtk.ResponseType.CLOSE:
          close = true;
        break;
        case Gtk.ResponseType.CANCEL:
          close = false;
        break;
      }
    }

    if (close)
    {
      List<iDocument> documents;

      documents = this.documents.copy ();
      foreach (iDocument document in documents)
      {
        document.is_save = true;
        document.close ();
      }
    }
    return (this.get_n_pages () == 0);
  }

  /**
   * Append new page in notebook.
   *
   * @param child the document added.
   */
  public new int append_page (iDocument child)
  {
    int page_num;

    this.documents.append (child);
    page_num = base.append_page (child, null);
    this.set_tab_reorderable (child, true);
    this.title_change (child);
    this.setup_ui ();
    this.tab_added (child);
    return page_num;
  }

  /**
   * Remove page from notebook.
   *
   * @param child a document.
   */
  public new void remove_page (iDocument child)
  {
    int page_num;

    page_num = this.page_num (child);
    if (page_num >= 0)
    {
      this.tab_removed (child);
      base.remove_page (page_num);
      this.documents.remove (child);
      this.setup_ui ();
    }
  }

  /**
   * Returns the document contained in page number page_num.
   *
   * @param page_num the index of a page in the notebook, or -1 to get the last page. 
   *
   * @return The document, or NULL if page_num is out of bounds.
   */
  public iDocument get_nth (int page_num)
  {
    return this.get_nth_page (page_num) as iDocument;
  }

  /**
   * Check if file is already open.
   *
   * @param filename the path of file.
   * @param pos return tab position.
   *
   * @return true if file is already open.
   */
  public bool is_open (string? filename, out int pos)
  {
    bool opened = false;

    if (filename != null)
    {
      pos = 0;
      foreach (iDocument d in this.documents)
      {
        if (d.path == filename)
        {
          opened = true;
          break;
        }
        pos++;
      }
    }
    return opened;
  }
}

