/* idocument.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public enum Valide.DocumentState
{
  NORMAL,
  EXTERNALLY_MODIFIED,
  BAD_ENCODING,
  INFO
}

/**
 * An interface for document
 */
public interface Valide.iDocument : Gtk.Widget
{
  /**
   * The signal savec is emited when the document is saved
   */
  public signal void saved ();

  /**
   * The signal savec is emited when the document is closed
   */
  public signal void closed ();

  /**
   * The signal savec is emited when the title of the document changed
   */
  public signal void title_change ();

  /**
   * The signal savec is emited when the state of the document changed
   */
  public signal void state_change (DocumentState state);

  /**
   * The state of the document
   */
  public abstract DocumentState state
  {
    get;
    set;
  }

  /**
   * The document is uptodate with the hard disk copy?
   */
  public abstract bool is_save
  {
    get;
    set;
  }

  /**
   * The absolute path of the document
   */
  public abstract string path
  {
    owned get;
    set;
  }

  /**
   * The title of the document
   */
  public abstract string title
  {
    owned get;
  }

  /**
   * The base name of the file
   */
  public abstract string filename
  {
    owned get;
  }

  /**
   * Open a document
   *
   * @param path A path of a file
   */
  public abstract void open (string path) throws Error;

  /**
   * Save the content of the widget in hard file file.
   * If the file has not yet been saved, this function show a dialog for enter
   * the path of the file.
   *
   * @param current_folder The current folder of chooser
   */
  public abstract void save (string? current_folder = null);

  /**
   * Save as the content of the widget in hard disk file
   *
   * @param current_folder The current folder of chooser
   */
  public abstract void save_as (string? current_folder = null);

  /**
   * Rename the hard disk file
   */
  public abstract void rename ();

  /**
   * Close the document
   */
  public abstract bool close ();

  /**
   * Redo the last user action
   */
  public abstract void redo ();

  /**
   * Show the searching dialog
   */
  public abstract void search ();

  /**
   * Find the next word
   */
  public abstract void find_next ();

  /**
   * Find the previous word
   */
  public abstract void find_prev ();

  /**
   * Show the replace dialog
   */
  public abstract void replace ();

  /**
   * Goto the specified line
   *
   * @param line A line number or -1 to show the dialog
   *
   * @return true if the cursor moved
   *
   */
  public abstract bool goto_line (int line = -1);

  /**
   * Get the pixbuf of the file
   *
   * @return A Gdk.Pixbuf
   */
  public abstract Gdk.Pixbuf get_icon ();

  /**
   * Get the informations about this file
   *
   * @return An information text
   */
  public abstract string get_tooltips ();

  /**
   * Update the user interface
   *
   * @param ui_manager The UI manager
   */
  public abstract void setup_ui (UIManager ui_manager);
}

