/* project-dialog-options.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

private class Valide.HelpDialog : Gtk.Dialog
{
  public HelpDialog (Gtk.Window parent)
  {
    string message;
    Gtk.Label label;

    this.add_button (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE);
    this.set_transient_for (parent);
    this.set_modal (true);
    this.set_border_width (5);

    label = new Gtk.Label ("");
    this.vbox.add (label);
    label.show ();

    try
    {
      Process.spawn_command_line_sync ("valac --help", out message, null, null);
    }
    catch (Error e)
    {
      debug (e.message);
      message = e.message;
    }
    label.set_text (message);
  }

  public new int run ()
  {
    this.show_all ();
    return base.run ();
  }
}

/**
 * A dialog box widget for show/edit the option of a project
 */
public class Valide.ProjectDialogOptions : AbstractProjectDialogOptions
{
  private Gtk.ListStore list_store_src;
  private Gtk.ListStore list_store_dst;
  private Gtk.ListStore list_store_vapi;
  private HelpDialog help_dialog;

  /**
   * The project
   */
  public Project project
  {
    get;
    construct;
  }

  /* Frame general */

  /**
   * The path of the project
   */
  public string path
  {
    get { return this.widgets.path.get_text (); }
    set { this.widgets.path.set_text (value); }
  }

  /**
   * The name of the project
   */
  public string name
  {
    get { return this.widgets.name.get_text (); }
    set { this.widgets.name.set_text (value); }
  }

  /**
   * The author of the project
   */
  public string author
  {
    get { return this.widgets.author.get_text (); }
    set
    {
      if (value == null)
      {
        this.widgets.author.set_text ("");
      }
      else
      {
        this.widgets.author.set_text (value);
      }
    }
  }

  /**
   * The license of the project
   */
  public string license
  {
    get { return this.widgets.license.get_text (); }
    set { this.widgets.license.set_text (value); }
  }

  /**
   * The version of the project
   */
  public string version
  {
    get { return this.widgets.version.get_text (); }
    set
    {
      if (value == null)
      {
        this.widgets.version.set_text ("");
      }
      else
      {
        this.widgets.version.set_text (value);
      }
    }
  }

  /* Frame compiler */

  /**
   * The debug option
   */
  public bool debuging
  {
    get { return this.widgets.debug.get_active (); }
    set { this.widgets.debug.set_active (value); }
  }

  /**
   * Active thread
   */
  public bool thread
  {
    get { return this.widgets.thread.get_active (); }
    set { this.widgets.thread.set_active (value); }
  }

  /**
   * Disable assert option
   */
  public bool disable_assert
  {
    get { return this.widgets.disable_assert.get_active (); }
    set { this.widgets.disable_assert.set_active (value); }
  }

  /**
   * Disable checking option
   */
  public bool disable_checking
  {
    get { return this.widgets.disable_checking.get_active (); }
    set { this.widgets.disable_checking.set_active (value); }
  }

  /**
   * Disable non null option
   */
  public bool disable_non_null
  {
    get { return this.widgets.disable_non_null.get_active (); }
    set { this.widgets.disable_non_null.set_active (value); }
  }

  /**
   * Save the temporary files
   */
  public bool save_temps
  {
    get { return this.widgets.save_temps.get_active (); }
    set { this.widgets.save_temps.set_active (value); }
  }

  /**
   * Quiet option
   */
  public bool quiet
  {
    get { return this.widgets.quiet.get_active (); }
    set { this.widgets.quiet.set_active (value); }
  }

  /**
   * Other options
   */
  public string other
  {
    owned get { return this.widgets.other.get_text (); }
    set { this.widgets.other.set_text (value); }
  }

  /* Frame vapi dir */

  /**
   * The list of the vapi directories
   */
  public List<string> vapi_dir
  {
    owned get
    {
      List<string> it;
      Gtk.TreeIter iter;
      string str;

      if (this.list_store_vapi.get_iter_first (out iter))
      {
        do
        {
          this.list_store_vapi.get (iter, 0, out str);
          it.append (str);
        } while (this.list_store_vapi.iter_next (ref iter));
      }
      return it;
    }

    set
    {
      foreach (string it in value)
      {
        Gtk.TreeIter iter;

        this.list_store_vapi.append (out iter);
        this.list_store_vapi.set (iter, 0, it);
      }
    }
  }

  /* Frame package */

  /**
   * The list of the packages
   */
  public List<string> pkg
  {
    owned get
    {
      List<string> it;
      Gtk.TreeIter iter;
      string str;

      if (this.list_store_dst.get_iter_first (out iter))
      {
        do
        {
          this.list_store_dst.get (iter, 0, out str);
          it.append (str);
        } while (this.list_store_dst.iter_next (ref iter));
      }
      return it;
    }

    set
    {
      foreach (string it in value)
      {
        Gtk.TreeIter iter;

        this.list_store_dst.append (out iter);
        this.list_store_dst.set (iter, 0, it);
      }
    }
  }

  private bool pkg_exist (string pkg)
  {
    bool exist = false;

    foreach (string p in this.pkg)
    {
      if (p == pkg)
      {
        exist = true;
        break;
      }
    }
    return exist;
  }

  private void populate_tree_view (Gtk.TreeView tree_view)
  {
    Gtk.ListStore list_store;
    Gtk.CellRendererText render;
    Gtk.TreeViewColumn column;

    list_store = new Gtk.ListStore (1, typeof (string));
    list_store.set_sort_column_id (0, Gtk.SortType.ASCENDING);

    tree_view.set_model (list_store);
    render = new Gtk.CellRendererText ();
    column = new Gtk.TreeViewColumn.with_attributes ("", render, "text", 0, null);
    tree_view.append_column (column);
    tree_view.headers_visible = false;
  }

  private void update_vapi ()
  {
    List<string> pkg;

    pkg = Valac.get_package_list (this.vapi_dir);
    this.list_store_src.clear ();
    foreach (string p in pkg)
    {
      Gtk.TreeIter iter;

      this.list_store_src.append (out iter);
      this.list_store_src.set (iter, 0, p);
    }
  }

  /* @todo Check duplicated items */
  /**
   * Callback for add a vapi directory
   */
  [CCode (instance_pos = -1)]
  protected void add_vapi_dir (Gtk.Widget sender)
  {
    string dirname;

    dirname = this.widgets.file_chooser_button.get_filename ();
    if (dirname != "")
    {
      Gtk.TreeIter iter;

      this.list_store_vapi.append (out iter);
      this.list_store_vapi.set (iter, 0, dirname);
      this.update_vapi ();
    }
  }

  /**
   * Callback for delete a vapi directory
   */
  [CCode (instance_pos = -1)]
  protected void del_vapi_dir (Gtk.Widget sender)
  {
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.widgets.tree_view_vapi.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      this.list_store_vapi.remove (iter);
      this.update_vapi ();
    }
  }

  /**
   * Callback for add a package with double-clic
   */
  [CCode (instance_pos = -1)]
  protected void row_add_pkg (Gtk.TreeView sender,
                              Gtk.TreePath path,
                              Gtk.TreeViewColumn column)
  {
    this.add_pkg (sender);
  }

  /**
   * Callback for delete a package with double-clic
   */
  [CCode (instance_pos = -1)]
  protected void row_del_pkg (Gtk.TreeView sender,
                              Gtk.TreePath path,
                              Gtk.TreeViewColumn column)
  {
    this.del_pkg (sender);
  }

  /* @todo Check duplicated items */
  /**
   * Callback for add a package
   */
  [CCode (instance_pos = -1)]
  protected void add_pkg (Gtk.Widget sender)
  {
    string pkg;
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.widgets.tree_view_src.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      this.list_store_src.get (iter, 0, out pkg);
      if (!this.pkg_exist (pkg))
      {
        this.list_store_dst.append (out iter);
        this.list_store_dst.set (iter, 0, pkg);
      }
    }
  }

  /* @todo Check duplicated items */
  /**
   * Callback for delete a package
   */
  [CCode (instance_pos = -1)]
  protected void del_pkg (Gtk.Widget sender)
  {
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.widgets.tree_view_dst.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      this.list_store_dst.remove (iter);
    }
  }

  /**
   * Callback for show the valac help
   */
  [CCode (instance_pos = -1)]
  protected void show_help (Gtk.Button sender)
  {
    HelpDialog help_dialog;

    help_dialog = new HelpDialog (this.widgets.dialog);
    help_dialog.run ();
    help_dialog.destroy ();
  }

  /**
   * Create a new Valide.ProjectDialogOption
   *
   * @param project The attached project
   * @param parent The parent window
   */
  public ProjectDialogOptions (Project project, Gtk.Window parent)
  {
    this.project = project;
    this.widgets.dialog.set_transient_for (parent);
  }

  construct
  {
    string opt;
    Gdk.Pixbuf icon;
    Gtk.TreeIter iter;
    Gtk.IconView icon_view;
    Gtk.ListStore list_store;

    icon_view = this.widgets.icon_view;
    list_store = new Gtk.ListStore (3, typeof (Gdk.Pixbuf),
                                       typeof (string),
                                       typeof (int));
    icon_view.set_model (list_store);
    icon_view.set_pixbuf_column (0);
    icon_view.set_text_column (1);

    this.widgets.label4.set_markup ("<b>%s</b>".printf (this.widgets.label4.get_text ()));
    this.widgets.label5.set_markup ("<b>%s</b>".printf (this.widgets.label5.get_text ()));
    try
    {
      list_store.append (out iter);
      icon = new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "pref-general.png"));
      list_store.set (iter, 0, icon, 1, _("General"), 2, 0, -1);
    }
    catch (Error e)
    {
      debug (e.message);
    }

    try
    {
      list_store.append (out iter);
      icon = new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "pref-compiler.png"));
      list_store.set (iter, 0, icon, 1, _("Compiler"), 2, 1, -1);
    }
    catch (Error e)
    {
      debug (e.message);
    }

    try
    {
      list_store.append (out iter);
      icon = new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "pref-pkg.png"));
      list_store.set (iter, 0, icon, 1, _("Packages"), 2, 2, -1);
    }
    catch (Error e)
    {
      debug (e.message);
    }

    icon_view.selection_changed.connect ((s) => {
      weak List<Gtk.TreePath> tabs;

      tabs = s.get_selected_items ();
      if (tabs != null)
      {
        int index;
        Gtk.TreeIter iter2;

        s.model.get_iter (out iter2, tabs.data);
        s.model.get (iter2, 2, out index, -1);
        this.widgets.notebook.set_current_page (index);
      }
    });
    /* Set first item */
    icon_view.select_path (new Gtk.TreePath.from_string ("0"));

    this.populate_tree_view (this.widgets.tree_view_vapi);
    this.list_store_vapi = (Gtk.ListStore)this.widgets.tree_view_vapi.model;
    this.populate_tree_view (this.widgets.tree_view_src);
    this.list_store_src = (Gtk.ListStore)this.widgets.tree_view_src.model;
    this.populate_tree_view (this.widgets.tree_view_dst);
    this.list_store_dst = (Gtk.ListStore)this.widgets.tree_view_dst.model;

    this.connect_signals ("valide_project_dialog_options_");

    this.path = this.project.path;
    this.name = this.project.name;
    this.author = this.project.author;
    this.version = this.project.version;
    this.pkg = this.project.pkg;
    this.license = this.project.license;
    this.vapi_dir = this.project.vapi_dir;
    this.debuging = this.project.debuging;
    this.thread = this.project.thread;
    this.disable_assert = this.project.disable_assert;
    this.disable_checking = this.project.disable_checking;
    this.disable_non_null = this.project.disable_non_null;
    this.save_temps = this.project.save_temps;
    this.quiet = this.project.quiet;
    opt = this.project.other;
    if (opt != null)
    {
      this.other = opt;
    }
    this.update_vapi ();
  }

  /**
   * @see Gtk.Dialog.run
   */
  public int run ()
  {
    this.widgets.dialog.show_all ();
    return this.widgets.dialog.run ();
  }

  /**
   * @see Gtk.Object.destroy
   */
  public new void destroy ()
  {
    this.widgets.dialog.destroy ();
  }
}

