/* project-manager.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public errordomain Valide.ProjectError
{
  NO_PROJECT,
  MAX_OPEN
}

/**
 * The project manager
 */
public class Valide.ProjectManager : Gtk.ScrolledWindow
{
  private enum Col
  {
    ICON,
    NAME,
    PATH,
    REAL_PATH,
    NB
  }

  private Gtk.Menu popup;
  private Project project;
  private weak Project _current;
  private Gtk.TreeView tree_view;
  private List<Project> projects;
  private Gtk.TreeStore tree_store;
  private ProjectDialog project_dialog;

  /**
   * This signal was emitted when an project is opened
   */
  public signal void project_opened (Project project);

  /**
   * This signal was emitted when an project is closed
   */
  public signal void project_closed ();

  /**
   * The ui manager
   */
  public UIManager ui_manager
  {
    get;
    construct;
  }

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    construct;
  }

  private string copy_template (Template template,
                                string path, string name, string author,
                                string version, string license) throws Error
  {
    string filename = null;
    string contents = null;

    // Transform XML
    Xml.Doc* xml;
    Xml.XPathContext xpath;
    Xml.XPathObject* xpath_obj;

    xml = Xml.Parser.parse_file (template.filename);
    xpath = new Xml.XPathContext (xml);

    xpath_obj = xpath.eval ("/template");
    xpath_obj->nodesetval->item (0)->name = "project";
    xpath_obj = xpath.eval ("/project/description");
    xpath_obj->nodesetval->item (0)->unlink ();
    xpath_obj = xpath.eval ("/project/icon");
    xpath_obj->nodesetval->item (0)->unlink ();
    xpath_obj = xpath.eval ("/project/name");
    xpath_obj->nodesetval->item (0)->set_content (name);
    xpath_obj = xpath.eval ("/project");
    xpath_obj->nodesetval->item (0)->new_text_child (null, "author", author);
    xpath_obj->nodesetval->item (0)->new_text_child (null, "version", version);
    xpath_obj->nodesetval->item (0)->new_text_child (null, "license", license);

    if (template.lang == "genie")
    {
      xpath_obj = xpath.eval ("/project/files/file/text ()");
      for (int i = 0; i < xpath_obj->nodesetval->length (); i++)
      {
        string file;

        file = xpath_obj->nodesetval->item (i)->get_content ();
        file = file.replace (".vala", ".gs");
        xpath_obj->nodesetval->item (i)->set_content (file);
      }
    }

    // Copy license
    if (license != "None")
    {
      FileUtils.get_contents (Path.build_filename (Config.LICENSES_DIR, license),
                              out contents, null);
      FileUtils.set_contents (Path.build_filename (path, "COPYING"),
                              contents, -1);
      xpath_obj = xpath.eval ("/project/files");
      xpath_obj->nodesetval->item (0)->new_text_child (null, "file", "COPYING");
    }

    // Copy xml
    filename = Path.build_filename (path, name + "." + Project.FILE_EXT[0]);
    xml->save_file (filename);

    // Copy source files
    foreach (string file in template.files)
    {
      File src, dst;

      src = File.new_for_path (file);
      dst = File.new_for_path (Path.build_filename (path, Path.get_basename (file)));
      src.copy (dst, FileCopyFlags.OVERWRITE, null, null);
    }

    return filename;
  }

  /**
   * The current project
   */
  public Project current
  {
    get
    {
      return this._current;
    }
    set
    {
      this._current = value;
      this.refresh ();
    }
  }

  private void file_select (Gtk.TreeView sender, Gtk.TreePath path,
                            Gtk.TreeViewColumn column)
  {
    Gtk.TreeIter iter;
    string filename = null;

    this.tree_store.get_iter (out iter, path);
    this.tree_store.get (iter, Col.REAL_PATH, out filename);
    if (!this.tree_store.iter_has_child (iter))
    {
      try
      {
        this.current.open_file (filename);
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
    else
    {
      // @todo Active project
      // @todo Unexpand other projects

      /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=536480 */
      if (Fix.GtkTreeView.row_is_expanded (this.tree_view, path))
      {
        this.tree_view.collapse_row (path);
      }
      else
      {
        this.tree_view.expand_row (path, false);
      }
    }
  }

  private bool button_press (Gtk.Widget sender, Gdk.EventButton event)
  {
    if (event.type == Gdk.EventType.BUTTON_PRESS && event.button == 3)
    {
      this.popup.popup (null, null, null, event.button, event.time);
      return true;
    }
    return false;
  }

  private void setup_ui ()
  {
    bool active = false;

    if (this.current != null)
    {
      active = true;
    }
    this.ui_manager.get_action ("project-add").sensitive = active;
    this.ui_manager.get_action ("project-remove").sensitive = active;
    this.ui_manager.get_action ("project-options").sensitive = active;
    this.ui_manager.get_action ("project-compile").sensitive = active;
    this.ui_manager.get_action ("project-execute").sensitive = active;
    this.ui_manager.get_action ("project-close").sensitive = active;
    this.sensitive = active;
  }

  /* @TODO Save all expanded rows (childrens of root) */
  private void refresh ()
  {
    bool expanded;
    bool was_empty = true;
    Gtk.TreePath tree_path;
    Gtk.TreeIter iter;
    HashTable<string, Gtk.TreeIter*> dir_iters;

    tree_path = new Gtk.TreePath ();
    /* Save expanded rows */
    if (this.tree_store.get_iter_first (out iter))
    {
      tree_path = this.tree_store.get_path (iter);
      /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=536480 */
      expanded = Fix.GtkTreeView.row_is_expanded (this.tree_view, tree_path);
      was_empty = false;
    }
    else
    {
      expanded = false;
      was_empty = true;
    }

    this.tree_store.clear ();
    foreach (Project p in projects)
    {
      string title = null;
      Gtk.TreeIter root;

      this.tree_store.append (out root, null);

      if (p == this.current)
      {
        this.tree_store.set (root, Col.NAME, "<b>" + p.name + "</b>", -1);
      }
      else
      {
        this.tree_store.set (root, Col.NAME, p.name, -1);
      }

      dir_iters = new HashTable<string, Gtk.TreeIter*> (str_hash, str_equal);
      foreach (string file in p.files)
      {
        string dir;
        string path;
        string home;
        string dirname;
        string filename;
        Gdk.Pixbuf pixbuf;
        Gtk.TreeIter* dir_iter;

        dir = Path.get_dirname (file);
        home = Environment.get_home_dir ();
        filename = Path.get_basename (file);
        if (dir.has_prefix (p.path) && dir != p.path)
        {
          dirname = dir.substring (p.path.length + 1);
        }
        else if (dir.has_prefix (home))
        {
          dirname = "~" + dir.substring (home.size (),
                                     dir.size () - home.size ());
        }
        else
        {
          dirname = dir;
        }

        if (dir == "." || dir == p.path)
        {
          dir_iter = &root;
        }
        else
        {
          dir_iter = dir_iters.lookup (dir);

          if (dir_iter == null)
          {
            Gtk.TreeIter tmp;

            this.tree_store.append (out tmp, root);
            dir_iter = (Gtk.TreeIter*)tmp.copy ();
            try
            {
              pixbuf = Utils.get_pixbuf_for_file (dir, Gtk.IconSize.MENU);
            }
            catch (Error e)
            {
              pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_CANCEL, Gtk.IconSize.MENU);
            }
            this.tree_store.set (*dir_iter, Col.ICON, pixbuf,
                                 Col.NAME, dirname, Col.PATH, dir, Col.REAL_PATH, dir);
            dir_iters.insert (dir, dir_iter);
          }
        }
        path = p.get_real_filename (file);
        this.tree_store.append (out iter, *dir_iter);
        try
        {
          pixbuf = Utils.get_pixbuf_for_file (path, Gtk.IconSize.MENU);
        }
        catch (Error e)
        {
          pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_CANCEL, Gtk.IconSize.MENU);
        }
        this.tree_store.set (iter, Col.NAME, filename, Col.ICON, pixbuf,
                             Col.PATH, file, Col.REAL_PATH, path);
      }
    }
    this.setup_ui ();

    /* Restore expanded rows */
    if (this.sensitive && expanded)
    {
      this.tree_view.expand_row (tree_path, true);
    }
    if (was_empty) 
    {
      if(this.tree_store.get_iter_first (out iter))
      {
        tree_path = this.tree_store.get_path (iter);
        this.tree_view.expand_row (tree_path, true);
      }
    }
  }

  private bool drag_motion_cb (Gtk.Widget sender,
                               Gdk.DragContext drag_context,
                               int x, int y,
                               uint time)
  {
    Gtk.TreePath path;
    Gtk.TreeSelection selection;

    selection = this.tree_view.get_selection ();
    /* Retrive the drag location */
    this.tree_view.get_path_at_pos (x, y, out path, null, null, null);
    if (path != null)
    {
      selection.select_path (path);
    }
    else
    {
      selection.unselect_all ();
    }
    return true;
  }

  private void drag_data_received_cb (Gtk.Widget sender,
                                      Gdk.DragContext drag_context,
                                      int x, int y,
                                      Gtk.SelectionData data,
                                      uint info, uint time)
  {
    bool del = false;
    Project project;
    SList<string> files;
    Gtk.TreePath path;
    Gtk.TreeIter iter;
    string filename;
    string name;
    string project_name;
    string destination;

    /* Retrive the drag location */
    this.tree_view.get_path_at_pos (x, y, out path, null, null, null);
    if (path == null)
    {
      project = this.current;
      destination = project.path;
    }
    else
    {
      this.tree_store.get_iter (out iter, path);
      this.tree_store.get (iter, Col.NAME, out project_name);
      this.tree_store.get (iter, Col.REAL_PATH, out filename);
      if (filename == null)
      {
        project = this.get_project_by_name (project_name);
        destination = project.path;
      }
      else
      {
        Gtk.TreeIter parent;

        if (!this.tree_store.iter_has_child (iter))
        {
          this.tree_store.iter_parent (out parent, iter);
          iter = parent;
        }
        this.tree_store.get (iter, Col.NAME, out name);

        this.tree_store.iter_parent (out parent, iter);
        this.tree_store.get (iter, Col.NAME, out project_name);
        project = this.get_project_by_name (name);

        destination = Path.build_filename (project.path, name);
      }
    }

    files = new SList<string> ();
    foreach (string uri in ((string)data.data).split ("\r\n"))
    {
      if (uri != "")
      {
        File src;
        string filepath;
        string basename;

        src = File.new_for_uri (uri);
        filepath = src.get_path ();
        basename = Path.get_basename (filepath);

        if (drag_context.suggested_action == Gdk.DragAction.LINK)
        {
          files.append (filepath);
        }
        else
        {
          try
          {
            File dest;
            string destpath;

            destpath = Path.build_filename (destination, src.get_basename ());
            dest = File.new_for_path (destpath);
            /* @TODO Ask for overwriting */
            src.copy (dest, FileCopyFlags.OVERWRITE, null, null);
            files.append (dest.get_path ());
            if (drag_context.suggested_action == Gdk.DragAction.MOVE)
            {
              src.delete (null);
              del = true;
            }
          }
          catch (Error e)
          {
            warning (e.message);
          }
          drag_context.suggested_action = drag_context.action;
        }
      }
    }
    if (files != null)
    {
      project.add_file (files);
    }
    Gtk.drag_finish (drag_context, true, del, time);
  }

  private int sort_func (Gtk.TreeModel model, Gtk.TreeIter a, Gtk.TreeIter b)
  {
    string path_a, path_b;

    model.get (a, Col.REAL_PATH, out path_a);
    model.get (b, Col.REAL_PATH, out path_b);

    return Utils.cmp_filename (path_a, path_b);
  }

  private Project get_project_by_name (string name)
  {
    /* @TODO */
    return this.current;
  }

  /**
   * Crea te a new project manager
   *
   * @param ui_manager The ui manager
   * @param documents The document manager
   */
  public ProjectManager (UIManager ui_manager, DocumentManager documents)
  {
    this.ui_manager = ui_manager;
    this.documents = documents;
    this.width_request = 200;
  }

  construct
  {
    Gtk.TreeViewColumn col;
    Gtk.CellRenderer renderer;
    Gtk.TargetEntry[] targets;

    this.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    /* create some nice padding */
    this.set_shadow_type (Gtk.ShadowType.IN);
    this.set_border_width (2);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.headers_visible = false;
    this.add (this.tree_view);

    this.tree_store = new Gtk.TreeStore (Col.NB, typeof (Gdk.Pixbuf),
                                         typeof (string), typeof(string),
                                         typeof (string));
    /* @todo Use sorting function for show directories before */
    this.tree_store.set_sort_column_id (Col.NAME, Gtk.SortType.ASCENDING);
    this.tree_store.set_sort_func (Col.NAME, this.sort_func);
    this.tree_view.set_model (this.tree_store);

    col = new Gtk.TreeViewColumn ();
    col.set_title (_("Projects"));

    renderer = new Gtk.CellRendererPixbuf ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "pixbuf", Col.ICON);

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "markup", Col.NAME);

    this.tree_view.append_column (col);

    this.tree_view.row_activated.connect (this.file_select);
    /* Add popup */
    this.popup = this.ui_manager.get_widget ("/project-popup") as Gtk.Menu;
    this.tree_view.button_press_event.connect (this.button_press);

    this.setup_ui ();

    /* Seting up the drag & drop */
    targets = new Gtk.TargetEntry[1];
    targets[0].target = "text/uri-list";
    targets[0].flags = 0;
    targets[0].info = 0;

    Gtk.drag_dest_set (this.tree_view, Gtk.DestDefaults.ALL, targets,
                       Gdk.DragAction.MOVE | Gdk.DragAction.COPY | Gdk.DragAction.LINK);
    this.tree_view.drag_motion.connect (drag_motion_cb);
    this.tree_view.drag_data_received.connect (drag_data_received_cb);
  }

  /**
   * Create a new project
   *
   * @return true if the project is created
   */
  public bool create ()
  {
    bool created = false;
    ProjectDialog dialog;

    dialog = new ProjectDialog ();
    if (dialog.run () == Gtk.ResponseType.APPLY)
    {
      string path;

      path = dialog.project_dir;
      if (!FileUtils.test (path, FileTest.EXISTS))
      {
        Gtk.Dialog msg;

        msg = new Gtk.Dialog.with_buttons (_("New project"), null,
                                           Gtk.DialogFlags.MODAL,
                                           Gtk.STOCK_NO, Gtk.ResponseType.NO,
                                           Gtk.STOCK_YES, Gtk.ResponseType.YES,
                                           null);

        msg.vbox.pack_start (new Gtk.Label (_("Project directory doesn't exist. Would you like to create it?")),
                             true, true, 10);
        msg.show_all ();
        if (msg.run () == Gtk.ResponseType.YES)
        {
          DirUtils.create_with_parents (path, 0755);
          created = true;
        }
        msg.destroy ();
      }
      else
      {
        created = true;
      }

      if (created == true)
      {
        path = dialog.project_path;
        if (FileUtils.test (path, FileTest.EXISTS))
        {
          Gtk.Dialog msg;

          msg = new Gtk.Dialog.with_buttons (_("New project"), null,
                                             Gtk.DialogFlags.MODAL,
                                             Gtk.STOCK_NO, Gtk.ResponseType.NO,
                                             Gtk.STOCK_YES, Gtk.ResponseType.YES,
                                             null);

          msg.vbox.pack_start (new Gtk.Label (_("Directory already exists. Would you like to overwrite it?")),
                               true, true, 10);
          msg.show_all ();
          if (msg.run () == Gtk.ResponseType.NO)
          {
            created = false;
          }
          msg.destroy ();
        }
        else
        {
          DirUtils.create (path, 0755);
        }
      }

      if (created == true)
      {
        try
        {
          string prj;

          prj = this.copy_template (new Template (dialog.template, dialog.lang),
                                    path, dialog.project_name,
                                    dialog.project_author,
                                    dialog.project_version,
                                    dialog.project_license);
          this.open (prj);
        }
        catch (Error e)
        {
          warning (e.message);
          created = false;
        }
      }
    }
    dialog.destroy ();
    return created;
  }

  /**
   * Open a project
   *
   * @param filename The name of project file
   *
   * @return ture if the project is opened
   *
   * @throws ProjectError
   */
  public bool open (string? filename = null) throws ProjectError
  {
    bool opened = false;
    Gtk.FileFilter filter;
    string file_name = filename;

    if (current == null)
    {
      if (file_name == null)
      {
        Gtk.FileChooserDialog dialog;

        dialog = new Gtk.FileChooserDialog (_("Open project"), null,
                                            Gtk.FileChooserAction.OPEN,
                                            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                            Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT,
                                            null);
        dialog.set_current_folder (ConfigManager.get_instance ().get_string ("Projects", "default-directory"));

        /* Add filter */
        filter = new Gtk.FileFilter ();
        filter.set_name (_("Val(a)IDE project file"));
        foreach (string ext in Project.FILE_EXT)
        {
          filter.add_pattern ("*." + ext);
        }
        dialog.add_filter (filter);

        if (dialog.run () == Gtk.ResponseType.ACCEPT)
        {
          try
          {
            file_name =  dialog.get_filename ();
          }
          catch (Error e)
          {
            warning (e.message);
          }
        }
        dialog.destroy ();
      }
      if (file_name != null)
      {
        if (FileUtils.test (file_name, FileTest.EXISTS))
        {
          Project p;

          p = new Project (this.documents, file_name);
          p.closed.connect (this.close);
          p.added_file.connect (this.refresh);
          p.removed_file.connect (this.refresh);
          p.options_changed.connect (this.refresh);
          this.projects.append (p);
          this.current = p;
          this.refresh ();

          this.project_opened (p);
          opened = true;
        }
        else
        {
          warning (_("The project file doesn't exist!"));
        }
      }
    }
    else
    {
      throw new ProjectError.MAX_OPEN (_("Maximum projects open!"));
    }
    return opened;
  }

  /**
   * Close the project
   *
   * @param project The project
   */
  public void close (Project project)
  {
    this.projects.remove (project);
    this.current = null;
    this.project_closed ();
  }

  /**
   * Open the selected file
   */
  public void open_selected_file ()
  {
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.tree_view.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      Gtk.TreePath path;
      Gtk.TreeViewColumn column;

      path = this.tree_store.get_path (iter);
      column = this.tree_view.get_column (0);
      this.file_select (this.tree_view, path, column);
    }
  }

  /**
   * Remove the selected file
   */
  public void remove_selected_file ()
  {

    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.tree_view.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      Gtk.TreePath path;

      path = this.tree_store.get_path (iter);
      if (path.get_depth () > 1)
      {
        string filename;
        List<string> files;

        files = new List<string> ();
        this.tree_store.get (iter, Col.PATH, out filename);
        files.append (filename);
        this.current.remove_file (files);
      }
    }
  }

  /**
   * Test if a filename is a project file.
   *
   * @param filename a filename.
   *
   * @return true if a filename is a project file.
   */
  public bool accept_file (string filename)
  {
    bool accept = false;
    string ext = Utils.get_extension (filename);

    foreach (string e in Project.FILE_EXT)
    {
      if (e == ext)
      {
        accept = true;
        break;
      }
    }
    return accept;
  }
}

