/* project.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * A class for project
 */
public class Valide.Project : Object
{
  /**
   * The file extension of projects
   */
  public const string[] FILE_EXT = {"vide"};

  private string _filename;
  private string working_dir;

  /**
   * The signal closed is emited when the project is closed
   */
  public signal void closed ();

  /**
   * The signal closed is emited when a file is removed from the project
   */
  public signal void removed_file ();

  /**
   * The signal closed is emited when a file is added from the project
   */
  public signal void added_file ();

  /**
   * The signal closed is emited when the project options changed
   */
  public signal void options_changed ();

  private bool exist (string filename)
  {
    bool ret = false;

    foreach (string file in this.files)
    {
      if (file == filename)
      {
        ret = true;
        break;
      }
    }
    return ret;
  }

  /**
   * The project file name
   */
  public string filename
  {
    owned get
    {
      return this._filename;
    }
    set construct
    {
      if (FileUtils.test (value, FileTest.IS_SYMLINK))
      {
        try
        {
          File file;
          FileInfo info;

          file = File.new_for_path (value);
          info = file.query_info ("*", FileQueryInfoFlags.NONE, null);
          this._filename = info.get_symlink_target ();
        }
        catch (Error e)
        {
          debug (e.message);
        }
      }
      else
      {
        this._filename = value;
      }
      this.working_dir = this.path;
    }
  }

  /* General options */

  /**
   * The path of the project
   */
  public string path
  {
    owned get
    {
      return Path.get_dirname (this.filename);
    }
  }

  /**
   * The name of the project
   */
  public string name
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/name/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/name");
      Markup.add_child (this.filename, "/project", "name", value);
    }
  }

  /**
   * The author of the project
   */
  public string author
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/author/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/author");
      Markup.add_child (this.filename, "/project", "author", value);
    }
  }

  /**
   * The version of the project
   */
  public string version
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/version/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/version");
      Markup.add_child (this.filename, "/project", "version", value);
    }
  }

  /**
   * The license of the project
   */
  public string license
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/license/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/license");
      Markup.add_child (this.filename, "/project", "license", value);
    }
  }

  /**
   * The file list of the project
   */
  public List<string> files
  {
    owned get
    {
      List<string> files_;
      Xml.NodeSet* nodes;

      files_ = new List<string> ();
      nodes = Markup.get_nodes (this.filename, "/project/files/file/text ()");
      for (int i = 0; i < nodes->length (); i++)
      {
        files_.append (nodes->item (i)->content);
      }
      return files_;
    }
  }

  /* Compiler options */

  /**
   * Produce debug information
   */
  public bool debuging
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/debug");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/debug"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "debug");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/debug");
      }
    }
  }

  /**
   * Enable multithreading support
   */
  public bool thread
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/thread");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/thread"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "thread");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/thread");
      }
    }
  }

  /**
   * Disable assertions
   */
  public bool disable_assert
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/disable_assert");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/disable_assert"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "disable_assert");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/disable_assert");
      }
    }
  }

  /**
   * Disable additional run-time checks
   */
  public bool disable_checking
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/disable_checking");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/disable_checking"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "disable_checking");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/disable_checking");
      }
    }
  }

  /**
   * disable experimental enhancements for non-null types
   */
  public bool disable_non_null
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/disable_non_null");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/disable_non_null"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "disable_non_null");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/disable_non_null");
      }
    }
  }

  /**
   * Keep temporary files
   */
  public bool save_temps
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/save_temps");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/save_temps"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "save_temps");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/save_temps");
      }
    }
  }

  /**
   * Do not print messages to the console
   */
  public bool quiet
  {
    get
    {
      return Markup.exist (this.filename, "/project/compiler/options/quiet");
    }
    set
    {
      if (value)
      {
        if (!Markup.exist (this.filename, "/project/compiler/options/quiet"))
        {
          Markup.add_child (this.filename, "/project/compiler/options", "quiet");
        }
      }
      else
      {
        Markup.remove (this.filename, "/project/compiler/options/quiet");
      }
    }
  }

  /**
   * Other options
   */
  public string other
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/compiler/options/other/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/compiler/options/other");
      Markup.add_child (this.filename, "/project/compiler/options", "other", value);
    }
  }

  /* Packages options */

  /**
   * The list of the packages
   */
  public List<string> pkg
  {
    owned get
    {
      List<string> pkg_;
      Xml.NodeSet* nodes;

      pkg_ = new List<string> ();
      nodes = Markup.get_nodes (this.filename, "/project/compiler/pkg/text ()");
      for (int i = 0; i < nodes->length (); i++)
      {
        pkg_.append (nodes->item (i)->content);
      }
      return pkg_;
    }
    set
    {
      Markup.remove (this.filename, "/project/compiler/pkg");

      foreach (string p in value)
      {
        Markup.add_child (this.filename, "/project/compiler", "pkg", p);
      }
    }
  }

  /**
   * The list of the vapi directory
   */
  public List<string> vapi_dir
  {
    owned get
    {
      List<string> vapi_;
      Xml.NodeSet* nodes;

      vapi_ = new List<string> ();
      nodes = Markup.get_nodes (this.filename, "/project/compiler/vapi/text ()");
      for (int i = 0; i < nodes->length (); i++)
      {
        vapi_.append (nodes->item (i)->content);
      }
      return vapi_;
    }
    set
    {
      Markup.remove (this.filename, "/project/compiler/vapi");

      foreach (string p in value)
      {
        Markup.add_child (this.filename, "/project/compiler", "vapi", p);
      }
    }
  }

  /**
   * The building system
   */
  public string builder
  {
    owned get
    {
      return Markup.get_content (this.filename, "/project/compiler/builder/text ()");
    }
    set
    {
      Markup.remove (this.filename, "/project/compiler/builder");
      Markup.add_child (this.filename, "/project/compiler", "builder", value);
    }
  }

  /**
   * The compiler options
   */
  public string compiler_options
  {
    owned get
    {
      string opt = "";

      opt += "-o \"%s\"".printf (this.executable_name);

      foreach (string p in this.pkg)
      {
        opt += " --pkg=" + p;
      }

      foreach (string p in this.vapi_dir)
      {
        opt += " --vapidir=\"" + p + "\"";
      }

      if (this.debuging)
      {
        opt += " -g";
      }

      if (this.thread)
      {
        opt += " --thread";
      }

      if (this.disable_assert)
      {
        opt += " --disable-assert";
      }

      if (this.disable_checking)
      {
        opt += " --disable-checking";
      }

      if (this.disable_non_null)
      {
        opt += " --disable-non-null";
      }

      if (this.save_temps)
      {
        opt += " --save-temps";
      }

      if (this.quiet)
      {
        opt += " -q";
      }

      opt += " " + this.other;

      foreach (string file in this.files)
      {
         if (file.has_suffix (".vala") || file.has_suffix (".gs"))
         {
           opt += " \"%s\"".printf (file);
         }
      }

      return opt;
    }
  }

  /**
   * The name of the executable
   */
  public string executable_name
  {
    owned get
    {
      string exe_name;

      exe_name = Path.build_filename (this.path, this.name);
      if (Config.OS == "win32")
      {
        exe_name += ".exe";
      }
      return exe_name;
    }
  }

  /**
   * The executable options
   */
  public string executable_options
  {
    owned get
    {
      string opt =  "";

      return opt;
    }
  }

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    construct;
  }

  private void document_changed (DocumentManager sender, iDocument document)
  {
    string xpath;
    string filename;
    string old_filename;

    old_filename = (string)document.get_data ("old_path");
    foreach (string file in this.files)
    {
      if (!Path.is_absolute (file))
      {
        filename = Path.build_filename (this.path, file);
      }
      else
      {
        filename = file;
      }
      if (filename == old_filename)
      {
        xpath = "/project/files/file[.='%s']".printf (file);
        Markup.remove (this.filename, xpath);
        Markup.add_child (this.filename, "/project/files", "file",
                          Utils.get_relative_path (document.path, this.path));
        this.added_file ();
        break;
      }
    }
  }

  /**
   * Create a new project
   *
   * @param documents The document manager
   * @param filename The project file name
   */
  public Project (DocumentManager documents, string filename)
  {
    this.documents = documents;
    this.filename = filename;
  }

  construct
  {
    this.documents.tab_state_changed.connect (this.document_changed);
  }

  /**
   * Close the project
   */
  public void close ()
  {
    foreach (string filename in this.files)
    {
      int pos;
      string path;

      path = this.get_real_filename (filename);

      if (this.documents.is_open (path, out pos))
      {
        iDocument document;

        document = this.documents.get_nth_page (pos) as iDocument;
        this.documents.remove_page (document);
      }
    }
    this.closed ();
  }

  /**
   * Create a new file
   */
  public void create_file () throws DocumentError
  {
    iDocument doc;
    Gtk.Dialog dialog;
    bool adding_file = false;

    dialog = new Gtk.Dialog ();
    dialog.title = _("Create file");
    dialog.vbox.pack_start (new Gtk.Label (_("Would you like to add the file to this project?")), true, true, 0);
    dialog.add_button (Gtk.STOCK_YES, Gtk.ResponseType.YES);
    dialog.add_button (Gtk.STOCK_NO, Gtk.ResponseType.NO);
    dialog.show_all ();
    if (dialog.run () == Gtk.ResponseType.YES)
    {
      adding_file = true;
    }
    dialog.destroy ();

    doc = this.documents.create ();
    if (adding_file)
    {
      if (doc is AbstractTextDocument)
      {
        (doc as AbstractTextDocument).buffer.set_modified (true);
      }
      doc.save (this.path);
      if (doc.is_save)
      {
        SList<string> files;

        files = new SList<string> ();
        files.append (doc.path);
        this.add_file (files);
      }
    }
  }

  /**
   * Open a file
   *
   * @param filename The name of the file
   *
   * @return The new document
   */
  public iDocument open_file (string filename) throws DocumentError
  {
    string path;
    iDocument document;

    path = this.get_real_filename (filename);
    if (FileUtils.test (path, FileTest.EXISTS))
    {
      document = this.documents.create (path);
    }
    else
    {
      throw new DocumentError.BAD_URI (_("The file %s doesn't exist!"), filename);
    }
    return document;
  }

  /**
   * Add a list of file in project
   *
   * @param files The list of files
   */
  public void add_file (SList<string>? files = null)
  {
    SList<string> f;

    if (files == null)
    {
      Gtk.FileChooserDialog dialog;

      dialog = new Gtk.FileChooserDialog (_("Add file"), null,
                                          Gtk.FileChooserAction.OPEN,
                                          Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                          Gtk.STOCK_ADD, Gtk.ResponseType.ACCEPT);

      dialog.set_current_folder (this.path);
      dialog.set_local_only (true);
      dialog.select_multiple = true;

      if (dialog.run () == Gtk.ResponseType.ACCEPT)
      {
        f = dialog.get_filenames ().copy ();
        files = f;
      }
      dialog.destroy ();
    }
    if (files != null)
    {
      foreach (string file in files)
      {
        string filename;

        try
        {
          if (!FileUtils.test (file, FileTest.EXISTS))
          {
            FileUtils.set_contents (file, "", -1);
          }

          filename = Utils.get_relative_path (file, this.path);
          if (!this.exist (filename))
          {
            if (!Markup.exist (this.filename, "/project/files"))
            {
              Markup.add_child (this.filename, "/project", "files");
            }
            Markup.add_child (this.filename, "/project/files", "file", filename);
            this.added_file ();
          }
          else
          {
            message (_("File '%s' already added!"), filename);
          }
        }
        catch (Error e)
        {
          warning (_("Couldn't create the file: %s"), file);
        }
      }
    }
  }

  /**
   * Remove a list of file in project
   *
   * @param files The list of files
   */
  public void remove_file (List<string>? files = null)
  {
    List<string> f;

    if (files == null)
    {
      ProjectDialogRemove dialog;

      dialog = new ProjectDialogRemove (this);
      if (dialog.run () == Gtk.ResponseType.APPLY)
      {
        f = dialog.selected_files;
        files = f;
      }
      dialog.destroy ();
    }
    if (files != null)
    {
      string msg;
      bool remove;
      int response;
      Gtk.MessageDialog dialog;

      if (files.length () < 2)
      {
        msg = _("This will remove the file from the project. Also delete the file?");
      }
      else
      {
        msg = _("This will remove the files from the project. Also delete the files?");
      }
      dialog = new Gtk.MessageDialog (null,
                                      Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.QUESTION,
                                      Gtk.ButtonsType.YES_NO,
                                      msg);
      dialog.add_button (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL);
      dialog.show_all ();
      response = dialog.run ();
      dialog.destroy ();
      if (response == Gtk.ResponseType.CANCEL)
      {
        return;
      }
      remove = (response == Gtk.ResponseType.YES);
      foreach (string file in files)
      {
        Markup.remove (this.filename, "/project/files/file[.='" + file + "']");
        if (remove)
        {
          int pos;
          string filename;

          filename = this.get_real_filename (file);
          if (this.documents.is_open (filename, out pos))
          {
            this.documents.get_nth (pos).close ();
          }
          if (FileUtils.unlink (filename) < 0)
          {
            warning (_("Couldn't remove the file '%s' from disk"), filename);
          }
        }
      }
      this.removed_file ();
    }
  }

  /**
   * The project is compiled?
   */
  public bool is_compiled ()
  {
    return FileUtils.test (this.executable_name, FileTest.EXISTS);
  }

  /**
   * The executable is up to date?
   */
  public bool is_uptodate ()
  {
    string filename;
    uint64 exec_mtime;
    uint64 file_mtime;
    bool uptodate = true;

    exec_mtime = Utils.get_mtime (this.executable_name);
    foreach (string file in this.files)
    {
      filename = this.get_real_filename (file);
      file_mtime = Utils.get_mtime (filename);
      if (file_mtime > exec_mtime)
      {
        uptodate = false;
        break;
      }
    }
    return uptodate;
  }

  /**
   * Show the project option dialog
   */
  public void options (Gtk.Window parent)
  {
    ProjectDialogOptions dialog;

    dialog = new ProjectDialogOptions (this, parent);
    if (dialog.run () == Gtk.ResponseType.APPLY)
    {
      this.name = dialog.name;
      this.author = dialog.author;
      this.version = dialog.version;
      this.debuging = dialog.debuging;
      this.thread = dialog.thread;
      this.disable_assert = dialog.disable_assert;
      this.disable_checking = dialog.disable_checking;
      this.save_temps = dialog.save_temps;
      this.quiet = dialog.quiet;
      this.other = dialog.other;
      this.pkg = dialog.pkg;
      this.vapi_dir = dialog.vapi_dir;
      this.options_changed ();
    }
    dialog.destroy ();
  }

  /**
   * Compile the project
   *
   * @param executables The executable manager
   */
  public void compile (ExecutableManager executables)
  {
    int pos;
    bool saved = true;

    foreach (string file in this.files)
    {
      string path;

      path = this.get_real_filename (file);
      if (this.documents.is_open (path, out pos))
      {
        iDocument document;

        document = this.documents.get_nth (pos);
        document.save (this.path);
        if (!document.is_save)
        {
          saved = false;
          break;
        }
      }
    }

    if (saved)
    {
      try
      {
        Compiler compiler;

        compiler = new Compiler (this);
        executables.run (compiler, this.path, this.compiler_options);
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
  }

  /**
   * Execute the project
   *
   * @param executables The executable manager
   */
  public void execute (ExecutableManager executables)
  {
    if (this.is_compiled ())
    {
      bool run = true;

      if (!this.is_uptodate ())
      {
        Gtk.MessageDialog dialog;

        dialog = new Gtk.MessageDialog (null, Gtk.DialogFlags.MODAL,
                                        Gtk.MessageType.QUESTION,
                                        Gtk.ButtonsType.YES_NO,
                                        _("The executable is too old, would you like really execute it?"));
        if (dialog.run () == Gtk.ResponseType.NO)
        {
          run = false;
        }
        dialog.destroy ();
      }
      if (run)
      {
        Executable executable;

        executable = new Executable ();
        executable.executable = "'%s'".printf (this.executable_name);
        executables.run (executable, this.working_dir, this.executable_options);
      }
    }
    else
    {
      warning (_("You should compile the project before execute it!"));
    }
  }

  /**
   * Test if a filename is a project file.
   *
   * @param filename a filename.
   *
   * @return true if a filename is a project file.
   */
  public static bool accept_file (string filename)
  {
    bool accept = false;
    string ext;

    ext = Utils.get_extension (filename);
    foreach (string e in Project.FILE_EXT)
    {
      if (e == ext)
      {
        accept = true;
        break;
      }
    }
    return accept;
  }

  /**
   * Get the real file name
   *
   * @param filename A file include in this project
   *
   * @return The real name of the file
   */
  public string get_real_filename (string filename)
  {
    string path;

    path = Path.build_filename (this.path, filename);
    if (!FileUtils.test (path, FileTest.EXISTS))
    {
      string xpath;
      string basename;

      xpath = "/project/files/file[contains(string(),'" + Path.DIR_SEPARATOR_S + filename + "')]/text()";
      basename = Markup.get_content (this.filename, xpath);
      if (basename != null)
      {
        path = Path.build_filename (this.path, basename);
      }
      else
      {
        path = filename;
      }
    }
    return path;
  }
}


