/* start-page.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/**
 * Private class
 */
private class Valide.LinkLabel : Gtk.EventBox
{
  private signal void open_uri (string? uri);

  public string uri
  {
    get;
    construct;
  }

  construct
  {
    this.set_visible_window (false);
  }

  public LinkLabel (string? uri, string title, Gdk.Pixbuf? pixbuf = null)
  {
    string text;
    Gtk.HBox hbox;
    Gtk.Label label;

    hbox = new Gtk.HBox (false, 0);
    this.add (hbox);

    if (pixbuf != null)
    {
      Gtk.Image image;

      image = new Gtk.Image.from_pixbuf (pixbuf);
      hbox.pack_start (image, false, false, 2);
      text = "";
    }
    else
    {
      text = "• ";
    }

    label = new Gtk.Label ("");
    label.set_alignment (0, 0.5f);
    text += "<span color=\"blue\"><u>%s</u></span>".printf (title);
    label.set_markup (text);
    hbox.pack_start (label, false, true, 0);
    this.uri = uri;
    this.button_press_event.connect (() => {
      this.open_uri (this.uri);
    });
    this.enter_notify_event.connect (() => {
      Gdk.Cursor cursor;

      cursor = new Gdk.Cursor (Gdk.CursorType.HAND2);
      this.window.set_cursor (cursor);
      return false;
    });
    this.leave_notify_event.connect (() => {
      Gdk.Cursor cursor;

      cursor = new Gdk.Cursor (Gdk.CursorType.ARROW);
      this.window.set_cursor (cursor);
      return false;
    });
  }
}

private struct Valide.Item
{
  public Gdk.Pixbuf icon;
  public string uri;
  public string text;
  public Callback callback;
}

/**
 * The start page widget
 */
public class Valide.StartPage : Gtk.Table, iDocument
{
  private Gtk.VBox recent_box;
  private Gtk.Frame file_frame;
  private Gtk.Frame project_frame;

  public signal void send_message (string message);

  private delegate void OpenFunc (Gtk.LinkButton sender);

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    construct;
  }

  /**
   * The project manager
   */
  public ProjectManager projects
  {
    get;
    construct;
  }

  /**
   * The recents files manager
   */
  public RecentManagerUI recents
  {
    get;
    construct;
  }

  /**
   * @see Valide.iDocument.state
   */
  public new DocumentState state
  {
    get
    {
      return DocumentState.INFO;
    }
    set
    {
    }
  }

  /**
   * @see Valide.iDocument.is_save
   */
  public bool is_save
  {
    get
    {
      return true;
    }
    set
    {
    }
  }

  /**
   * @see Valide.iDocument.path
   */
  public new string path
  {
    owned get
    {
      return "";
    }
    set
    {
    }
  }

  /**
   * @see Valide.iDocument.title
   */
  public string title
  {
    owned get
    {
      return this.filename;
    }
  }

  /**
   * @see Valide.iDocument.filename
   */
  public string filename
  {
    owned get
    {
      return _("Start page");
    }
  }

  [CCode (instance_pos = -1)]
  private void create_project (LinkLabel sender, string? uri)
  {
    if (this.projects.create ())
    {
      this.close ();
    }
  }

  [CCode (instance_pos = -1)]
  private void open_project (LinkLabel sender, string? uri)
  {
    try
    {
      if(this.projects.open (uri))
      {
        this.close ();
      }
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  [CCode (instance_pos = -1)]
  private void create_file (LinkLabel sender, string? uri)
  {
    if (this.projects.create ())
    {
      this.close ();
    }
  }

  [CCode (instance_pos = -1)]
  private void open_file (LinkLabel sender, string? uri)
  {
    this.close ();
    try
    {
      this.documents.create (uri);
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  [CCode (instance_pos = -1)]
  private void open_link (LinkLabel sender, string? uri)
  {
    try
    {
      AppInfo.launch_default_for_uri (uri, null);
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  private bool label_enter_notify_event (Gtk.Widget sender,
                                         Gdk.EventCrossing event)
  {
    string status_message;
    LinkLabel link_label;

    link_label = sender as LinkLabel;
    if (link_label.uri != null)
    {
      status_message = _("Open %s").printf (link_label.uri);
    }
    else
    {
      status_message = "";
    }
    this.send_message (status_message);
    return false;
  }

  private Gtk.Frame create_frame (string title, List<Item?> items)
  {
    Gtk.VBox vbox;
    Gtk.Frame frame;
    Gtk.Label label;
    Gtk.Alignment alignment;
    Pango.AttrList attr_list;

    frame = new Gtk.Frame (title);
    attr_list = new Pango.AttrList ();
    attr_list.insert (Pango.attr_weight_new (Pango.Weight.BOLD));
    label = frame.label_widget as Gtk.Label;
    label.set_attributes (attr_list);
    frame.set_shadow_type (Gtk.ShadowType.NONE);

    alignment = new Gtk.Alignment (0, 0, 0, 0);
    alignment.left_padding = 12;
    frame.add (alignment);

    vbox = new Gtk.VBox (false, 5);
    vbox.set_border_width (5);
    alignment.add (vbox);

    if (items != null)
    {
      foreach (Item item in items)
      {
        LinkLabel link_label;

        link_label = new LinkLabel (item.uri, item.text, item.icon);
        link_label.enter_notify_event.connect (this.label_enter_notify_event);
  
        Signal.connect (link_label, "open_uri", item.callback, this);
        vbox.pack_start (link_label, false, false, 0);
      }
    }
    else
    {
      label = new Gtk.Label (_("No item"));
      vbox.pack_start (label, false, false, 0);
    }
    return frame;
  }

  private Gtk.Frame populate_frame (List<Gtk.RecentInfo> infos, string title, Callback callback)
  {
    Item item;
    List<Item?> items;

    items = new List<Item?> ();
    foreach (Gtk.RecentInfo info in infos)
    {
      item = Item ();
      item.icon = null;
      item.uri = info.get_uri ();
      item.text = Path.get_basename (item.uri);
      item.callback = callback;

      items.append (item);
    }
    return this.create_frame (title, items);
  }

  private void populate_file_frame ()
  {
    if (this.file_frame != null)
    {
      this.file_frame.destroy ();
    }
    this.file_frame = this.populate_frame (this.recents.recent_files,
                                           _("Recent files"),
                                           (Callback)this.open_file);
    this.recent_box.pack_start (this.file_frame, false, false, 0);
    this.recent_box.reorder_child (this.file_frame, 2);
    this.file_frame.show_all ();
  }

  private void populate_project_frame ()
  {
    if (this.project_frame != null)
    {
      this.project_frame.destroy ();
    }
    this.project_frame = this.populate_frame (this.recents.recent_projects,
                                              _("Recent projects"),
                                              (Callback)this.open_project);
    this.recent_box.pack_start (this.project_frame, false, false, 0);
    this.recent_box.reorder_child (this.project_frame, 0);
    this.project_frame.show_all ();
  }

  /**
   * Create a new Valide.StartPage
   *
   * @param documents The document manager
   * @param projects The project manager
   * @param recents The recents files manager
   */
  public StartPage (DocumentManager documents, ProjectManager projects,
                    RecentManagerUI recents)
  {
    this.documents = documents;
    this.projects = projects;
    this.recents = recents;
  }

  construct
  {
    Gtk.VBox vbox;
    Gtk.Image logo;
    Gtk.Label label;
    Gtk.Frame frame;
    List<Item?> items;
    Gtk.HSeparator sep;

    this.resize (7, 2);

    logo = new Gtk.Image.from_file (Path.build_filename (Config.PIXMAPS_DIR, "logo.png"));
    this.attach (logo, 0, 2, 0, 1, Gtk.AttachOptions.EXPAND,
                 Gtk.AttachOptions.EXPAND, 0, 0);

    vbox = new Gtk.VBox (false, 5);
    vbox.set_border_width (5);
    this.attach (vbox, 0, 1, 1, 5, Gtk.AttachOptions.EXPAND,
                 Gtk.AttachOptions.EXPAND, 0, 0);

    items = new List<Item?> ();

    /* New project */
    Item item = Item ();
    item.icon = Utils.get_pixbuf_for_stock (Gtk.STOCK_NEW, Gtk.IconSize.BUTTON);
    item.uri = null;
    item.text = _("New project");
    item.callback = (Callback)this.create_project;
    items.append (item);

    /* Open project */
    item = Item ();
    item.icon = Utils.get_pixbuf_for_stock (Gtk.STOCK_OPEN, Gtk.IconSize.BUTTON);
    item.uri = null;
    item.text = _("Open project");
    item.callback = (Callback)this.open_project;
    items.append (item);

    frame = this.create_frame ("", items);
    vbox.pack_start (frame, true, true, 10);

    vbox.pack_start (new Gtk.HSeparator (), false, false, 10);

    items = new List<Item?> ();

    /* Website */
    item = Item ();
    item.icon = Utils.get_pixbuf_for_stock (Gtk.STOCK_HOME, Gtk.IconSize.BUTTON);
    item.uri = Config.WEBSITE;
    item.text = _("Website");
    item.callback = (Callback)this.open_link;
    items.append (item);

    /* Report a bug */
    item = Item ();
    item.icon = Utils.get_pixbuf_for_stock (Gtk.STOCK_CANCEL, Gtk.IconSize.BUTTON);
    item.uri = "http://code.google.com/p/valide/issues/entry";
    item.text = _("Report a bug");
    item.callback = (Callback)this.open_link;
    items.append (item);


    /* Visit the user group */
    item = Item ();
    item.icon = Utils.get_pixbuf_for_stock (Gtk.STOCK_HELP, Gtk.IconSize.BUTTON);
    item.uri = "http://groups.google.com/group/valide-users";
    item.text = _("Visit the user group");
    item.callback = (Callback)this.open_link;
    items.append (item);

    frame = this.create_frame ("", items);
    vbox.pack_start (frame, true, true, 10);

    this.recent_box = new Gtk.VBox (false, 5);
    this.recent_box.set_border_width (5);
    this.attach (this.recent_box, 1, 2, 1, 5, Gtk.AttachOptions.EXPAND,
                 Gtk.AttachOptions.EXPAND, 0, 0);

    this.populate_project_frame ();
    this.recent_box.pack_start (new Gtk.HSeparator (), false, false, 10);
    this.populate_file_frame ();

    this.recents.update_recent_projects.connect (() => {
      this.populate_project_frame ();
    });

    this.recents.update_recent_files.connect (() => {
      this.populate_file_frame ();
    });

    sep = new Gtk.HSeparator ();
    this.attach (sep, 0, 2, 5, 6, Gtk.AttachOptions.FILL,
                 Gtk.AttachOptions.FILL, 0, 5);

    label = new Gtk.Label (Config.COPYRIGHT);
    this.attach (label, 0, 2, 6, 7, Gtk.AttachOptions.FILL,
                 Gtk.AttachOptions.FILL, 0, 0);
  }

  /**
   * @see Valide.iDocument.open
   */
  public void open (string uri) throws Error
  {
  }

  /**
   * @see Valide.iDocument.save
   */
  public void save (string? current_folder = null)
  {
  }

  /**
   * @see Valide.iDocument.save_as
   */
  public void save_as (string? current_folder = null)
  {
  }

  /**
   * @see Valide.iDocument.rename
   */
  public void rename ()
  {
  }

  /**
   * @see Valide.iDocument.close
   */
  public bool close ()
  {
    Gdk.Cursor cursor;

    cursor = new Gdk.Cursor (Gdk.CursorType.ARROW);
    this.window.set_cursor (cursor);
    this.closed ();
    return true;
  }

  /**
   * @see Valide.iDocument.redo
   */
  public void redo ()
  {
  }

  /**
   * @see Valide.iDocument.search
   */
  public void search ()
  {
  }

  /**
   * @see Valide.iDocument.find_next
   */
  public void find_next ()
  {
  }

  /**
   * @see Valide.iDocument.find_prev
   */
  public void find_prev ()
  {
  }

  /**
   * @see Valide.iDocument.replace
   */
  public void replace ()
  {
  }

  /**
   * @see Valide.iDocument.goto_line
   */
  public bool goto_line (int line = -1)
  {
    return true;
  }

  /**
   * @see Valide.iDocument.setup_ui
   */
  public void setup_ui (UIManager ui_manager)
  {
    ui_manager.get_action ("file-save").sensitive = false;
    ui_manager.get_action ("file-save-as").sensitive = false;
    ui_manager.get_action ("file-rename").sensitive = false;
    ui_manager.get_action ("edit-copy").sensitive = false;
    ui_manager.get_action ("edit-cut").sensitive = false;
    ui_manager.get_action ("edit-paste").sensitive = false;
    ui_manager.get_action ("edit-search").sensitive = false;
    ui_manager.get_action ("edit-find-next").sensitive = false;
    ui_manager.get_action ("edit-find-prev").sensitive = false;
    ui_manager.get_action ("edit-replace").sensitive = false;
    ui_manager.get_action ("edit-goto-line").sensitive = false;
    ui_manager.get_action ("edit-goto-definition").sensitive = false;
    ui_manager.get_action ("edit-undo").sensitive = false;
    ui_manager.get_action ("edit-redo").sensitive = false;
  }

  /**
   * @see Valide.iDocument.get_icon
   */
  public Gdk.Pixbuf get_icon ()
  {
    return Utils.get_pixbuf_for_stock (Gtk.STOCK_INFO, Gtk.IconSize.MENU);
  }

  /**
   * @see Valide.iDocument.get_tooltips
   */
  public string get_tooltips ()
  {
    return _("Start here");
  }
}

