/* cache.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *   Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public errordomain Valide.Symbols.CacheError
{
  KIND,
  UNSUPPORTED
}

public enum Valide.Symbols.SortType
{
  KIND,
  SCOPE
}

public class Valide.Symbols.CachePreferences : Gtk.VBox
{
  private Gtk.Button clear_button;
  private Gtk.Button build_button;
  private Gtk.ListStore list_store;

  private enum Col
  {
    PATH,
    COUNT,
    NB
  }

  public Cache cache
  {
    get;
    construct;
  }

  private bool populate_tree_view ()
  {
    try
    {
      this.list_store.clear ();
      this.cache.get_statistics ((k, v) => {
        Gtk.TreeIter iter;

        this.list_store.append (out iter);
        this.list_store.set (iter, Col.PATH, k, Col.COUNT, v);
        Utils.process_gtk_events ();
        return true;
      });
      if (this.list_store.iter_n_children (null) > 0)
      {
        this.clear_button.set_sensitive (true);
      }
      else
      {
        this.clear_button.set_sensitive (false);
      }
    }
    catch (Error e)
    {
      warning (e.message);
    }
    return false;
  }

  public CachePreferences (Cache cache)
  {
    this.cache = cache;
  }

  construct
  {
    Gtk.VBox vbox;
    Gtk.HBox hbox;
    Gtk.TreeView tree_view;
    Gtk.TreeViewColumn col;
    Gtk.CellRenderer renderer;
    Gtk.ScrolledWindow scrolled_window;

    vbox = new Gtk.VBox (false, 5);

    scrolled_window = new Gtk.ScrolledWindow (null, null);
    scrolled_window.set_policy (Gtk.PolicyType.AUTOMATIC,
                                Gtk.PolicyType.AUTOMATIC);
    vbox.pack_start (scrolled_window, true, true, 0);

    this.list_store = new Gtk.ListStore (Col.NB, typeof (string), typeof (int));
    list_store.set_sort_column_id (Col.PATH, Gtk.SortType.ASCENDING);

    tree_view = new Gtk.TreeView.with_model (this.list_store);
    scrolled_window.add (tree_view);

    col = new Gtk.TreeViewColumn ();
    col.set_title(_("File"));

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "text", Col.PATH);

    tree_view.append_column (col);

    col = new Gtk.TreeViewColumn ();
    col.set_title(_("Symbols"));

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "text", Col.COUNT);

    tree_view.append_column (col);

    hbox = new Gtk.HBox (true, 0);
    vbox.pack_start (hbox, false, true, 0);

    this.clear_button = new Gtk.Button.with_label (_("Clear cache"));
    hbox.pack_start (this.clear_button, false, false, 0);
    this.clear_button.clicked.connect (() => {
      this.cache.clear ();
      this.populate_tree_view ();
    });
    this.build_button = new Gtk.Button.with_label (_("Rebuild cache"));
    hbox.pack_start (this.build_button, false, false, 0);
    this.build_button.clicked.connect (() => {
      this.cache.rebuild ();
      this.populate_tree_view ();
    });

    Idle.add (this.populate_tree_view);
  }
}

public class Valide.Symbols.Cache : GLib.Object
{
  private static Cache instance;
  private Engine engine;
  private bool building;

  /**
   * The signal cache_update is emited when the database is updated
   */
  public signal void cache_updated ();

  /**
   * Delegate for retrieve the cache statistisques
   */
  public delegate bool StatisticsCallback (string path, int count);

  /**
   * Delegate for retrieve the symbol list
   */
  public delegate bool ListSymbolCallback (Symbol symbol);

  /**
   * This function needs work. Used to be custom parsed for ex searches
   * which could potentially contain tabs, however, I've decided that isn't
   * really needed. But, it's still using Symbol. We could just pass back
   * **fields array.
   */ 
  private Symbol parse_line (string line) throws CacheError
  {
    /* (parent.)name file ex_cmd;" kind:[enumvalue|errorcode|errordomain|enum|struct|namespace] line:x language:x access:x */
    /* (parent.)name file ex_cmd;" kind:field line:x language:x typeref:x access:x */
    /* (parent.)name file ex_cmd;" kind:class line:x language:x inherits:x access:x implementation:x */
    /* (parent.)name file ex_cmd;" kind:[delegate|method|signal] line:x language:x typeref:x access:x implementation:x signature:x */
    /* (parent.)name file ex_cmd;" kind:interface line:x language:x inherits:x access:x */
    /* (parent.)name file ex_cmd;" kind:property line:x language:x typeref:x access:x implementation:x */

    string[] fields;
    Symbol symbol = null;

    fields = line.split ("\t", 0);
    if (fields[0] != null)
    {
      string filename;
      string name;
      string full_name;
      string kind;
      int file_position;
      string language;
      string access;
      string[] inherits = null;
      string typeref = null;
      string signature = null;
      string implementation = null;
      string[] splits;

      symbol = new Symbol ();
      full_name = fields[0];
      if (full_name.rchr (-1, '.') != null)
      {
        name = full_name.rchr (-1, '.').offset (1);
      }
      else
      {
        name = full_name;
      }
      filename = fields[1];
      splits = fields[3].split(":", 0);
      kind = splits[1];

      splits = fields[4].split(":", 0);
      file_position = splits[1].to_int ();

      splits = fields[5].split(":", 0);
      language = fields[1];

      switch (kind)
      {
        case "enumvalue":
        case "errorcode":
        case "errordomain":
        case "enum":
        case "namespace":
        case "struct":
          splits = fields[6].split(":", 0);
          access = splits[1];
        break;
        case "field":
          splits = fields[6].split(":", 0);
          typeref = splits[1];
          splits = fields[7].split(":", 0);
          access = splits[1];
        break;
        case "class":
          splits = fields[6].split(":", 0);
          if (splits != null && splits[1] != null)
          {
            inherits = splits[1].split (",", 0);
          }
          splits = fields[7].split(":", 0);
          access = splits[1];
          splits = fields[8].split(":", 0);
          implementation = splits[1];
        break;
        case "delegate":
        case "method":
        case "signal":
          splits = fields[6].split(":", 0);
          typeref = splits[1];
          splits = fields[7].split(":", 0);
          access = splits[1];
          splits = fields[8].split(":", 0);
          implementation = splits[1];
          splits = fields[9].split(":", 0);
          signature = splits[1];
        break;
        case "interface":
          splits = fields[6].split(":", 0);
          if (splits != null && splits[1] != null)
          {
            inherits = splits[1].split (",", 0);
          }
          splits = fields[7].split(":", 0);
          access = splits[1];
        break;
        case "property":
          splits = fields[6].split(":", 0);
          typeref = splits[1];
          splits = fields[7].split(":", 0);
          access = splits[1];
          splits = fields[8].split(":", 0);
          implementation = splits[1];
        break;
        default:
          throw new CacheError.KIND (_("Unknow kind '%s' for symbol '%s'"), kind, name);
        break;
      }

      symbol.name = name;
      symbol.full_name = full_name;
      symbol.file = new File (filename, language);
      symbol.file_position = file_position;
      symbol.signature = signature;
      symbol.inherits = inherits;
      symbol.typeref = typeref;
      symbol.root = null;
      if (kind != null)
      {
        symbol.kind = new Kind (kind);
      }
      if (access != null)
      {
        symbol.access = new Access (access);
      }
      if (implementation != null)
      {
        symbol.implementation = new Implementation (implementation);
      }
    }
    return symbol;
  }

  /**
   * Execute the ctags command and return stdout.
   */ 
  private string exec_ctags (string filename) throws Error
  {
    string output;
    string command;

    /* build ctags command */
    command = "%s --extra=+q --fields=afmiKlnsStz -f - \"%s\""
                       .printf (Config.VALIDE_CTAGS_EXEC, filename);
    /* execute command */
    Process.spawn_command_line_sync (command, out output, null, null);

    return output;
  }

  private bool need_rebuild ()
  {
    File file;
    string filename;

    filename = Valac.get_vapi_filename ("glib-2.0");
    file = new File (filename, "vala");
    return !(file.exist () && file.up_to_date);
  }

  private Cache ()
  {
  }

  construct
  {
    this.building = false;
    this.engine = Engine.get_instance ();
    if (this.need_rebuild ())
    {
      this.rebuild ();
    }
  }

  /**
   * Get the uniq instance of the Valide.Symbols.Cache
   *
   * @return The Valide.Symbols.Cache instance
   */
  public static Cache get_instance ()
  {
    if (instance == null)
    {
      instance = new Cache ();
    }
    return instance;
  }

  /**
   * Add a new vapi file
   *
   * @param vapi The vapi file path
   */
  public void add_vapi (string vapi)
  {
    string filename;

    filename = Valac.get_vapi_filename (vapi);
    this.add_file (filename, "vala");
  }

  /**
   * Check if the vapi already exist in the database
   *
   * @param vapi The vapi file path
   *
   * @return true if the file exist
   */
  public bool vapi_exist (string vapi)
  {
    File file;
    string filename;

    filename = Valac.get_vapi_filename (vapi);
    file = new File (filename, "vala");
    return file.exist ();
  }

  /**
   * Add a new source file
   *
   * @param filename The file path
   * @param language The programming language of the file
   */
  public void add_file (string filename, string language = "vala")
  {
    try
    {
      bool update = false;
      File file;

      this.engine.begin_transaction ();
      file = new File (filename, language);
      if (file.exist () && !file.up_to_date)
      {
        file.remove ();
      }
      if (!file.exist ())
      {
        file.add ();
        update = true;
      }
      if (update)
      {
        this.update_symbols (filename);
      }
      this.engine.end_transaction ();
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  /**
   * Update the symbol list for a file
   *
   * @param filename The name of the file
   */
  public void update_symbols (string filename)
  {
    string output;
    Symbol symbol;
    string[] lines;
    Symbol root = null;

    output = this.exec_ctags (filename);
    lines = output.split ("\n", 0);
    for (int i = 0; lines[i] != null; i++)
    {
      try
      {
        symbol = this.parse_line (lines[i]);
        if (symbol != null)
        {
          if (root != null)
          {
            string parent;

            do
            {
              parent = root.full_name;
              if (symbol.full_name.has_prefix (parent + "."))
              {
                symbol.root = root;
              }
              else
              {
                root = root.root;
              }
            } while (symbol.root == null && root != null);
          }
          symbol.add ();
          if (symbol.kind.name == "class"
              || symbol.kind.name == "struct"
              || symbol.kind.name == "enum"
              || symbol.kind.name == "errordomain"
              || symbol.kind.name == "namespace"
              || symbol.kind.name == "interface")
          {
            root = symbol;
          }
        }
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
    this.cache_updated ();
  }

  /**
   * Retrieve all symbols for a file
   *
   * @param filename The file path
   * @param sort The type sorting
   * @param callback A callback for get the symbol list
   */
  public void get_symbols_for_file (string filename, SortType sort,
                                    ListSymbolCallback callback) throws EngineError
  {
    Symbol symbol;
    string sql = null;
    Sqlite.Statement statement;

    switch (sort)
    {
      case SortType.KIND:
        sql = """SELECT symbol.id
                   FROM symbol
                     JOIN file ON symbol.file_id = file.id
                     JOIN kind ON symbol.kind_id = kind.id
                   WHERE file.path = ?
                   ORDER BY kind.name, symbol.name;""";
      break;
      case SortType.SCOPE:
        sql = """SELECT symbol.id
                   FROM symbol
                     JOIN file ON symbol.file_id = file.id
                   WHERE file.path = ?
                   ORDER BY file.path, symbol.file_position;""";
      break;
    }
    statement = this.engine.prepare (sql);
    statement.bind_text (1, filename);
    while (statement.step () == Sqlite.ROW)
    {
      symbol = new Symbol.from_id (statement.column_int (0));
      if (!callback (symbol))
      {
        break;
      }
    }
  }

  /**
   * Retrieve all completion symbols for a symbols
   *
   * @param filename The file path
   * @param type The type of the symbol
   * @param filter A filter for symbol
   * @param callback A callback for get the symbol list
   */
  public void get_completion_for_symbol (string filename, Symbol type, string filter,
                                         ListSymbolCallback callback) throws EngineError
  {
    if (type.exist ())
    {
      Symbol symbol;
      string full_name;
      string sql = null;
      Sqlite.Statement statement;

      full_name = type.full_name.replace ("_", "\\_");
      sql = """SELECT symbol.id
                 FROM symbol
                   INNER JOIN file ON file.id = symbol.file_id
                   INNER JOIN access ON access.id = symbol.access_id
                 WHERE symbol.full_name LIKE '%s.%s%%' ESCAPE '\'
                   AND (file.path=  ? OR access.name = 'public')
                 ORDER BY symbol.name;""".printf (full_name, filter);

      statement = this.engine.prepare (sql);
      statement.bind_text (1, filename);
      while (statement.step () == Sqlite.ROW)
      {
        symbol = new Symbol.from_id (statement.column_int (0));
        if (!callback (symbol))
        {
          break;
        }
      }
      foreach (string d in type.inherits)
      {
        Symbol base_type;

        base_type = new Symbol.from_full_name (d);
        this.get_completion_for_symbol (filename, base_type, filter, callback);
      }
    }
  }

  /**
   * Get the preferences widget
   *
   * @return The preferences widget
   */
  public Gtk.Widget preferences ()
  {
    return new CachePreferences (this);
  }

  /**
   * Clear the database
   */
  public void clear ()
  {
    this.engine.delete_db ();
    this.cache_updated ();
  }

  /**
   * Get the cache statistics
   *
   * @param callback A callback for get the statistics
   */
  public void get_statistics (StatisticsCallback callback) throws EngineError
  {
    Sqlite.Statement statement;

    statement = this.engine.prepare ("SELECT file.path, COUNT (symbol.id) FROM file INNER JOIN symbol ON file.id = symbol.file_id GROUP BY file.id;");
    while (statement.step () == Sqlite.ROW)
    {
      if (!callback (Path.get_basename (statement.column_text (0)),
                     statement.column_int (1)))
      {
        break;
      }
    }
  }

  /**
   * Rebuild the symbols cache
   */
  public void rebuild ()
  {
    if (!this.building)
    {
      Gtk.Label label;
      Gtk.Dialog dialog;
      Gtk.Button no_button;
      Gtk.Button yes_button;
      Gtk.Button cancel_button;
      Gtk.ProgressBar progress_bar;

      this.building = true;
      dialog = new Gtk.Dialog ();
      yes_button = dialog.add_button (Gtk.STOCK_YES, Gtk.ResponseType.YES) as Gtk.Button;
      no_button = dialog.add_button (Gtk.STOCK_NO, Gtk.ResponseType.NO) as Gtk.Button;
      cancel_button = dialog.add_button (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL) as Gtk.Button;

      label = new Gtk.Label (_("Would you like to create/update the symbols database now?"));
      dialog.vbox.pack_start (label, false, false, 0);

      progress_bar = new Gtk.ProgressBar ();
      progress_bar.set_no_show_all (true);
      dialog.vbox.pack_start (progress_bar, false, true, 5);

      dialog.show_all ();
      cancel_button.hide ();
      if (dialog.run () == Gtk.ResponseType.YES)
      {
        List<string> pkg;

        pkg = Valac.get_package_list ();
        if (pkg.length () > 0)
        {
          int i = 0;
          string msg;
          double fraction;

          no_button.hide ();
          yes_button.hide ();
          cancel_button.show ();
          cancel_button.clicked.connect ((s) => {
            s.sensitive = false;
            this.building = false;
          });
          dialog.destroy.connect ((s) => {
            this.building = false;
          });
          label.set_text (_("Update in progress..."));
          progress_bar.show ();
          fraction = 1.0/pkg.length ();
          foreach (string p in pkg)
          {
            if (!this.building)
            {
              break;
            }
            Utils.process_gtk_events ();
            i++;
            msg = "%s (%d/%u)".printf (p, i, pkg.length ());
            progress_bar.set_text (msg);
            this.add_vapi (p);
            progress_bar.set_fraction (fraction * i);
          }
          this.cache_updated ();
        }
      }
      dialog.destroy ();
      this.building = false;
    }
    else
    {
      message (_("Rebuild in progress"));
    }
  }

  /**
   * Got to the definition of the current symbol
   *
   * @param documents The document manager
   */
  public void goto_definition (DocumentManager documents)
  {
    if (documents.current is AbstractTextDocument)
    {
      Parser parser;
      Symbol symbol;
      SourceBuffer buffer;

      buffer = (documents.current as AbstractTextDocument).buffer;

      parser = new Parser ();
      symbol = parser.get_current_symbol (buffer);
      if (symbol != null)
      {
        iDocument document;

        document = documents.create (symbol.file.path);
        document.goto_line (symbol.file_position - 1);
      }
    }
  }
}

