/* engine.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public errordomain Valide.Symbols.EngineError
{
  BIND,
  TRANSACTION,
  PREPARE,
  EXEC,
  FILENAME,
  OPEN
}

/**
 * The symbols engine
 */
public class Valide.Symbols.Engine : GLib.Object
{
  private static Engine instance;
  private Sqlite.Database _db;

  private string db_file
  {
    owned get
    {
      return Path.build_filename (ConfigManager.get_instance ().home_dir, ".valide_symbol_db");
    }
  }

  private unowned Sqlite.Database db
  {
    get
    {
      if (this._db == null)
      {
        this._db = this.open_database ();
      }
      return this._db;
    }
  }

  private Sqlite.Database open_database () throws EngineError
  {
    Sqlite.Database database;
    bool create_table = false;

    if (!FileUtils.test (this.db_file, FileTest.EXISTS))
    {
      create_table = true;
    }

    if (Sqlite.Database.open (this.db_file, out database) == Sqlite.OK)
    {
      if (create_table)
      {
        string msg;

        /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=579364 */
        if (Fix.SqliteDatabase.exec (database, this.create_query, null, out msg) != Sqlite.OK)
        {
          throw new EngineError.EXEC (_("Create database: %s"), msg);
        }
      }
    }
    else
    {
      throw new EngineError.OPEN (_("Open database: %s"), database.errmsg ());
    }
    return database;
  }

  private Engine ()
  {
  }

  construct
  {
    try
    {
      this._db = this.open_database ();
    }
    catch (EngineError e)
    {
      error (e.message);
    }
  }

  /**
   * Get the uniq instance of the Valide.Symbols.Engine
   *
   * @return The Valide.Symbols.Engine instance
   */
  public static Engine get_instance ()
  {
    if (instance == null)
    {
      instance = new Engine ();
    }
    return instance;
  }

  /**
   * Prepare SQL query
   *
   * @param sql The sql query
   *
   * @return The prepared statement
   */
  public Sqlite.Statement prepare (string sql) throws EngineError
  {
    Sqlite.Statement statement;

    if (this.db.prepare_v2 (sql, -1, out statement) != Sqlite.OK)
    {
      throw new EngineError.PREPARE ("Couldn't prepare the sql query '%s': %s",
                                     sql, this.errmsg ());
    }
    return statement;
  }

  /**
   * Execute a SQL query
   *
   * @param sql The SQL query
   */
  public void exec (string sql) throws EngineError
  {
    string msg;

    /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=579364 */
    Fix.SqliteDatabase.exec (this.db, sql, null, out msg);
    if (msg != null)
    {
      throw new EngineError.EXEC (msg);
    }
  }

  /**
   * Get the database error
   */
  public string errmsg ()
  {
    return this.db.errmsg ();
  }

  /**
   * Begin a transaction
   */
  public void begin_transaction () throws EngineError
  {
    string msg;

    /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=579364 */
    if (Fix.SqliteDatabase.exec (this.db, "begin", null, out msg) != Sqlite.OK)
    {
      throw new EngineError.TRANSACTION (msg);
    }
  }

  /**
   * End a transaction
   */
  public void end_transaction () throws EngineError
  {
    string msg;

    /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=579364 */
    if (Fix.SqliteDatabase.exec (this.db, "commit", null, out msg) != Sqlite.OK)
    {
      throw new EngineError.TRANSACTION (msg);
    }
  }

  /**
   * Delete the database
   */
  public void delete_db ()
  {
    if (FileUtils.remove (this.db_file) == 0)
    {
      this._db = null;
    }
    else
    {
      debug (_("Couldn't remove the '%s' file"), this.db_file);
    }
  }

  private const string create_query = """
CREATE TABLE file (id integer PRIMARY KEY AUTOINCREMENT,
                   path TEXT not null unique,
                   language_id integer REFERENCES language (id),
                   mtime DATE
                  );

CREATE TABLE language (id integer PRIMARY KEY AUTOINCREMENT,
                       name varchar (50) not null unique
                      );

CREATE TABLE symbol (id integer PRIMARY KEY AUTOINCREMENT,
                     name varchar (256) not null,
                     full_name varchar (256) not null,
                     file_id integer not null REFERENCES file (file_id),
                     file_position integer,
                     signature varchar (256),
                     root_id integer REFERENCES symbol (id),
                     kind_id integer REFERENCES kind (id),
                     access_id integer REFERENCES access (id),
                     implementation_id integer REFERENCES implementation (id),
                     typeref varchar (256),
                     unique (name, file_id, file_position)
                    );

CREATE TABLE kind (id integer PRIMARY KEY AUTOINCREMENT,
                   name varchar (50) not null unique
                  );

CREATE TABLE access (id integer PRIMARY KEY AUTOINCREMENT,
                     name varchar (50) not null unique
                    );

CREATE TABLE implementation (id integer PRIMARY KEY AUTOINCREMENT,
                             name varchar (50) not null unique
                            );

CREATE TABLE heritage (id integer PRIMARY KEY AUTOINCREMENT,
                       symbol_id_base integer REFERENCES symbol (id),
                       symbol_name_derived varchar (256) REFERENCES symbol (full_name),
                       unique (symbol_id_base, symbol_name_derived)
                      );

CREATE INDEX symbol_idx_1 ON symbol (name);

CREATE INDEX symbol_idx_2 ON symbol (name, file_id, file_position);

CREATE INDEX symbol_idx_3 ON symbol (full_name);

CREATE INDEX file_idx_1 ON file (path);

CREATE TRIGGER delete_file_trg BEFORE DELETE ON file
FOR EACH ROW
BEGIN
  DELETE FROM symbol WHERE file_id = (SELECT id FROM file WHERE path = old.path);
END;

CREATE TRIGGER delete_symbol_trg BEFORE DELETE ON symbol
FOR EACH ROW
BEGIN
  DELETE FROM heritage WHERE heritage.symbol_id_base = old.id OR heritage.symbol_name_derived = old.full_name;
  UPDATE symbol SET root_id = '-1' WHERE symbol.root_id = old.id;
END;""";
}

