/* file.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public class Valide.Symbols.File : Object
{
  /**
   * The file path
   */
  public string path;

  /**
   * The programming language
   */
  public Language language;

  /**
   * The last modification time
   */
  public uint64 mtime
  {
    get;
    private set;
  }

  /**
   * Check if the file is up to date
   */
  public bool up_to_date
  {
    get
    {
      bool ret = true;

      try
      {
        ret = (Utils.get_mtime (this.path) <= this.mtime);
      }
      catch (Error e)
      {
        debug (e.message);
      }
      return ret;
    }
  }

  /**
   * @see Valide.Symbols.Object.table_name
   */
  protected override string table_name
  {
    owned get
    {
      return "file";
    }
  }

  /**
   * @see Valide.Symbols.Object.table_name
   */
  protected override string add_query
  {
    owned get
    {
      return "INSERT INTO file (path, language_id, mtime) VALUES (?, ?, ?);";
    }
  }

  /**
   * @see Valide.Symbols.Object.exist_query
   */
  protected override string exist_query
  {
    owned get
    {
      return "SELECT id FROM file WHERE path = ?;";
    }
  }

  /**
   * @see Valide.Symbols.Object.bind_exist_stmt
   */
  protected override void bind_exist_stmt ()
  {
    this.exist_stmt.bind_text (1, this.path);
  }

  /**
   * Create a new file
   *
   * @param path The file path
   * @param language The programmation language of the file
   */
  public File (string path, string language)
  {
    this.path = path;
    this.language = new Language (language);
    if (this.exist ())
    {
      File file;

      file = new File.from_id (this.id);
      this.mtime = file.mtime;
    }
    else
    {
      try
      {
        this.mtime = Utils.get_mtime (this.path);
      }
      catch (Error e)
      {
        this.mtime = 0;
      }
    }
  }

  /**
   * Create a new file from id
   *
   * @param id The file id
   */
  public File.from_id (int id)
  {
    string sql;
    Sqlite.Statement statement;

    sql = """SELECT id, path, language_id, mtime
               FROM file
               WHERE id = ?;""";
    statement = this.engine.prepare (sql);
    statement.bind_int (1, id);
    if (statement.step () == Sqlite.ROW)
    {
      this.id = statement.column_int (0);
      this.path = statement.column_text (1);
      this.language = new Language.from_id (statement.column_int (2));
      this.mtime = (uint64)statement.column_int64 (3);
    }
  }

  /**
   * @see Valide.Symbols.Object.add
   */
  public override int add () throws EngineError
  {
    if (FileUtils.test (this.path, FileTest.EXISTS))
    {
      if (!this.exist ())
      {
        int language_id = -1;

        if (!this.language.exist ())
        {
          language_id = this.language.add ();
        }
        this.add_stmt.bind_text (1, this.path);
        this.add_stmt.bind_int (2, language_id);
        this.add_stmt.bind_int64 (3, (int64)Utils.get_mtime (this.path));
        if (this.add_stmt.step () == Sqlite.DONE)
        {
          this.id = this.get_last_id ();
        }
        else
        {
          throw new EngineError.EXEC (_("Add file '%s': %s"),
                                      this.path, this.engine.errmsg ());
        }
      }
      else
      {
        throw new EngineError.FILENAME (_("The file '%s' doesn't exist!"),
                                        this.path);
      }
      this.add_stmt.reset ();
    }
    return this.id;
  }
}

