/* object.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public abstract class Valide.Symbols.Object : GLib.Object
{
  /**
   * Statement for the add query
   */
  protected Sqlite.Statement add_stmt;

  /**
   * Statement for the exist query
   */
  protected Sqlite.Statement exist_stmt;

  /**
   * Statement for the remove query
   */
  protected Sqlite.Statement remove_stmt;

  /**
   * Statement for get the last id
   */
  protected Sqlite.Statement last_id_stmt;

  /**
   * The SQL engine
   */
  protected static Engine engine;

  /**
   * The name of SQL table
   */
  protected abstract string table_name { owned get; }

  /**
   * The SQL query for add a new row
   */
  protected abstract string add_query { owned get; }

  /**
   * The SQL query for test the exist of a row
   */
  protected abstract string exist_query { owned get; }

  /**
   * The SQL query for remove a row
   */
  protected string remove_query
  {
    owned get
    {
      return "DELETE FROM %s WHERE id = ?;".printf (this.table_name);
    }
  }

  /**
   * The internal id
   */
  public int id
  {
    get;
    protected set;
    default = -1;
  }

  /**
   * Bind the parameters for the exist query
   */
  protected abstract void bind_exist_stmt ();

  /**
   * Get the last inserted id in this table
   */
  protected int get_last_id () throws EngineError
  {
    int ret;
    int last_id = 0;

    if (this.last_id_stmt.bind_text (1, this.table_name) == Sqlite.OK)
    {
      ret = this.last_id_stmt.step ();
      if (ret == Sqlite.ROW)
      {
        last_id = this.last_id_stmt.column_int (0);
      }
      else if (ret != Sqlite.DONE)
      {
        throw new EngineError.EXEC (this.engine.errmsg ());
      }
    }
    else
    {
      throw new EngineError.BIND (this.engine.errmsg ());
    }
    this.last_id_stmt.reset ();
    return last_id;
  }

  construct
  {
    try
    {
      this.engine = Engine.get_instance ();
      this.add_stmt = this.engine.prepare (this.add_query);
      this.exist_stmt = this.engine.prepare (this.exist_query);
      this.remove_stmt = this.engine.prepare (this.remove_query);
      this.last_id_stmt = this.engine.prepare ("SELECT seq FROM sqlite_sequence WHERE name = ?;");
    }
    catch (Error e)
    {
      error (e.message);
    }
  }

  /**
   * Add a new row
   */
  public abstract int add () throws EngineError;

  /**
   * Test if this object exist in the SQL database
   */
  public bool exist () throws EngineError
  {
    if (this.id == -1)
    {
      int ret;

      this.bind_exist_stmt ();
      ret = this.exist_stmt.step ();
      if (ret == Sqlite.ROW)
      {
        this.id = this.exist_stmt.column_int (0);
      }
      else if (ret != Sqlite.DONE)
      {
        throw new EngineError.EXEC (this.engine.errmsg ());
      }
      else
      {
        this.id = 0;
      }
      this.exist_stmt.reset ();
    }
    return (this.id > 0);
  }

  /**
   * Remove this object from the database
   */
  public void remove () throws EngineError
  {
    if (this.exist ())
    {
      this.remove_stmt.bind_int (1, this.id);
      if (this.remove_stmt.step () == Sqlite.DONE)
      {
        this.id = -1;
      }
      else
      {
        throw new EngineError.EXEC (_("Remove object '%d': %s"),
                                    this.id, this.engine.errmsg ());
      }
      this.remove_stmt.reset ();
    }
  }
}

