/* parser.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *  Nicolas Joseph <nicolas.joseph@valaide.org>
 */

private class Valide.Symbols.Report : Vala.Report
{
  public override void err (Vala.SourceReference? source, string message)
  {
  }

  public override void warn (Vala.SourceReference? source, string message)
  {
  }
}

public class Valide.Symbols.Parser
{
  private void parse_current_line (Gtk.TextBuffer doc, out string symbolname,
                                   out int lineno, out int colno)
  {
    string line = "";
    Gtk.TextIter end;
    Gtk.TextIter start;
    string last_symbolname = null;
    unowned Gtk.TextMark mark;

    mark = doc.get_insert ();
    doc.get_iter_at_mark (out start, mark);

    lineno = start.get_line ();
    colno = start.get_line_offset ();
    symbolname = "";

    end = start;
    start.set_line_offset (0);
    line = start.get_text (end);

    if (colno > 0)
    {
      unichar ch;
      int status = 1;
      StringBuilder tmp;
      unichar prev_char;
      int string_lev = 0;
      int bracket_lev_1 = 0;
      int bracket_lev_2 = 0;
      int bracket_lev_3 = 0;

      do
      {
        ch = end.get_char ();
        if (Utils.char_is_separator (ch))
        {
          break;
        }
      } while (end.forward_char ());

      prev_char = end.get_char ();
      tmp = new StringBuilder ();
      while (true)
      {
        if (!end.backward_char ())
        {
          break; // end of buffer
        }

        if (end.get_line_offset () == 0 || end.get_line () < lineno)
        {
          break;
        }

        ch = end.get_char ();

        if (status == 1 && (ch != ' ' && ch != '\t'))
        {
          status = 0; //back to normal
        }
        if (status == 0
            && (ch == ' ' || ch == '\t'
                || (ch == '(' && bracket_lev_1 == 0)
                || (ch == '[' && bracket_lev_2 == 0)
                || (ch == '{' && bracket_lev_3 == 0)))
        {
          break; //word interruption
        }

        switch (status)
        {
          case 0: //normal state
            if (ch == ')')
            {
              status = 2;
              bracket_lev_1++;
            }
            else if (ch == ']')
            {
              status = 2;
              bracket_lev_2++;
            }
            else if (ch == '}')
            {
              status = 2;
              bracket_lev_3++;
            }
            else if (ch == '.' || (ch == '-' && prev_char == '>'))
            {
              if (last_symbolname == null)
                last_symbolname = tmp.str.reverse ();
              if (ch == '-' && prev_char == '>')
                tmp.append_unichar ('.'); //all dots
              else
                tmp.append_unichar (ch);
              status = 1; //skip spaces
            }
            else if (ch == '=')
            {
              if (last_symbolname == null)
                last_symbolname = tmp.str.reverse ();

              tmp.truncate (0);
              status = 1; //skip spaces
            }
            else if (ch == '\"')
            {
              string_lev++;
              status = 3;
            }
            else if (ch != '\n')
            {
              tmp.append_unichar (ch);
            }
          break;
          case 1: //skip spaces
          break;
          case 2: //skip to close bracket
            if (bracket_lev_1 > 0 && ch == '(')
            {
              bracket_lev_1--;
            }
            else if (bracket_lev_2 > 0 && ch == '[')
            {
              bracket_lev_2--;
            }
            else if (bracket_lev_3 > 0 && ch == '{')
            {
              bracket_lev_3--;
            }
            else if (ch == ')')
            {
              bracket_lev_1++;
            }
            else if (ch == ']')
            {
              bracket_lev_2++;
            }
            else if (ch == '{')
            {
              bracket_lev_3++;
            }
            if (bracket_lev_1 <= 0 && bracket_lev_2 <= 0 && bracket_lev_3 <= 0)
            {
              status = 1; //back to skip spaces
            }
          break;
          case 3: //inside string literal
            if (ch == '\"')
            {
              tmp.append ("string".reverse ());
              string_lev--;
              if (string_lev <= 0)
              {
                status = 0;
              }
            }
          break;
        }
        prev_char = ch;
      }
      symbolname = tmp.str.reverse ();
      if (last_symbolname == null)
      {
        last_symbolname = symbolname;
      }
    }
  }

  private bool node_contains_position (Vala.CodeNode node, int line, int column)
  {
    if ((node.source_reference.first_line < line
           && node.source_reference.last_line > line)
        || (node.source_reference.first_line == line
            && node.source_reference.first_column <= column)
        || (node.source_reference.last_line == line
            && node.source_reference.last_column <= column))
    {
      return true;
    }
    return false;
  }

  private Vala.CodeNode? find_sub_codenode (Vala.CodeNode node, int line,
                                            int column,
                                            Vala.CodeNode? parent_node)
  {
    Vala.CodeNode? result = null;

    //this is a HACK since vala set the parent_node property only for expression
    if (parent_node != null && node.parent_node == null && node != parent_node)
    {
      node.parent_node = parent_node;
    }

    if (node is Vala.Property)
    {
      Vala.Property prop;

      prop = node as Vala.Property;
      if (prop.get_accessor != null)
      {
        result = this.find_sub_codenode (prop.get_accessor, line, column, node);
        if (result != null)
        {
          return result;
        }
      }
      if (prop.set_accessor != null)
      {
        result = this.find_sub_codenode (prop.set_accessor, line, column, node);
        if (result != null)
        {
          return result;
        }
      }

      if (this.node_contains_position (prop, line, column))
      {
        return prop;
      }
    }
    else if (node is Vala.PropertyAccessor)
    {
      Vala.PropertyAccessor pa;

      pa = node as Vala.PropertyAccessor;
      result = this.find_sub_codenode (pa.body, line, column, node);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (pa, line, column))
      {
        return pa;
      }
    }
    else if (node is Vala.Constructor)
    {
      Vala.Constructor cs;

      cs = node as Vala.Constructor;
      result = this.find_sub_codenode (cs.body, line, column, node);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (cs.body, line, column)
          || this.node_contains_position (cs, line, column))
      {
        return cs;
      }
    }
    else if (node is Vala.Destructor)
    {
      Vala.Destructor ds;

      ds = node as Vala.Destructor;
      result = this.find_sub_codenode (ds.body, line, column, node);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (ds.body, line, column)
          || this.node_contains_position (ds, line, column))
      {
        return ds;
      }
    }
    else if (node is Vala.Method)
    {
      Vala.Method md;

      md = node as Vala.Method;

      if (md.body != null)
      {
        result = this.find_sub_codenode (md.body, line, column, node);
        if (result != null)
        {
          return result;
        }

        if (this.node_contains_position (md.body, line, column)
            || this.node_contains_position (md, line, column))
        {
          return md;
        }
      }
    }
    else if (node is Vala.TryStatement)
    {
      Vala.TryStatement tr;

      tr = node as Vala.TryStatement;
      result = this.find_sub_codenode (tr.body, line, column, tr);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (tr.body, line, column))
      {
        return tr.body;
      }

      if (tr.finally_body != null)
      {
        result = this.find_sub_codenode (tr.finally_body, line, column, tr);
        if (result != null)
        {
          return result;
        }

        if (this.node_contains_position (tr.finally_body, line, column))
        {
          return tr.finally_body;
        }
      }

      foreach (Vala.CatchClause clause in tr.get_catch_clauses ())
      {
        result = this.find_sub_codenode (clause, line, column, tr);
        if (result != null)
        {
          return result;
        }
      }

      if (this.node_contains_position (tr, line, column))
      {
        return tr;
      }
    }
    else if (node is Vala.CatchClause)
    {
      Vala.CatchClause cc;

      cc = node as Vala.CatchClause;
      result = this.find_sub_codenode (cc.body, line, column, cc);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (cc.body, line, column)
          || this.node_contains_position (cc, line, column))
      {
        return cc;
      }
    }
    else if (node is Vala.ExpressionStatement)
    {
      result = this.find_sub_codenode (((Vala.ExpressionStatement) node).expression, line, column, node);
      if (result != null)
      {
        return result;
      }
    }
    else if (node is Vala.LambdaExpression)
    {
      Vala.LambdaExpression lambda;

      lambda = node as Vala.LambdaExpression;
      if (this.node_contains_position (lambda.statement_body, line, column))
      {
        result = this.find_sub_codenode (lambda.statement_body, line, column, node);
        if (result != null)
        {
          return result;
        }
        return lambda;
      }
    }
    else if (node is Vala.Assignment)
    {
      result = this.find_sub_codenode (((Vala.Assignment) node).right, line, column, node);
      if (result != null)
      {
        return result;
      }

      if (this.node_contains_position (((Vala.Assignment) node).right, line, column))
      {
        return node;
      }
    }
    else if (node is Vala.IfStatement)
    {
      Vala.IfStatement ifs;

      ifs = node as Vala.IfStatement;
      if (ifs.true_statement != null)
      {
        result = this.find_sub_codenode (ifs.true_statement, line, column, node);
        if (result != null)
        {
          return result;
        }

        if  (this.node_contains_position (ifs.true_statement, line, column))
        {
          return ifs.true_statement;
        }
      }
      if (ifs.false_statement != null)
      {
        result = this.find_sub_codenode (ifs.false_statement, line, column, node);
        if (result != null)
        {
          return result;
        }

        if (this.node_contains_position (ifs.false_statement, line, column))
        {
          return ifs.false_statement;
        }
      }

      if (this.node_contains_position (ifs, line, column))
      {
        return ifs;
      }
    }
    else if (node is Vala.ForeachStatement)
    {
      Vala.ForeachStatement fe;

      fe = node as Vala.ForeachStatement;
      if (this.node_contains_position (fe.body, line, column)
          || this.node_contains_position (fe, line, column))
      {
        result = this.find_sub_codenode (fe.body, line, column, node);
        if (result != null)
        {
          return result;
        }

        return fe;
      }
    }
    else if (node is Vala.WhileStatement)
    {
      Vala.WhileStatement ws;

      ws = node as Vala.WhileStatement;
      if (this.node_contains_position (ws.body, line, column)
          || this.node_contains_position (ws, line, column))
      {
        result = this.find_sub_codenode (ws.body, line, column, node);
        if (result != null)
        {
          return result;
        }

        return ws;
      }
    }
    else if (node is Vala.ForStatement)
    {
      Vala.ForStatement fs;

      fs = node as Vala.ForStatement;
      if (this.node_contains_position (fs.body, line, column)
          || this.node_contains_position (fs, line, column))
      {
        result = this.find_sub_codenode (fs.body, line, column, node);
        if (result != null)
        {
          return result;
        }
        return fs;
      }
    }
    else if (node is Vala.Block)
    {
      Vala.Block block;

      block = node as Vala.Block;
      //check first in inner sub-nodes
      foreach (Vala.CodeNode subnode in block.get_statements ())
      {
        result = this.find_sub_codenode (subnode, line, column, node);
        if (result != null)
        {
          return result;
        }
      }
    }
    else if (node is Vala.ReturnStatement || node is Vala.DeclarationStatement
             || node is Vala.MethodCall)
    {
      return null;
    }
    else if (node is Vala.CastExpression)
    {
      if (this.node_contains_position (node, line, column))
      {
        return node;
      }
    }
    else
    {
      //debug ("incomplete support for %s", node.type_name);
    }
    return null;
  }


  private Vala.CodeNode? find_codenode (Vala.SourceFile source, int line, int column,
                                        out Vala.Class cl, out Vala.Method method)
  {
    cl = null;
    method = null;
    foreach (Vala.CodeNode node in source.get_nodes ())
    {
      Vala.CodeNode? result = null;

      method = null;
      if (node is Vala.Method)
      {
        method = node as Vala.Method;
        result = this.find_sub_codenode (node, line, column, null);
        if (result != null)
        {
          return result;
        }
      }
      else if (node is Vala.Class)
      {
        cl = node as Vala.Class;
        if (cl.constructor != null)
        {
          result = this.find_sub_codenode (cl.constructor, line, column, cl);
          if (result != null)
          {
            return result;
          }
        }

        if (cl.destructor != null)
        {
          result = this.find_sub_codenode (cl.destructor, line, column, cl);
          if (result != null)
          {
            return result;
          }
        }

        foreach (Vala.Method md in cl.get_methods ())
        {
          method = md as Vala.Method;
          result = this.find_sub_codenode (md, line, column, cl);
          if (result != null)
          {
            return result;
          }
        }
        method = null;
        foreach (Vala.Property prop in cl.get_properties ())
        {
          result = this.find_sub_codenode (prop, line, column, cl);
          if (result != null)
          {
            return result;
          }
        }
      }
    }
    cl = null;
    method = null;
    return null;
  }

  private Vala.DataType get_datatype_for_name (string symbolname,
                                               string filename, int line,
                                               int column, string content)
  {
    Vala.Class cl;
    Vala.Method md;
    Vala.Parser parser;
    Vala.SourceFile file;
    Vala.CodeContext context;
    Vala.DataType type = null;

    context = new Vala.CodeContext ();
    file = new Vala.SourceFile (context, filename,
                                filename.has_suffix (".vapi"), content);
    context.add_source_file (file);
    context.assert = false;
    context.checking = false;
    context.non_null_experimental = false;
    context.compile_only = true;
    context.report = new Report ();
    Vala.CodeContext.push (context);

    parser = new Vala.Parser ();
    parser.parse (context);

    this.find_codenode (file, line, column, out cl, out md);
    if (cl != null)
    {
      if (symbolname == "this")
      {
        type = new Vala.ClassType (cl);
      }
      else if (symbolname == "base")
      {
        if (cl.base_class != null)
        {
          type = new Vala.ClassType (cl.base_class);
        }
        else
        {
          foreach (Vala.DataType item in cl.get_base_types ())
          {
            type = item;
            break;
          }
        }
      }

      if (type == null)
      {
        //field class
        foreach (Vala.Field field in cl.get_fields ())
        {
          if (field.name == symbolname)
          {
            type = field.field_type;
            break;
          }
        }
      }

      if (type == null)
      {
        //properties
        foreach (Vala.Property prop in cl.get_properties ())
        {
          if (prop.name == symbolname)
          {
            type = prop.property_type;
            break;
          }
        }
      }

      if (type == null)
      {
        //methods
        foreach (Vala.Method clmt in cl.get_methods ())
        {
          if (clmt.name == symbolname)
          {
            type = new Vala.MethodType (clmt);
            break;
          }
        }
      }
    }
    return type;
  }

  private Vala.DataType get_current_datatype (SourceBuffer source_buffer,
                                              bool skip_last)
  {
    int col;
    int line;
    string symbolname;
    Vala.DataType type = null;

    this.parse_current_line (source_buffer, out symbolname, out line, out col);
    if (skip_last)
    {
      Utils.remove_last_element (ref symbolname, '.');
    }
    symbolname = Utils.get_last_element (symbolname, '.');
    try
    {
      type = this.get_datatype_for_name (symbolname, source_buffer.path, line,
                                         col, source_buffer.get_contents ());
    }
    catch (Error e)
    {
      debug (e.message);
    }
    return type;
  }

  private string get_qualified_name_for_datatype (Vala.DataType dt)
  {
    string typename = null;

    if (dt is Vala.ClassType)
    {
      typename = (dt as Vala.ClassType).class_symbol.get_full_name ();
    }
    else
    {
      typename = dt.to_qualified_string ();
    }

    if (typename != null)
    {
      if (typename.has_suffix ("?"))
      {
        typename = typename.substring (0, typename.length - 1);
      }
      if (typename.str ("<") != null && typename.has_suffix (">"))
      {
        //generic type definition. delete the generic part from the typename
        typename = typename.split ("<", 2)[0];
      }
      typename = typename.replace ("[]", "");
    }

    return typename;
  }

  /**
   * Get the type of the current symbol
   * Note: for methods, the type is a return type
   *
   * @param source_buffer A Valide.SourceBuffer
   *
   * @return A type
   */
  public Symbol get_current_type (SourceBuffer source_buffer)
  {
    Vala.DataType type;
    Symbol symbol = null;

    type = this.get_current_datatype (source_buffer, true);
    if (type is Vala.MethodType)
    {
      type = type.get_return_type ();
    }

    if (type != null)
    {
      string type_name;

      type_name = this.get_qualified_name_for_datatype (type);
      symbol = new Symbol.from_full_name (type_name);
      if (symbol.id == -1)
      {
        symbol = null;
      }
    }
    return symbol;
  }

  /**
   * Get the current symbol
   *
   * @param source_buffer A Valide.SourceBuffer
   *
   * @return A symbol
   */
  public Symbol get_current_symbol (SourceBuffer source_buffer)
  {
    Vala.DataType type;
    Symbol symbol = null;

    type = this.get_current_datatype (source_buffer, false);
    if (type != null)
    {
      string type_name;

      type_name = this.get_qualified_name_for_datatype (type);
      symbol = new Symbol.from_full_name (type_name);
      if (symbol.id == -1)
      {
        symbol = null;
      }
    }
    return symbol;
  }
}

