/* symbol.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public class Valide.Symbols.Symbol : Object
{
  private static HashTable<string, Gdk.Pixbuf> icons;

  /**
   * The symbol name
   */
  public string name;

  /**
   * The symbol full name
   */
  public string full_name;

  /**
   * The file where the symbol is defined
   */
  public File file;

  /**
   * The line where the symbol is defined
   */
  public int file_position;

  /**
   * The symbol signature
   */
  public string signature;

  /**
   * The symbol root
   */
  public Symbol root;

  /**
   * The symbol kind
   */
  public Kind kind;

  /**
   * The symbol access
   */
  public Access access;

  /**
   * The symbol implementation
   */
  public Implementation implementation;

  /**
   * The symbol inheritance
   */
  public string[] inherits;

  /**
   * The symbol typeref
   */
  public string typeref;

  /**
   * The symbol icon
   */
  public Gdk.Pixbuf icon
  {
    owned get
    {
      Gdk.Pixbuf pixbuf;

      pixbuf = this.icons.lookup (kind.name);
      if (pixbuf == null)
      {
        try
        {
          string filename;
          string path;

          filename = "%s.png".printf (kind.name);
          path = Path.build_filename (Config.PIXMAPS_DIR, "symbols", filename);
          pixbuf = new Gdk.Pixbuf.from_file (path);
          this.icons.insert (kind.name, pixbuf);
        }
        catch (Error e)
        {
          debug ("Could not load pixbuf: %s\n", e.message);
          pixbuf = Utils.get_pixbuf_for_stock (Gtk.STOCK_MISSING_IMAGE, Gtk.IconSize.MENU);
        }
      }
      return pixbuf;
    }
  }

  /**
   * @see Valide.Symbols.Object.table_name
   */
  protected override string table_name
  {
    owned get
    {
      return "symbol";
    }
  }

  /**
   * @see Valide.Symbols.Object.add_query
   */
  protected override string add_query
  {
    owned get
    {
      return "INSERT INTO symbol (name, full_name, file_id, file_position, signature, root_id, kind_id, access_id, implementation_id, typeref) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?);";
    }
  }

  /**
   * @see Valide.Symbols.Object.exist_query
   */
  protected override string exist_query
  {
    owned get
    {
      return "SELECT id FROM symbol WHERE name = ? AND file_id = ? AND file_position = ?;";
    }
  }

  /**
   * @see Valide.Symbols.Object.bind_exist_stmt
   */
  protected override void bind_exist_stmt ()
  {
    this.exist_stmt.bind_text (1, this.name);
    if (this.file != null)
    {
      this.exist_stmt.bind_int (2, this.file.id);
    }
    else
    {
      this.exist_stmt.bind_int (2, -1);
    }
    this.exist_stmt.bind_int (3, this.file_position);
  }

  private void _new_from_id (int id)
  {
    if (id > 0)
    {
      string sql;
      Sqlite.Statement statement;

      sql = """SELECT id, name, full_name, file_id, file_position, signature,
                      root_id, kind_id, access_id, implementation_id, typeref
                 FROM symbol
                 WHERE id = ?;""";
      statement = this.engine.prepare (sql);
      statement.bind_int (1, id);
      if (statement.step () == Sqlite.ROW)
      {
        this.id = statement.column_int (0);
        this.name = statement.column_text (1);
        this.full_name = statement.column_text (2);
        this.file = new File.from_id (statement.column_int (3));
        this.file_position = statement.column_int (4);
        this.signature = statement.column_text (5);
        if (statement.column_int (6) != -1)
        {
          this.root = new Symbol.from_id (statement.column_int (6));
        }
        this.kind = new Kind.from_id (statement.column_int (7));
        this.access = new Access.from_id (statement.column_int (8));
        this.implementation = new Implementation.from_id (statement.column_int (9));
        this.inherits = Heritage.get_derived (this);
        this.typeref = statement.column_text (10);
      }
    }
  }

  construct
  {
    if (this.icons == null)
    {
      this.icons = new HashTable<string, Gdk.Pixbuf> (str_hash, str_equal);
    }
  }

  /**
   * Create a new symbol from id
   *
   * @param id The symbol id
   */
  public Symbol.from_id (int id)
  {
    this._new_from_id (id);
  }

  /**
   * Create a new symbol from full name
   *
   * @param full_name The symbol full name
   */
  public Symbol.from_full_name (string full_name)
  {
    string sql;
    Sqlite.Statement statement;

    sql = """SELECT id
               FROM symbol
               WHERE full_name = ?;""";
    statement = this.engine.prepare (sql);
    statement.bind_text (1, full_name);
    if (statement.step () == Sqlite.ROW)
    {
      this._new_from_id (statement.column_int (0));
    }
  }

  /**
   * @see Valide.Symbols.Object.add
   */
  public override int add () throws EngineError
  {
    int file_id = -1;

    return_if_fail (this.file != null);
    return_if_fail (this.kind != null);
    return_if_fail (this.access != null);

    if (!this.file.exist ())
    {
      this.file.add ();
    }
    else
    {
      file_id = this.file.id;
    }

    if (this.name != null && this.name != ""
        && this.full_name != null && this.full_name != ""
        && !this.exist ())
    {
      int kind_id = -1;
      int root_id = -1;
      int access_id = -1;
      int implementation_id = -1;

      if (!this.kind.exist ())
      {
        kind_id = this.kind.add ();
      }
      else
      {
        kind_id = this.kind.id;
      }
      if (!this.access.exist ())
      {
        access_id = this.access.add ();
      }
      else
      {
        access_id = this.access.id;
      }
      if (this.implementation != null)
      {
        if (!this.implementation.exist ())
        {
          implementation_id = this.implementation.add ();
        }
        else
        {
          implementation_id = this.implementation.id;
        }
      }
      if (this.signature == null)
      {
        this.signature = "";
      }
      if (this.root != null)
      {
        root_id = this.root.id;
      }

      this.add_stmt.bind_text (1, this.name);
      this.add_stmt.bind_text (2, this.full_name);
      this.add_stmt.bind_int (3, file_id);
      this.add_stmt.bind_int (4, this.file_position);
      this.add_stmt.bind_text (5, this.signature);
      this.add_stmt.bind_int (6, root_id);
      this.add_stmt.bind_int (7, kind_id);
      this.add_stmt.bind_int (8, access_id);
      this.add_stmt.bind_int (9, implementation_id);
      this.add_stmt.bind_text (10, this.typeref);
      if (this.add_stmt.step () == Sqlite.DONE)
      {
        this.id = this.get_last_id ();

        foreach (string base_class in inherits)
        {
          Heritage heritage;

          heritage = new Heritage (this.id, base_class);
          heritage.add ();
        }
      }
      else
      {
        throw new EngineError.EXEC (_("Add symbol '%s': %s"),
                                    this.full_name, this.engine.errmsg ());
      }
      this.add_stmt.reset ();
    }
    return this.id;
  }

  /**
   * Get the symbol prototype
   *
   * @return The propotype
   */
  public string get_prototype ()
  {
    string prototype;

    if (this.kind.name == "method" || this.kind.name == "delegate"
        || this.kind.name == "signal")
    {
      prototype = "%s %s %s".printf (this.typeref, this.full_name, this.signature);
    }
    else
    {
      prototype = this.full_name;
    }
    return prototype;
  }
}

