/* window.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

private class Valide.FullscreenToolbar : Gtk.Window
{
  private bool animation_enter;
  private uint animation_timeout_id;

  public new Window parent
  {
    get;
    construct;
  }

  private bool on_enter_notify_event (Gtk.Widget sender, Gdk.EventCrossing event)
  {
    this.show_hide (true, 0);
    return false;
  }

  private bool on_leave_notify_event (Gtk.Widget sender, Gdk.EventCrossing event)
  {
    int w, h;
    int x, y;
    Gdk.Screen screen;
    Gdk.Display display;

    display = Gdk.Display.get_default ();
    screen = this.parent.get_screen ();

    this.get_size (out w, out h);
    display.get_pointer (out screen, out x, out y, null);

    /*
     * gtk seems to emit leave notify when clicking on tool items,
     * work around it by checking the coordinates
     */
    if (y >= h)
    {
      this.show_hide (false, h);
    }

    return false;
  }

  private bool run_animation ()
  {
    int x, y;
    Gdk.Screen screen;
    Gdk.Rectangle fs_rect;

    screen = this.parent.get_screen ();
    screen.get_monitor_geometry (screen.get_monitor_at_window (this.parent.window),
                                 out fs_rect);

    this.get_position (out x, out y);

    if (this.animation_enter)
    {
      if (y == fs_rect.y)
      {
        this.animation_timeout_id = 0;
        return false;
      }
      else
      {
        this.move (x, y + 1);
        return true;
      }
    }
    else
    {
      int w, h;

      this.get_size (out w, out h);

      if (y == fs_rect.y - h + 1)
      {
        this.animation_timeout_id = 0;
        return false;
      }
      else
      {
        this.move (x, y - 1);
        return true;
      }
    }
  }

  private void show_hide (bool show, int height)
  {
    Gtk.Settings settings;
    bool enable_animations;

    settings = this.parent.get_settings ();
    enable_animations = settings.gtk_enable_animations;

    if (enable_animations)
    {
      this.animation_enter = show;

      if (this.animation_timeout_id == 0)
      {
        this.animation_timeout_id = Timeout.add (4, this.run_animation);
      }
    }
    else
    {
      Gdk.Screen screen;
      Gdk.Rectangle fs_rect;

      screen = this.parent.get_screen ();
      screen.get_monitor_geometry (screen.get_monitor_at_window (this.parent.window),
                                   out fs_rect);

      if (show)
      {
        this.move (fs_rect.x, fs_rect.y);
      }
      else
      {
        this.move (fs_rect.x, fs_rect.y - height + 1);
      }
    }

  }

  public FullscreenToolbar (Window parent)
  {
    this.parent = parent;
  }

  construct
  {
    Gtk.Window window;
    Gtk.Action action;
    Gtk.Toolbar toolbar;

    this.type = Gtk.WindowType.POPUP;
    this.set_transient_for (this.parent);
    
    toolbar = this.parent.ui_manager.get_widget ("/fullscreen-toolbar") as Gtk.Toolbar;
    toolbar.set_style (Gtk.ToolbarStyle.BOTH_HORIZ);
    this.add (toolbar);

    action = this.parent.ui_manager.get_action ("leave-fullscreen");
    action.set ("is_important", true);

    this.enter_notify_event.connect (this.on_enter_notify_event);
    this.leave_notify_event.connect (this.on_leave_notify_event);
  }

  public new void show ()
  {
    this.show_all ();
  }

  public new void show_all ()
  {
    int w, h;
    Gdk.Screen screen;
    Gdk.Rectangle fs_rect;

    screen = this.parent.get_screen ();
    screen.get_monitor_geometry (screen.get_monitor_at_window (this.parent.window),
                                 out fs_rect);

    this.get_size (out w, out h);
    this.resize (fs_rect.width, h);
    this.move (fs_rect.x, fs_rect.y - h + 1);
    base.show_all ();
  }
}

/**
 * The main window of the application
 */
public class Valide.Window : Gtk.Window
{
  public uint workspace;

  private uint status_context;
  private StartPage start_page;
  private Gtk.WindowGroup owned_group;
  private Gdk.WindowState window_state;
  private FullscreenToolbar fullscreen_toolbar;

  /**
   * The UI manager
   */
  public UIManager ui_manager
  {
    get;
    private set;
  }

  /**
   * The recent file manager
   */
  public RecentManagerUI recent_manager
  {
    get;
    private set;
  }

  /**
   * The config manager
   */
  public ConfigManager config_manager
  {
    get;
    private set;
  }

  /**
   * The document manager
   */
  public DocumentManager documents
  {
    get;
    private set;
  }

  /**
   * The project manager
   */
  public ProjectManager projects
  {
    get;
    private set;
  }

  /**
   * The executable manager
   */
  public ExecutableManager executables
  {
    get;
    private set;
  }

  /**
   * The plugin manager
   */
  public PluginManager plugins
  {
    get;
    private set;
  }

  /**
   * the symbol cache
   */
  public Symbols.Cache symbols
  {
    get;
    private set;
  }

  /**
   * The completion manager
   */
  public Completion.Manager completion
  {
    get;
    private set;
  }

  /**
   * The menu bar
   */
  public Gtk.MenuBar menu
  {
    get;
    private set;
  }

  /**
   * The tool bar
   */
  public Gtk.Toolbar toolbar
  {
    get;
    private set;
  }

  /**
   * The side panel
   */
  public Panel side_panel
  {
    get;
    private set;
  }

  /**
   * The bottom panel
   */
  public Panel bottom_panel
  {
    get;
    private set;
  }

  /**
   * The statusbar
   */
  public Statusbar statusbar
  {
    get;
    private set;
  }

  private void message (string? log_domain, LogLevelFlags log_levels,
                        string message)
  {
    Gtk.Dialog dialog;
    Gtk.MessageType type;

    switch (log_levels)
    {
      case LogLevelFlags.LEVEL_MESSAGE:
        type = Gtk.MessageType.INFO;
      break;
      case LogLevelFlags.LEVEL_WARNING:
        type = Gtk.MessageType.WARNING;
      break;
      case LogLevelFlags.LEVEL_CRITICAL:
        type = Gtk.MessageType.ERROR;
      break;
      default:
        type = Gtk.MessageType.OTHER;
      break;
    }

    dialog = new Gtk.MessageDialog (this, Gtk.DialogFlags.MODAL, type,
                                    Gtk.ButtonsType.OK, message);
    dialog.run ();
    dialog.destroy ();
  }

  private bool on_delete_event ()
  {
    int width;
    int height;

    this.get_size (out width, out height);
    this.config_manager.set_integer ("General", "width", width);
    this.config_manager.set_integer ("General", "height", height);
    try
    {
      this.executables.stop ();
    }
    catch (Error e)
    {
      debug (e.message);
    }
    return !this.documents.close_all ();
  }

  private void menu_item_select_cb (Gtk.Item item)
  {
    string message;
    Gtk.Action action;

    action = item.get_data ("gtk-action") as Gtk.Action;
    return_if_fail (action != null);

    action.get ("tooltip", out message);
    if (message != null)
    {
      this.statusbar.push (this.status_context, message);
    }
  }

  private void menu_item_deselect_cb (Gtk.Item item)
  {
    this.statusbar.pop (this.status_context);
  }

  private void connect_proxy_cb (Gtk.UIManager sender, Gtk.Action action,
                                 Gtk.Widget proxy)
  {
    if (proxy is Gtk.MenuItem)
    {
      Gtk.MenuItem item;

      item = proxy as Gtk.MenuItem;
      item.select.connect (this.menu_item_select_cb);
      item.deselect.connect (this.menu_item_deselect_cb);
    }
  }

  private void disconnect_proxy_cb (Gtk.UIManager sender, Gtk.Action action,
                                    Gtk.Widget proxy)
  {
    if (proxy is Gtk.MenuItem)
    {
      Gtk.MenuItem item;

      item = proxy as Gtk.MenuItem;
      item.select.disconnect (this.menu_item_select_cb);
      item.deselect.disconnect (this.menu_item_deselect_cb);
    }
  }

  private void setup_ui_manager ()
  {
    Gtk.ToggleAction action;

    this.ui_manager = new UIManager ();
    this.ui_manager.connect_proxy.connect (this.connect_proxy_cb);
    this.ui_manager.disconnect_proxy.connect (this.disconnect_proxy_cb);
    this.ui_manager.populate ();

    this.menu = this.ui_manager.get_widget ("/menubar") as Gtk.MenuBar;
    this.toolbar = this.ui_manager.get_widget ("/toolbar") as Gtk.Toolbar;

    this.ui_manager.action_set_toggled ("view-statusbar", true);
    this.ui_manager.action_set_toggled ("view-toolbar", true);
    this.ui_manager.action_set_toggled ("view-side-panel", true);
    this.ui_manager.action_set_toggled ("view-bottom-panel", true);

    this.add_accel_group (this.ui_manager.get_accel_group ());
  }

  private void populate_window ()
  {
    Gtk.VBox main_box;
    Gtk.VPaned vpaned;
    Gtk.HPaned hpaned;

    main_box = new Gtk.VBox (false, 0);
    this.add (main_box);

    this.recent_manager.update_file ();
    this.recent_manager.update_project ();

    main_box.pack_start (this.menu, false, true, 0);
    main_box.pack_start (this.toolbar, false, true, 0);
    this.fullscreen_toolbar = new FullscreenToolbar (this);

    vpaned = new Gtk.VPaned ();
    main_box.pack_start (vpaned, true, true, 0);

    this.bottom_panel = new Panel (Gtk.Orientation.HORIZONTAL);
    this.bottom_panel.add_item_with_stock_icon (this.executables, _("Messages"),
                                                Gtk.STOCK_EXECUTE);

    this.bottom_panel.hide.connect (() => {
      this.ui_manager.action_set_toggled ("view-bottom-panel", false);
    });

    vpaned.pack2 (this.bottom_panel, false, false);

    hpaned = new Gtk.HPaned ();
    vpaned.pack1 (hpaned, true, true);

    this.side_panel = new Panel (Gtk.Orientation.VERTICAL);
    this.side_panel.add_item (this.projects, _("Project"),
                              new Gtk.Image.from_file (Path.build_filename (Config.PIXMAPS_DIR, "icone-16.xpm")));
    this.side_panel.hide.connect (() => {
      this.ui_manager.action_set_toggled ("view-side-panel", false);
    });

    hpaned.add1 (this.side_panel);
    hpaned.add2 (this.documents);

    main_box.pack_start (this.statusbar, false, true, 0);
  }

  private void on_action_activated (Gtk.UIManager sender, Gtk.Action action)
  {
    try
    {
      switch (action.name)
      {
        /* File menu */
        case "file-new":
          if (this.projects.current != null)
          {
            this.projects.current.create_file ();
          }
          else
          {
            this.documents.create ();
          }
        break;
        case "file-open":
          this.documents.open ();
        break;
        case "file-save":
          if (this.documents.current != null)
          {
            string current_folder = null;

            if (this.projects.current != null)
            {
              current_folder = this.projects.current.path;
            }
            this.documents.current.save (current_folder);
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "file-save-as":
          if (this.documents.current != null)
          {
            string current_folder = null;

            if (this.projects.current != null)
            {
              current_folder = this.projects.current.path;
            }
            this.documents.current.save_as (current_folder);
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "file-save-all":
          if (this.documents.current != null)
          {
            this.documents.save_all ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "file-rename":
          if (this.documents.current != null)
          {
            this.documents.current.rename ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "file-close":
          if (this.documents.current != null)
          {
            this.documents.current.close ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "file-close-all":
          if (this.documents.current != null)
          {
            this.documents.close_all ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "quit":
          if (!this.on_delete_event ())
          {
            this.destroy ();
          }
        break;

        /* Edit menu */
        case "edit-undo":
          Gtk.bindings_activate (this.get_focus (), Gdk.KeySyms.z,
                                 Gdk.ModifierType.CONTROL_MASK);
        break;
        case "edit-redo":
          /* @bug No key binding found */
/*
          Gtk.bindings_activate (this.get_focus (), Gdk.KeySyms.z,
                                 Gdk.ModifierType.CONTROL_MASK
                                 | Gdk.ModifierType.SHIFT_MASK);
*/
          if (this.documents.current != null)
          {
            this.documents.current.redo ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-copy":
          Gtk.bindings_activate (this.get_focus (), Gdk.KeySyms.c,
                                 Gdk.ModifierType.CONTROL_MASK);
        break;
        case "edit-cut":
          Gtk.bindings_activate (this.get_focus (), Gdk.KeySyms.x,
                                 Gdk.ModifierType.CONTROL_MASK);
        break;
        case "edit-paste":
          Gtk.bindings_activate (this.get_focus (), Gdk.KeySyms.v,
                                 Gdk.ModifierType.CONTROL_MASK);
        break;
        case "edit-search":
          if (this.documents.current != null)
          {
            this.documents.current.search ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-find-next":
          if (this.documents.current != null)
          {
            this.documents.current.find_next ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-find-prev":
          if (this.documents.current != null)
          {
            this.documents.current.find_prev ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-replace":
          if (this.documents.current != null)
          {
            this.documents.current.replace ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-goto-line":
          if (this.documents.current != null)
          {
            this.documents.current.goto_line ();
          }
          else
          {
            throw new DocumentError.NO_DOCUMENT (_("No document open!"));
          }
        break;
        case "edit-goto-definition":
          this.symbols.goto_definition (this.documents);
        break;
        case "edit-preferences":
          config_manager.preferences (this);
        break;

        /* View menu */
        case "view-start-page":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.show_start_page ();
          }
          else
          {
            this.hide_start_page ();
          }
        break;
        case "view-toolbar":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.toolbar.show ();
          }
          else
          {
            this.toolbar.hide ();
          }
        break;
        case "view-statusbar":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.statusbar.show ();
          }
          else
          {
            this.statusbar.hide ();
          }
        break;
        case "view-side-panel":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.side_panel.show ();
          }
          else
          {
            this.side_panel.hide ();
          }
        break;
        case "view-bottom-panel":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.bottom_panel.show ();
          }
          else
          {
            this.bottom_panel.hide ();
          }
        break;
        case "view-fullscreen":
          Gtk.ToggleAction toggle_action;

          toggle_action = action as Gtk.ToggleAction;
          if (toggle_action.active)
          {
            this.fullscreen ();
          }
          else
          {
            this.unfullscreen ();
          }
        break;

        /* Project menu */
        case "project-new":
          this.projects.create ();
        break;
        case "project-open":
          this.projects.open ();
        break;
        case "project-add":
          if (this.projects.current != null)
          {
            this.projects.current.add_file ();
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-remove":
          if (this.projects.current != null)
          {
            this.projects.current.remove_file ();
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-options":
          if (this.projects.current != null)
          {
            this.projects.current.options (this);
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-compile":
          if (this.projects.current != null)
          {
            this.projects.current.compile (this.executables);
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-execute":
          if (this.projects.current != null)
          {
            this.projects.current.execute (this.executables);
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-abort":
          this.executables.stop ();
        break;
        case "project-close":
          if (this.projects.current != null)
          {
            this.projects.current.close ();
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;

        /* Help menu */
        case "help-about":
          AboutDialog dialog = new AboutDialog ();
          dialog.run ();
          dialog.destroy ();
        break;

        /* Project popup */
        case "project-file-new":
          if (this.projects.current != null)
          {
            this.projects.current.create_file ();
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-file-open":
          this.projects.open_selected_file ();
        break;
        case "project-file-add":
          if (this.projects.current != null)
          {
            this.projects.current.add_file ();
          }
          else
          {
            throw new ProjectError.NO_PROJECT (_("No project open!"));
          }
        break;
        case "project-file-remove":
          this.projects.remove_selected_file ();
        break;

        /* Fullscreen toolbar */
        case "leave-fullscreen":
          this.unfullscreen ();
          this.ui_manager.action_set_toggled ("view-fullscreen", false);
        break;
      }
    }
    catch (Error e)
    {
      warning (e.message);
    }
  }

  private void update_title ()
  {
    string title = "";
    iDocument document;
    Project project;

    document = this.documents.current;
    if (document != null)
    {
      if (document.path != "")
      {
        string dir;
        string filename;
        string home;

        dir = Path.get_dirname (document.path);
        home = Environment.get_home_dir ();
        if (dir.has_prefix (home))
        {
          dir = "~" + dir.substring (home.size (),
                                     dir.size () - home.size ());
        }

        filename = Path.get_basename (document.path);
        title += "%s (%s) - ".printf (filename, dir);
      }
    }

    project = this.projects.current;
    if (project != null)
    {
      if (project.name != "")
      {
        title += project.name + " - ";
      }
    }

    title += "Val(a)IDE";
    this.set_title (title);
  }

  private bool is_fullscreen ()
  {
    return ((this.window_state & Gdk.WindowState.FULLSCREEN) != 0);
  }

  construct
  {
    this.owned_group = new Gtk.WindowGroup ();
    this.group = this.owned_group;
    this.group.add_window (this);

    this.delete_event.connect (this.on_delete_event);

    this.window_state_event.connect ((s, e) => {
      this.window_state = e.new_window_state;
      if ((e.changed_mask &
          (Gdk.WindowState.MAXIMIZED | Gdk.WindowState.FULLSCREEN)) != 0)
      {
        bool full_screen;

        full_screen = ((e.new_window_state &
          (Gdk.WindowState.MAXIMIZED | Gdk.WindowState.FULLSCREEN)) != 0);

        this.statusbar.set_has_resize_grip (!full_screen);
        this.config_manager.set_boolean ("General", "full-screen", full_screen);
      }
    });

    this.setup_ui_manager ();

    this.config_manager = ConfigManager.get_instance ();

    this.documents = new DocumentManager (this.ui_manager);
    this.documents.tab_changed.connect (this.update_title);
    this.documents.tab_state_changed.connect (this.update_title);
    this.documents.tab_added.connect (this.update_title);
    this.documents.tab_removed.connect (this.update_title);
    this.documents.tab_cursor_moved.connect ((s, d, r, c) => {
      this.statusbar.set_cursor_position (r, c);
    });

    this.projects = new ProjectManager (this.ui_manager, this.documents);
    this.projects.project_opened.connect (() => {
      this.side_panel_show (this.projects);
    });
    this.projects.project_opened.connect (this.update_title);
    this.projects.project_closed.connect (this.update_title);

    this.recent_manager = new RecentManagerUI (this.ui_manager, this.documents,
                                               this.projects);

    this.executables = new ExecutableManager (this.ui_manager);
    this.executables.run_exec.connect (() => {
      this.bottom_panel_show (this.executables);
    });

    this.symbols = Symbols.Cache.get_instance ();
    this.completion = new Completion.Manager (this.documents);

    this.statusbar = new Statusbar ();
    this.status_context = this.statusbar.get_context_id ("valide_window");

    this.start_page = new StartPage (this.documents, this.projects,
                                     this.recent_manager);
    this.start_page.closed.connect (() => {
      this.hide_start_page ();
    });
    this.start_page.send_message.connect ((s, m) => {
      this.statusbar.flash_message (this.status_context, m);
    });

    this.ui_manager.post_activate.connect (this.on_action_activated);
    this.populate_window ();
    this.update_title ();

    try
    {
      List<Gdk.Pixbuf> list;

      list.append (new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "icone-16.xpm")));
      list.append (new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "icone-32.xpm")));
      list.append (new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "icone-48.xpm")));
      list.append (new Gdk.Pixbuf.from_file (Path.build_filename (Config.PIXMAPS_DIR, "icone-64.xpm")));
      Gtk.Window.set_default_icon_list (list);
    }
    catch (Error e)
    {
      debug (e.message);
    }
  }

  /**
   * Launch the application
   */
  public void run ()
  {
    bool full_screen;

    this.plugins = new PluginManager (this);
    this.show_all ();
    this.show_start_page ();
    this.fullscreen_toolbar.hide ();

    full_screen = this.config_manager.get_boolean ("General", "full-screen");
    if (full_screen)
    {
      this.maximize ();
    }
    else
    {
      int width;
      int height;

      width = this.config_manager.get_integer ("General", "width");
      height = this.config_manager.get_integer ("General", "height");
      if (width > 1 && height > 1)
      {
        this.resize (width, height);
      }
    }
    Log.set_handler (null, LogLevelFlags.LEVEL_WARNING
                           | LogLevelFlags.LEVEL_MESSAGE
                           | LogLevelFlags.LEVEL_CRITICAL, this.message);
  }

  /**
   * Show the side panel
   *
   * @param widget The widget to show, or null for the current widget
   */
  public void side_panel_show (Gtk.Widget? widget = null)
  {
    if (widget != null)
    {
      this.side_panel.set_current_page (widget);
    }
    this.side_panel.show ();
    this.ui_manager.action_set_toggled ("view-side-panel", true);
  }

  /**
   * Hide the side panel
   */
  public void side_panel_hide ()
  {
    this.side_panel.hide ();
    this.ui_manager.action_set_toggled ("view-side-panel", false);
  }

  /**
   * Show the bottom panel
   *
   * @param widget The widget to show, or null for the current widget
   */
  public void bottom_panel_show (Gtk.Widget? widget = null)
  {
    if (widget != null)
    {
      this.bottom_panel.set_current_page (widget);
    }
    this.bottom_panel.show ();
    this.ui_manager.action_set_toggled ("view-bottom-panel", true);
  }

  /**
   * Hide the bottom panel
   */
  public void bottom_panel_hide ()
  {
    this.bottom_panel.hide ();
    this.ui_manager.action_set_toggled ("view-bottom-panel", false);
  }

  /**
   * @see Gtk.Window.fullscreen
   */
  public new void fullscreen ()
  {
    if (!this.is_fullscreen ())
    {
      base.fullscreen ();
      this.menu.hide ();
      this.toolbar.hide ();
      this.statusbar.hide ();
      this.fullscreen_toolbar.show ();
    }
  }

  /**
   * @see Gtk.Window.unfullscreen
   */
  public new void unfullscreen ()
  {
    if (this.is_fullscreen ())
    {
      base.unfullscreen ();
      this.menu.show ();
      this.toolbar.show ();
      this.statusbar.show ();
      this.fullscreen_toolbar.hide ();
    }
  }

  /**
   * Show the start page
   */
  public void show_start_page ()
  {
    int page_num;

    page_num = this.documents.page_num (this.start_page);
    if (page_num < 0)
    {
      this.documents.append_page (this.start_page);
      this.start_page.show_all ();
    }
    this.documents.current = this.start_page;
    this.ui_manager.action_set_toggled ("view-start-page", true);
  }

  /**
   * Hide the start page
   */
  public void hide_start_page ()
  {
    this.documents.remove_page (this.start_page);
    this.ui_manager.action_set_toggled ("view-start-page", false);
  }
}

