/* opened-documents.vala
 *
 * Copyright (C) 2009 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

using Valide;

public class OpenedDocuments : Plugin
{
  /* enum for tree model columns, not all is currently used */
  enum Col
  {
    ICON,
    NAME,
    DOCUMENT,
    STYLE,
    TOOLTIP,
    NB
  }

  private Gtk.VBox box;
  private Gtk.TreeView tree_view;
  private DocumentManager documents;

  private bool populate_asc ()
  {
    Gtk.TreeIter iter;
    Gtk.ListStore model;

    model = this.tree_view.get_model () as Gtk.ListStore;
    model.clear ();

    foreach (iDocument document in this.documents.documents)
    {
      string tooltip;
      Pango.Style style;

      if (document.is_save)
      {
        style = Pango.Style.NORMAL;
      }
      else
      {
        style = Pango.Style.ITALIC;
      }
      tooltip = document.get_tooltips ();
      model.append (out iter);
      model.set (iter, Col.ICON, document.get_icon (),
                 Col.NAME, document.filename,
                 Col.DOCUMENT, document,
                 Col.STYLE, style,
                 Col.TOOLTIP, tooltip);
    }
    this.on_tab_changed ();
    return false;
  }

  private void populate ()
  {
    Idle.add (this.populate_asc);
  }

  private void on_cursor_change (Gtk.TreeView sender)
  {
    iDocument document;
    Gtk.TreeModel model;
    Gtk.TreeIter iter;
    Gtk.TreeSelection selection;

    selection = this.tree_view.get_selection ();

    if (selection.get_selected (out model, out iter))
    {
      model.get (iter, Col.DOCUMENT, out document);
      this.documents.current = document;
    }
  }

  private void on_tab_changed ()
  {
    Gtk.TreeIter iter;
    iDocument document;
    Gtk.ListStore list_store;

    list_store = this.tree_view.get_model () as Gtk.ListStore;
    if (list_store.get_iter_first (out iter))
    {
      do
      {
        list_store.get (iter, Col.DOCUMENT, out document);
        if (document == this.documents.current)
        {
          this.tree_view.get_selection ().select_iter (iter);
          break;
        }
      } while (list_store.iter_next (ref iter));
    }
  }

  construct
  {
    Gtk.Image image;
    Gtk.ListStore store;
    Gtk.TreeViewColumn col;
    Gtk.CellRenderer renderer;
    Gtk.ScrolledWindow scrolled;

    this.box = new Gtk.VBox (false, 0);

    /* create some nice padding */
    this.box.set_spacing (2);
    this.box.set_border_width (2);
 
    /* create and pack tree view */
    scrolled = new Gtk.ScrolledWindow (null, null);
    scrolled.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    scrolled.set_shadow_type (Gtk.ShadowType.IN);
    this.box.pack_start (scrolled, true, true, 0);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.set_rules_hint (false);
    this.tree_view.set_headers_visible (false);
    scrolled.add (this.tree_view);

    /* create model */
    store = new Gtk.ListStore (Col.NB, typeof (Gdk.Pixbuf), typeof (string),
                               typeof (iDocument), typeof (Pango.Style),
                               typeof (string));
    store.set_sort_column_id (Col.NAME, Gtk.SortType.ASCENDING);

    /* main column */
    col = new Gtk.TreeViewColumn ();
    col.set_title(_("File"));

    renderer = new Gtk.CellRendererPixbuf ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "pixbuf", Col.ICON);

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "markup", Col.NAME, "style", Col.STYLE);

    this.tree_view.append_column (col);
    this.tree_view.set_model (store);

    this.tree_view.set_tooltip_column (Col.TOOLTIP);

    /* attach signal callbacks for tree view */
    this.tree_view.cursor_changed.connect (this.on_cursor_change);

    this.documents = this.window.documents;

    this.documents.tab_removed.connect (this.populate);
    this.documents.tab_added.connect (this.populate);
    this.documents.tab_state_changed.connect (this.populate);
    this.documents.tab_changed.connect (this.on_tab_changed);

    image = new Gtk.Image.from_file (Path.build_filename (this.path,
                                                          "opened-documents.png"));
    this.window.side_panel.add_item (this.box, _("Documents"), image);

    this.populate ();
    this.box.show_all ();
  }

  ~OpenedDocuments ()
  {
    this.documents.tab_removed.disconnect (this.populate);
    this.documents.tab_added.disconnect (this.populate);
    this.documents.tab_state_changed.disconnect (this.populate);
    this.documents.tab_changed.disconnect (this.on_tab_changed);
    this.window.side_panel.remove_item (this.box);
  }
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (OpenedDocuments);
}

