/* symbol-browser.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/*
 * This code is re-writing in the vala language of GeditSymbolBrowserPlugin
 *
 * Copyright (C) 2007 - Micah Carrick
 */

using Valide;

public class SymbolBrowser : Plugin
{
  /* enum for tree model columns, not all is currently used */
  enum Col
  {
    ICON,
    NAME,
    FILE,
    LINE,
    LANGUAGE,
    NB
  }

  private Gtk.VBox box;
  private Gtk.TreeView tree_view;
  private Gtk.TreeStore model;
  private Gtk.Label label;
  private SList<string> expanded_rows;
  private uint status_context;
  private Gtk.TreeViewColumn language_column;
  private Gtk.TreeViewColumn source_file_column;
  private Gtk.TreeViewColumn line_number_column;
  private HashTable<string, Gtk.TreeIter*> root_nodes;
  private int atom;

  private bool show_language_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-language");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-language",
                                                 value);
      this.language_column.visible = value;
    }
  }

  private bool show_source_file_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-source-file");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-source-file",
                                                 value);
      this.source_file_column.visible = value;
    }
  }

  private bool show_line_number_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-line-number");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-line-number",
                                                 value);
      this.line_number_column.visible = value;
    }
  }

  private bool only_load_active_tab
  {
    get
    {
      return true;
    }
  }

  private bool hierarchic_view
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "hierarchic-view");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "hierarchic-view", value);
      this.populate_symbols ();
    }
  }

  private void save_expanded_rows ()
  {
/*
        GConfClient             *client;
        GError                  *error=NULL;
        
        client = gconf_client_get_default ();
        
        gconf_client_set_list (client, 
                        GEDIT_SYMBOL_BROWSER_PLUGIN_GCONF_PATH"expanded-rows",
                        GCONF_VALUE_STRING, panel->priv->expanded_rows,
                        &error);
                        
        if (error != NULL)
        {
                g_warning ("Error saving gedit-symbol-browser-plugin setting: %s", 
                                error->message);
                                
                g_error_free (error);
        }

        g_object_unref (client);
*/
  }

  private bool is_expanded_row (string name)
  {
    bool ret = false;

    foreach (string i in this.expanded_rows)
    {
      if (i == name)
      {
        ret = true;
        break;
      }
    }
    return ret;
  }

  private void add_expanded_row (string name)
  {
    if (!this.is_expanded_row (name))
    {
      this.expanded_rows.prepend (name);
      this.save_expanded_rows ();
    }
  }

  private void remove_expanded_row (string name)
  {
    weak SList<string> l;

    if (this.is_expanded_row (name) == true)
    {
      for (l = this.expanded_rows; l != null; l = l.next)
      {
        if (l.data == name)
        {
          this.expanded_rows.remove_all (l.data);
        }
      }
      this.save_expanded_rows ();
    }
  }

  private void clear_data ()
  {
    this.model.clear ();
    this.root_nodes = new HashTable<string, Gtk.TreeIter*> (str_hash, str_equal);
    this.box.sensitive = false;
    this.label.set_text ("");
  }

  /**
   * Add a single symbol to the treeview. Parents rows are added as the type
   * whenever they aren't already present.
   */
  private bool add_symbol (Symbols.Symbol symbol)
  {
    string markup;
    Gdk.Pixbuf pixbuf;
    Gtk.TreeIter iter;
    Gtk.TreeIter* parent;
    Gtk.TreeView tree_view;

    Utils.process_gtk_events ();
    symbol.name = symbol.name.replace ("<", "&lt;");
    symbol.name = symbol.name.replace (">", "&gt;");

    pixbuf = symbol.icon;

    if (!this.hierarchic_view)
    {
      /* we could do bold here */
      markup = "<b>%s</b>".printf (symbol.kind.name);

      /* check for existing parent iter */
      parent = this.root_nodes.lookup (markup);
      /* create the parent if it wasn't already found */
      if (parent == null)
      {
        Gtk.TreeIter tmp;

        this.model.append (out tmp, null);
        parent = (Gtk.TreeIter*)tmp.copy ();
        this.model.set (*parent, Col.ICON, pixbuf, Col.NAME, markup);
        this.root_nodes.insert (markup, parent);
      }

      this.model.append (out iter, *parent);
      /* add the new symbol */
      this.model.set (iter, Col.ICON, pixbuf,
                            Col.NAME, symbol.name,
                            Col.FILE, symbol.file.path,
                            Col.LINE, symbol.file_position,
                            Col.LANGUAGE, symbol.file.language.name);
    }
    else
    {
      if (symbol.root != null)
      {
        parent = this.root_nodes.lookup (symbol.root.full_name);
        this.model.append (out iter, *parent);
      }
      else
      {
        this.model.append (out iter, null);
        parent = (Gtk.TreeIter*)iter.copy ();
      }
      this.model.set (iter, Col.ICON, pixbuf,
                            Col.NAME, symbol.name,
                            Col.FILE, symbol.file.path,
                            Col.LINE, symbol.file_position,
                            Col.LANGUAGE, symbol.file.language.name);
      this.root_nodes.insert (symbol.full_name, (Gtk.TreeIter*)iter.copy ());
    }
    return true;
  }

  private bool populate_symbols_asc ()
  {
    Gtk.TreeStore model;

    if (AtomicInt.compare_and_exchange (ref this.atom, 0, 1))
    {
      this.model = this.tree_view.get_model () as Gtk.TreeStore;
      this.tree_view.set_model (null);
      this.tree_view.freeze_child_notify ();
      if (this.window.documents.current != null)
      {
        List<iDocument> doc;
        Symbols.Symbol root = null;

        doc = new List<iDocument> ();
        /* find out if we want to load all symbols or just active tab */
        if (this.only_load_active_tab)
        {
          doc.prepend (this.window.documents.current);
        }
        else
        {
          doc = this.window.documents.documents.copy ();
        }

        this.clear_data ();
        /* run ctags on each document in the list and add symbols */
        foreach (iDocument d in doc)
        {
          if (!this.hierarchic_view)
          {
            this.window.symbols.get_symbols_for_file (d.path, Symbols.SortType.KIND, this.add_symbol);
          }
          else
          {
            this.window.symbols.get_symbols_for_file (d.path, Symbols.SortType.SCOPE, this.add_symbol);
          }
        }
        if (this.model.iter_n_children (null) > 0)
        {
          this.box.sensitive = true;
        }
      }
      else
      {
        this.clear_data ();
      }
      this.tree_view.set_model (this.model);
      this.tree_view.thaw_child_notify ();
//      this.restore_expanded_rows ();
      this.tree_view.expand_all ();
      AtomicInt.set (ref this.atom, 0);
    }
    return false;
  }

  /**
   * Call to re-populate tree view.
   */
  private void populate_symbols ()
  {
    Idle.add (this.populate_symbols_asc);
  }

  /**
   * Re-expand any row when the list is repopulated if that row was previously
   * expanded.
   */
  private void restore_expanded_rows ()
  {
    /* is this going to fix the crash? */
    if (this.window.documents.sensitive == true)
    {
      Gtk.TreeIter iter;
      Gtk.TreeModel model = this.tree_view.get_model ();

      if (model.get_iter_first (out iter) == true)
      {
        for (bool r = model.get_iter_first (out iter); 
             r == true;
             r = model.iter_next (ref iter))
        {
          string name;

          model.get (iter, Col.NAME, out name,  -1);
          if (this.is_expanded_row (name))
          {
            Gtk.TreePath path = model.get_path (iter);
            this.tree_view.expand_row (path, false);
          }
        }
      }
    }
  }

  private void on_tab_changed ()
  {
    /* only needed in single-file mode  */
    if (this.only_load_active_tab)
    {
      this.populate_symbols ();
    }
  }

  private void on_row_activated (Gtk.TreeView sender, Gtk.TreePath path,
                                 Gtk.TreeViewColumn column)
  {
    Gtk.TreeIter iter;
        
    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get_iter (out iter, path);

    /* expand/collapse row if it has children (and return) */
    if (!this.hierarchic_view && model.iter_has_child (iter))
    {
      /* @bug http://bugzilla.gnome.org/show_bug.cgi?id=536480 */
      if (Fix.GtkTreeView.row_is_expanded (this.tree_view, path))
      {
        this.tree_view.collapse_row (path);
      }
      else
      {
        this.tree_view.expand_row (path, false);
      }
    }
    else
    {
      /* otherwise, show the file name and line number and flash statusbar */
      int line;
      string name;
      string filename;

      model.get (iter, Col.NAME, out name,  Col.LINE, out line,
                 Col.FILE, out filename);
      try
      {
        iDocument document;

        document = this.window.documents.create (filename);
        document.goto_line (line - 1);
        document.grab_focus ();
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
  }

  private void on_cursor_changed (Gtk.TreeView sender)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;

    model = this.tree_view.get_model ();
    selection = this.tree_view.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      int line;
      string name;
      string filename;

      model.get (iter, Col.NAME, out name, Col.FILE, out filename,
                       Col.LINE, out line);

			if (filename != null)
			{
		    string label_text;
		    string status_message;

		    label_text = "%s Line:%d".printf (Path.get_basename (filename), line);
		    this.label.set_text (label_text);

		    status_message = "Double-click to jump to '%s'".printf (name);
		    this.window.statusbar.flash_message (this.status_context, status_message);
		  }
    }
  }

  private void on_row_expanded (Gtk.TreeView sender, Gtk.TreeIter iter,
                                Gtk.TreePath path)
  {
    string name;        

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    this.add_expanded_row (name);
  }

  private void on_row_collapsed (Gtk.TreeView sender, Gtk.TreeIter iter,
                                 Gtk.TreePath path)
  {
    string name;

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    remove_expanded_row (name);
  }

  construct
  {
    Gtk.ScrolledWindow scrolled;
    Gtk.TreeStore store;
    Gtk.TreeViewColumn col;
    Gtk.CellRendererText renderer;
    Gtk.CellRendererPixbuf renderer_pixbuf;
    Gtk.Image image;

    AtomicInt.set (ref this.atom, 0);

    this.box = new Gtk.VBox (false, 0);

    /* create some nice padding */
    this.box.set_spacing (2);
    this.box.set_border_width (2);
 
    /* create and pack tree view */
    scrolled = new Gtk.ScrolledWindow (null, null);
    scrolled.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    scrolled.set_shadow_type (Gtk.ShadowType.IN);
    this.box.pack_start (scrolled, true, true, 0);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.set_rules_hint (false);
    this.tree_view.set_headers_visible (true);
    this.tree_view.set_headers_clickable (true);
    scrolled.add (this.tree_view);

    /* create and pack label */
    this.label = new Gtk.Label ("");
    this.label.set_alignment ((float)0.0, (float)0.5);
    this.box.pack_start (this.label, false, true, 0);

    /* create model */
    store = new Gtk.TreeStore (Col.NB, typeof (Gdk.Pixbuf), typeof (string),
                               typeof (string), typeof (int), typeof (string),
                               typeof (Object));
    //store.set_sort_column_id (1, Gtk.SortType.ASCENDING);

    /* main column */
    col = new Gtk.TreeViewColumn ();
    col.set_title(_("Symbol"));

    renderer_pixbuf = new Gtk.CellRendererPixbuf ();
    col.pack_start (renderer_pixbuf, false);
    col.set_attributes (renderer_pixbuf, "pixbuf", Col.ICON);

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "markup", Col.NAME);
    this.tree_view.append_column (col);

    /* additional columns */

    this.language_column = new Gtk.TreeViewColumn ();
    this.show_language_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-language");
    this.language_column.set_title (_("Language"));
    renderer = new Gtk.CellRendererText ();
    this.language_column.pack_start (renderer, true);
    this.language_column.set_attributes (renderer, "text", Col.LANGUAGE);
    this.tree_view.append_column (this.language_column);

    this.source_file_column = new Gtk.TreeViewColumn ();
    this.show_source_file_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-source-file");
    this.source_file_column.set_title (_("File"));
    renderer = new Gtk.CellRendererText ();
    this.source_file_column.pack_start (renderer, true);
    this.source_file_column.set_attributes (renderer, "text", Col.FILE);
    this.tree_view.append_column (this.source_file_column);

    this.line_number_column = new Gtk.TreeViewColumn ();
    this.show_line_number_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-line-number");
    this.line_number_column.set_title (_("Line"));
    renderer = new Gtk.CellRendererText ();
    this.line_number_column.pack_start (renderer, true);
    this.line_number_column.set_attributes (renderer, "text", Col.LINE);
    this.tree_view.append_column (this.line_number_column);

    this.tree_view.set_model (store);

    /* attach signal callbacks for tree view */
    this.tree_view.row_activated.connect (on_row_activated);
    this.tree_view.cursor_changed.connect (on_cursor_changed);
    this.tree_view.row_expanded.connect (on_row_expanded);
    this.tree_view.row_collapsed.connect (on_row_collapsed);

    this.status_context = this.window.statusbar.get_context_id ("valide_symbol_browser");

    this.box.sensitive = false;

    this.window.documents.tab_removed.connect (this.populate_symbols);
    this.window.documents.tab_changed.connect (this.on_tab_changed);

    this.window.symbols.cache_updated.connect (this.populate_symbols);

    image = new Gtk.Image.from_file (Path.build_filename (this.path,
                                                          "symbol-browser.png"));

    this.window.side_panel.add_item (this.box, "Symbol Browser", image);

    this.box.show_all ();
  }

  public SymbolBrowser (string path)
  {
    this.path = path;
  }

  ~SymbolBrowser ()
  {
    this.window.documents.tab_removed.disconnect (this.populate_symbols);
    this.window.documents.tab_changed.disconnect (this.on_tab_changed);
    this.window.symbols.cache_updated.disconnect (this.populate_symbols);
    this.window.side_panel.remove_item (this.box);
  }

	public Gtk.Widget create_configure_dialog ()
	{
	  Gtk.Frame frame;
	  Gtk.VBox vbox;
	  Gtk.CheckButton check;

	  frame = new Gtk.Frame (_("Additionnal columns"));

	  vbox = new Gtk.VBox (true, 0);
	  frame.add (vbox);

	  check = new Gtk.CheckButton.with_label (_("Programming language"));
	  check.set_active (this.show_language_column);
	  check.toggled.connect ((s) => {
	    this.show_language_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Source file"));
	  check.set_active (this.show_source_file_column);
	  check.toggled.connect ((s) => {
	    this.show_source_file_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Line number"));
	  check.set_active (this.show_line_number_column);
	  check.toggled.connect ((s) => {
	    this.show_line_number_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Hierarchic view"));
	  check.set_active (this.hierarchic_view);
	  check.toggled.connect ((s) => {
	    this.hierarchic_view = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  return frame;
	}
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (SymbolBrowser);
}

public Gtk.Widget create_configure_dialog (SymbolBrowser self)
{
  return self.create_configure_dialog ();
}

