/* main.vala
 *
 * Copyright (C) 2008-2009 Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/*string encoding_charset;*/
/*bool new_window_option;*/
bool new_document_option;
bool show_version;
string[] remaining_args;

const OptionEntry[] options = {
{ "version", 'V', 0, OptionArg.NONE, ref show_version, N_("Show the application's version"), null },
/* { "encoding", '\0', 0, OptionArg.STRING, ref encoding_charset, N_("Set the character encoding to be used to open the files listed on the command line"), N_("ENCODING")}, */
/*{ "new-window", '\0', 0, OptionArg.NONE, ref new_window_option, N_("Create a new toplevel window in an existing instance of valide"), null },*/
{ "new-document", '\0', 0, OptionArg.NONE, ref new_document_option, N_("Create a new document in an existing instance of valide"), null },
/* "" = G_OPTION_REMAINING */
{ "", '\0', 0, OptionArg.FILENAME_ARRAY, ref remaining_args, null, N_("[FILE...]") }, /* collects file arguments */
{null}
};

public class Valide.Application : Object
{
  private bool show_version;
  private bool new_document;
  private string[] file_list;
  private HashTable<string, Window> windows;

#if (UNIQUE)
  private Unique.Response message (Unique.App sender, int command,
                                   Unique.MessageData data, uint time_)
  {
    string ws;
    Window window;
    uint workspace;
    bool opened = false;
    Unique.Response res = Unique.Response.INVALID;

    workspace = data.get_workspace ();
    /* The first instance doesn't set the workspace */
    if (workspace == 0)
    {
      workspace = Unique.Backend.create ().get_workspace ();
    }
    ws = workspace.to_string ();
    window = this.windows.lookup (ws);
    if (window == null)
    {
      window = this.create_new_window (workspace);
      this.windows.insert (ws, window);
    }
    switch (command)
    {
      case Unique.Command.ACTIVATE:
        weak Gdk.Screen screen = data.get_screen();
        if(screen != null) {
          window.set_screen (screen);
        }
        window.present ();
        res = Unique.Response.OK;
      break;
      case Unique.Command.NEW:
        window.documents.create ();
        opened = true;
        res = Unique.Response.OK;
      break;
      case Unique.Command.OPEN:
        string[] uris;

        uris = data.get_uris ();
        if (uris != null)
        {
          for (int i = 0; uris[i] != null; i++)
          {
            try
            {
              if (window.projects.accept_file (uris[i]))
              {
                window.projects.open (uris[i]);
              }
              else
              {
                window.documents.create (uris[i]);
              }
              opened = true;
            }
            catch (Error e)
            {
              debug (e.message);
            }
          }
         }
        res = Unique.Response.OK;
      break;
    }
    if (opened)
    {
      window.hide_start_page ();
    }
    return res;
  }
#endif

  private Window create_new_window (uint workspace)
  {
    Window window;

    window = new Window ();
    window.workspace = workspace;
    window.destroy.connect ((s) => {
      string ws;

      ws = (s as Window).workspace.to_string ();
      this.windows.remove (ws);
      if (this.windows.size () == 0)
      {
        Gtk.main_quit ();
        Xml.Parser.cleanup ();
      }
    });

    window.run ();
    return window;
  }

  public Application (string[] args)
  {
    OptionContext context;

    Xml.Parser.init ();
    Gtk.init (ref args);
    Intl.bindtextdomain (Config.GETTEXT_PACKAGE, Config.LOCALE_DIR);
    Intl.bind_textdomain_codeset (Config.GETTEXT_PACKAGE, "UTF-8");

    this.windows = new HashTable<string, Window> (str_hash, str_equal);

    /* Setup command line options */
    context = new OptionContext (_("- Integrated Development Environment for Vala"));
    context.add_main_entries (options, Config.GETTEXT_PACKAGE);
    context.add_group (Gtk.get_option_group (false));

    try
    {
      context.parse (ref args);
      if (remaining_args != null)
      {
        for (int i = 0; remaining_args[i] != null; i++)
        {
          this.file_list += remaining_args[i];
        }
      }
      this.show_version = show_version;
      this.new_document = new_document;
    }
    catch (OptionError e)
    {
      print (_("%s\nRun '%s --help' to see a full list of available command line options.\n"),
             e.message, args[0]);
    }
  }

#if (!UNIQUE)
  public int run ()
  {
    if (this.show_version)
    {
      print ("%s - Version %s\n", Environment.get_application_name (),
                                  Config.VERSION);
    }
    else
    {
      Window window;

      window = this.create_new_window (0);
      Gtk.main ();
    }
    return 0;
  }
#else
  public int run ()
  {
    Unique.App app;

    if (this.show_version)
    {
      print ("%s - Version %s\n", Environment.get_application_name (),
                                  Config.VERSION);
    }
    else
    {
      app = new Unique.App ("org.valide.valaide", null);
      if (app.is_running)
      {
        Unique.Command command;
        Unique.MessageData data;

        data = new Unique.MessageData ();
        if (this.new_document)
        {
          command = Unique.Command.NEW;
        }
        else if (this.file_list != null)
        {
          command = Unique.Command.OPEN;
          data.set_uris (this.file_list);
        }
        else
        {
          command = Unique.Command.ACTIVATE;
        }
        app.send_message (command, data);
      }
      else
      {
        Unique.Command command;
        Unique.MessageData data;

        data = new Unique.MessageData ();
        if (this.new_document)
        {
          command = Unique.Command.NEW;
        }
        else if (this.file_list != null)
        {
          command = Unique.Command.OPEN;
          data.set_uris (this.file_list);
        }
        else
        {
          command = Unique.Command.ACTIVATE;
        }
        this.message (app, command, data, 0);
        app.message_received.connect (message);
        Gtk.main ();
      }
    }
    return 0;
  }
#endif
}

int main (string[] args)
{
  return new Valide.Application (args).run ();
}

